/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend-bibtex.c defines the bibtex backend of refdbd
  markus@mhoenicka.de 8-8-00
  $Id: backend-bibtex.c,v 1.16 2004/01/29 01:42:06 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <stdio.h>
#include <dbi/dbi.h> /* database abstraction layer */

#include "refdb.h"
#include "linklist.h"
#include "backend.h"
#include "strfncs.h"
#include "tokenize.h"
#include "refdbd.h" /* for struct ADDRESULT */
#include "backend-bibtex.h"
#include "backend-dbib.h"
#include "dbfncs.h"

/* some globals */
extern struct BSTRING outomem;
extern struct BSTRING outomem_n;
extern struct BSTRING connerr;
extern struct BSTRING connerr_n;
extern struct BSTRING positive;
extern struct BSTRING negative;

/*  extern char server_ip[]; */ /* the IP address of the database server */
extern int n_log_level; /* numeric version of log_level */

/* the style information if load_style() was successful */
extern dbi_result dbi_style_res;

/* forward declarations of local functions */
static char* bibtexify_author(char** ptr_bibauthor, char* author);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_bibtex(): writes a header for the bibtex output of a
                         query

  char prepare_render_bibtex returns a ptr to the result string
                             if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* prepare_render_bibtex(struct renderinfo* ptr_rendinfo) {
  /* we just make sure that we start with a clean string */
  (*(ptr_rendinfo->ptr_ref))[0] = '\0';
  return *(ptr_rendinfo->ptr_ref);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_bibtex(): writes a footer for the bibtex output of a query

  char* finish_render_bibtex returns ptr to the buffer if successful,
                             NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* finish_render_bibtex(struct renderinfo* ptr_rendinfo) {
  char* new_ref;

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return new_ref;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_bibtex() renders a RIS dataset for bibtex output

  char* render_bibtex returns ptr to the buffer if successful,
                             NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* render_bibtex(struct renderinfo* ptr_rendinfo) {
  int nperiodical_type;
  int sql_command_len;
  int error;
  int i; /* what the hell is this? */
  int errcode; /* receives error code when retrieving periodical */
  int n_multiple_db = 0; /* if 1, driver supports multiple databases per conn */
  unsigned long long n_id;
  char have_author = 0;
  char *new_ref;
  char *bibauthor;
  char *sql_command; /* buffer for assembling SQL queries */
  char *quoted_journal;
  const char *item;
  const char *drivername;
  const char *type;
  char date_buffer[256];
  char id[32];
  dbi_result dbires;
  dbi_result dbirescit;
  struct BIBCONNS bibconns;
  dbi_driver driver;
  struct REPRINT reprint;

  bibconns.conn = NULL;
  bibconns.conn_refdb = NULL;
  bibconns.conn_source = NULL;

  if ((type = get_refdb_type(ptr_rendinfo->dbires)) == NULL) {
    return NULL;
  }

  bibconns.conn = dbi_result_get_conn(ptr_rendinfo->dbires);
  bibconns.conn_source = bibconns.conn;

  driver = dbi_conn_get_driver(bibconns.conn);
  drivername = dbi_driver_get_name(driver);

  if (ptr_rendinfo->nuse_citestyle) {
    sql_command_len = 512;
    sql_command = (char*)malloc(sql_command_len);
    if (!sql_command) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }

    /* retrieve basic information about the bibliography style from CITSTYLE into ptr_result */
    quoted_journal = mstrdup((ptr_rendinfo->ptr_biblio_info)->format_string);
    if (!dbi_driver_quote_string(driver, &quoted_journal)) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      free(sql_command);
      return NULL;
    }

    /* db capability: multiple databases per connection */
    if (!strcmp(my_dbi_conn_get_cap(bibconns.conn, "multiple_db"), "t")) {
      bibconns.conn_refdb = bibconns.conn;
      n_multiple_db = 1;
      sprintf(sql_command, "SELECT ID from %s.CITSTYLE where JOURNAL=%s", MAIN_DB, quoted_journal);
    }
    else {
      sprintf(sql_command, "SELECT ID from CITSTYLE where JOURNAL=%s", quoted_journal);
      /* need another connection to the database refdb */
      if ((bibconns.conn_refdb = connect_to_db(ptr_rendinfo->ptr_clrequest, MAIN_DB, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, connerr.text);
	free(sql_command);
	return NULL;
      }
    }

    free(quoted_journal);
    dbirescit = dbi_conn_query(bibconns.conn_refdb, sql_command);

    LOG_PRINT(LOG_DEBUG, sql_command);
    free(sql_command);

    if (!dbirescit) {
      LOG_PRINT(LOG_WARNING, "CITSTYLE select failed");
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }
  
    if (dbi_result_next_row(dbirescit)) {
      n_id = my_dbi_result_get_idval_idx(dbirescit, 1);
      if (!load_style(type, n_id, bibconns.conn_refdb)) {
	LOG_PRINT(LOG_WARNING, "could not load style information");
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
    }
    else {
      LOG_PRINT(LOG_WARNING, "could not load style information");
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }

    LOG_PRINT(LOG_DEBUG, "style information loaded");
  }

  /* the first entries should always fit, don't hassle with mstrcat() */
  /*----------------------------------------------------------------*/
  /* type */
  /* the bibtex backend uses the following mapping between RIS types and
     bibtex types */
  /* ABST  @misc
     ADVS  @misc
     ART   @misc
     BILL  @misc
     BOOK  @book
     CASE  @misc
     CHAP  @inbook
     COMP  @misc
     CONF  @proceedings
     CTLG  @misc
     DATA  @misc
     ELEC  @misc
     GEN   @misc
     HEAR  @proceedings
     ICOMM @misc
     INPR  @misc
     JFULL @article
     JOUR  @article
     MAP   @misc
     MGZN  @article
     MPCT  @misc
     MUSIC @misc
     NEWS  @misc
     PAMP  @misc
     PAT   @misc
     PCOMM @misc
     RPRT  @techreport
     SER   @book
     SLIDE @misc
     SOUND @misc
     STAT  @misc
     THES  @phdthesis (could be mastersthesis as well, ambiguous!)
     UNBILL@misc
     UNPB  @unpublished
     VIDEO @misc */
     
  if (strcmp(type, "JOUR") == 0 || strcmp(type, "JFULL") == 0 || strcmp(type, "MGZN") == 0) {
/*     LOG_PRINT(LOG_DEBUG, "start JOUR"); */
    strcat(*(ptr_rendinfo->ptr_ref), "@ARTICLE{ ");
/*     LOG_PRINT(LOG_DEBUG, "done JOUR"); */
  }
  else if (strcmp(type, "BOOK") == 0 || strcmp(type, "SER") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@BOOK{ ");
  }
  else if (strcmp(type, "CHAP") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@INBOOK{ ");
  }
  else if (strcmp(type, "THES") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@PHDTHESIS{ ");
  }
  else if (strcmp(type, "CONF") == 0 || strcmp(type, "HEAR") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@PROCEEDINGS{ ");
  }
  else if (strcmp(type, "RPRT") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@TECHREPORT{ ");
  }
  else if (strcmp(type, "UNPB") == 0) {
    strcat(*(ptr_rendinfo->ptr_ref), "@UNPUBLISHED{ ");
  }
  else {
    strcat(*(ptr_rendinfo->ptr_ref), "@MISC{ ");
  }

  /*----------------------------------------------------------------*/
  /* ID */
/*   if (ptr_rendinfo->dbname) { */
/*     strcat(*(ptr_rendinfo->ptr_ref), ptr_rendinfo->dbname); */
/*     strcat(*(ptr_rendinfo->ptr_ref), "-ID"); */
/*   } */
/*   else { */
/*     strcat(*(ptr_rendinfo->ptr_ref), "ID"); */
/*   } */
/*   get_refdb_id(ptr_rendinfo->dbires, id); */
/*   if (*id) { */
/*     strcat(*(ptr_rendinfo->ptr_ref), id); */ /* ID */
/*     strcat(*(ptr_rendinfo->ptr_ref), ",\n"); */
  item = get_refdb_citekey(ptr_rendinfo->dbires);
  if (item && *item) {
    strcat(*(ptr_rendinfo->ptr_ref), item); /* citekey */
    strcat(*(ptr_rendinfo->ptr_ref), ",\n");
  }
  else {
    unload_style(); /* will do no harm if load_style() was not called so we
		       don't have to wrap it with a conditional each time */
    if (!n_multiple_db) {
      dbi_conn_close(bibconns.conn_refdb);
    }
    return NULL;
  }

  /*----------------------------------------------------------------*/
  /* pubyear */
  item = get_refdb_pubyear(ptr_rendinfo->dbires, date_buffer);
  if (item != NULL) {
    strcat(*(ptr_rendinfo->ptr_ref), "  YEAR = {");
    strcat(*(ptr_rendinfo->ptr_ref), date_buffer); /* year */
    strcat(*(ptr_rendinfo->ptr_ref), "},\n");
  }

  /*----------------------------------------------------------------*/
  /* authors */
  /* we loop here to first create authors and then editors if needed */
  /* I'm at a loss what to do with tertiary authors (series editors)
     at this point. They don't seem to fit well into the BibTeX model */

  for (i = 0; i < 2; i++) {
    have_author = 0;

    dbires = request_authors(bibconns.conn, i+1, NULL, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"));

    if (!dbires) {
      unload_style();
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }

    /* we're done if there are no authors or editors to list */
    if (dbi_result_get_numrows(dbires) == 0) {
      clean_request(dbires);
      continue;
    }

    if (i == 0) { /* authors */
      new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  AUTHOR = {", ptr_rendinfo->ptr_ref_len, 0);
    }
    else { /* editors */
      new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  EDITOR = {", ptr_rendinfo->ptr_ref_len, 0);
    }
    if (new_ref == NULL) {
      clean_request(dbires);
      unload_style();
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    /* fetch all authors of this article */
    while ((item = get_author(dbires)) != NULL) {
      have_author = 1; /* at least one author is present */


      if (bibtexify_author(&bibauthor, (char*)item) == NULL) {
	clean_request(dbires);
	unload_style();
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), bibauthor, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	clean_request(dbires);
	free(bibauthor);
	unload_style();
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	free(bibauthor);
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " and ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	clean_request(dbires);
	unload_style();
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    clean_request(dbires);

    if (have_author) { /* eliminate the trailing and after the last author */
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 5)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    else {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ",\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* title */
  item = get_refdb_title_copy(ptr_rendinfo->dbires);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  TITLE = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      unload_style();
      LOG_PRINT(LOG_WARNING, outomem.text);
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* title */
      unload_style();
      LOG_PRINT(LOG_WARNING, outomem.text);
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      unload_style();
      LOG_PRINT(LOG_WARNING, outomem.text);
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    free((char*)item);
  }

  if (strcmp(type, "BOOK") == 0 || strcmp(type, "CHAP") == 0 ||
      strcmp(type, "SER") == 0 || strcmp(type, "THES") == 0 ||
      strcmp(type, "RPRT") == 0) {
    /*----------------------------------------------------------------*/
    /* book title */
    item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  BOOKTITLE = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* book title */
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      free((char*)item);
    }

    /* series title */
    item = get_refdb_title_series_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  SERIES = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* series title */
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      free((char*)item);
    }
  }
  else {
  /*----------------------------------------------------------------*/
  /* periodical */

    if (ptr_rendinfo->nuse_citestyle) { /* do some formatting */
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  JOURNAL = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) { 
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((new_ref = format_journalname(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, &bibconns, ptr_rendinfo->dbires, dbi_style_res, ptr_rendinfo->database, 3/* without markup */)) == NULL) {
	LOG_PRINT(LOG_WARNING, "JOURNALNAME formatting failed");
	unload_style();
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
    }
    else { /* plain output */
      nperiodical_type = 4; /* this retrieves the first available periodical
			       name in the order full-abbrev-cust1-cust2. This
			       may not always be what we want so this should
			       be beefed up somehow */

      /* fetch journal of this article */
      if ((item = get_periodical(bibconns.conn, date_buffer, NULL, nperiodical_type, &errcode, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"))) != NULL) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  JOURNAL = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      unload_style();
      LOG_PRINT(LOG_WARNING, outomem.text);
      if (!n_multiple_db) {
	dbi_conn_close(bibconns.conn_refdb);
      }
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  
    /*----------------------------------------------------------------*/
    /* volume */
    item = get_refdb_volume(ptr_rendinfo->dbires);
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  VOLUME = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* volume */
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  
    /*----------------------------------------------------------------*/
    /* issue */
    /* ToDo: integrate into volume? */
    item = get_refdb_issue(ptr_rendinfo->dbires);  
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  NUMBER = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* issue */
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	unload_style();
	LOG_PRINT(LOG_WARNING, outomem.text);
	if (!n_multiple_db) {
	  dbi_conn_close(bibconns.conn_refdb);
	}
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /* we're done using the style info. unload and close conn */
  unload_style();
  if (!n_multiple_db) {
    dbi_conn_close(bibconns.conn_refdb);
  }

  /*----------------------------------------------------------------*/
  /* startpage, endpage */
  item = get_refdb_startpage(ptr_rendinfo->dbires);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "  PAGES = {", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* start page */
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    item = get_refdb_endpage(ptr_rendinfo->dbires);
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "-", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* end page */
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "},\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /* eliminate the trailing comma and finish with an empty line */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n}\n", ptr_rendinfo->ptr_ref_len, 2)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }
/*      printf("%s\n", ref); */

  return *(ptr_rendinfo->ptr_ref);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  gettexbib(): executes client command gettexbib

  int gettexbib returns 0 on success and 1 on error

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct bibinfo* ptr_biblio_info ptr to a structure containing formatting
                  information

  struct ADDRESULT* ptr_addresult ptr to a structure that will receive
                   the result of the command (number of successful
                   and failed references)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int gettexbib(struct CLIENT_REQUEST* ptr_clrequest, struct bibinfo* ptr_biblio_info, struct ADDRESULT* ptr_addresult) {
  int numbyte; /* number of bytes written */
  int error;
  int type;
  int result;
  int not_done = 1;
  size_t n_bufsize;
  size_t nsql_command_len = 4096; /* should be enough */
  char *sql_command = NULL;
  char *id_list = NULL;
  char *token = NULL;
  char *token_start = NULL;
  char *eof_string = NULL;
  char *new_sql_command = NULL;
  char *dbstring;
  char *idstring;
  char *the_db;
  char *drivername;
  char prev_db[DBNAME_LENGTH] = "";
  dbi_conn conn;
/*   dbi_result dbires; */
  struct lilimem sentinel;
  struct renderinfo rendinfo;

  rendinfo.username = ptr_clrequest->username;
  rendinfo.database = ptr_clrequest->current_db;
  rendinfo.ptr_biblio_info = ptr_biblio_info;
  rendinfo.ref_format = -1;
  rendinfo.nuse_citestyle = 1;
  rendinfo.pdfroot = NULL;
  rendinfo.ptr_clrequest = ptr_clrequest;
  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  ptr_addresult->success = 0;
  ptr_addresult->failure = 0;

  n_bufsize = (size_t)atoi(ptr_clrequest->argument);

  id_list = (char*)malloc(n_bufsize);
  if (id_list == NULL || insert_lilimem(&sentinel, (void**)&id_list, NULL)) {
    delete_all_lilimem(&sentinel);
    iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
    LOG_PRINT(LOG_CRIT, outomem.text);
    return 1;
  }

  sql_command = (char*)malloc(nsql_command_len);
  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    delete_all_lilimem(&sentinel);
    iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
    LOG_PRINT(LOG_CRIT, outomem.text);
    return 1;
  }

  /* send acknowledgement to client */
  numbyte = tiwrite(ptr_clrequest->fd, positive.text, TERM_YES);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, "could not write to client");
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* read whitespace-separated id list from client*/
  numbyte = tread(ptr_clrequest->fd, id_list, n_bufsize);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, "could not read from client");
    delete_all_lilimem(&sentinel);
    return 1;
  }

  token = id_list; 
  eof_string = &id_list[strlen(id_list)]; /* ptr to end of string */

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
    LOG_PRINT(LOG_ERR, "could not connect to database server");
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* keep a copy of drivername as we may have to switch connections */
  drivername = mstrdup((char*)dbi_driver_get_name(dbi_conn_get_driver(conn)));
  if (drivername == NULL || insert_lilimem(&sentinel, (void**)&drivername, NULL)) {
    delete_all_lilimem(&sentinel);
    iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
    LOG_PRINT(LOG_CRIT, outomem.text);
    return 1;
  }

  while (not_done) { 
    /* search for the start of a token */
    while (*token < 33 && token < eof_string) { /* all whitespace has ASCII values 32 or less */
      token++;
    }
    
    if (token == eof_string){
      break;
    }

    token_start = token; /* save start of token */

    /* Look for the end */
    while (*token > 32 && token < eof_string) {
      token++;
    }
    
    if (token == eof_string){
      not_done = 0;
    }
    else {
      *token = '\0'; /* terminate token */
      token++;
    }

/*     idstring = getid(token_start, &dbstring, &type); */ /* retrieve database and id parts */
/*     if (idstring == NULL) { */
/*       continue; */ /* invalid ID string, try the next one */
/*     } */
    dbstring = NULL;
    idstring = token_start;

    if (dbstring == NULL) {
      the_db = ptr_clrequest->current_db;
    }
    else {
      the_db = dbstring;
    }

    /* prepare query string */
/*     sprintf(sql_command, "SELECT DISTINCT t_refdb.refdb_id, t_refdb.refdb_citekey, t_refdb.refdb_type, t_refdb.refdb_pubyear, t_refdb.refdb_startpage, t_refdb.refdb_endpage, t_refdb.refdb_abstract, t_refdb.refdb_title, t_refdb.refdb_volume, t_refdb.refdb_issue, t_refdb.refdb_booktitle, t_refdb.refdb_city, t_refdb.refdb_publisher, t_refdb.refdb_title_series, t_refdb.refdb_address, t_refdb.refdb_url, t_refdb.refdb_issn, t_refdb.refdb_periodical_id, t_refdb.refdb_pyother_info, t_refdb.refdb_secyear, t_refdb.refdb_secother_info, t_refdb.refdb_user1, t_refdb.refdb_user2, t_refdb.refdb_user3, t_refdb.refdb_user4, t_refdb.refdb_user5, t_refdb.refdb_misc1, t_refdb.refdb_misc2, t_refdb.refdb_misc3 FROM t_refdb WHERE t_refdb.refdb_id=%s", idstring); */

    sprintf(sql_command, "SELECT DISTINCT t_refdb.refdb_id, t_refdb.refdb_citekey, t_refdb.refdb_type, t_refdb.refdb_pubyear, t_refdb.refdb_startpage, t_refdb.refdb_endpage, t_refdb.refdb_abstract, t_refdb.refdb_title, t_refdb.refdb_volume, t_refdb.refdb_issue, t_refdb.refdb_booktitle, t_refdb.refdb_city, t_refdb.refdb_publisher, t_refdb.refdb_title_series, t_refdb.refdb_address, t_refdb.refdb_url, t_refdb.refdb_issn, t_refdb.refdb_periodical_id, t_refdb.refdb_pyother_info, t_refdb.refdb_secyear, t_refdb.refdb_secother_info, t_refdb.refdb_user1, t_refdb.refdb_user2, t_refdb.refdb_user3, t_refdb.refdb_user4, t_refdb.refdb_user5, t_refdb.refdb_misc1, t_refdb.refdb_misc2, t_refdb.refdb_misc3 FROM t_refdb WHERE t_refdb.refdb_citekey='%s'", idstring);

    if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
      /* change the database if necessary. The string comparison is presumably
	 faster than a forced db switch in each cycle */
      if (strcmp(prev_db, the_db)) {
	if (dbi_conn_select_db(conn, the_db)) {
	  iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  LOG_PRINT(LOG_WARNING, "could not connect to database");
	  return 1;
	}
      }
      strcpy(prev_db, the_db); /* save current db for next cycle */
    }
    else {
      if (strcmp(prev_db, the_db)) {
	dbi_conn_close(conn);
	if ((conn = connect_to_db(ptr_clrequest, the_db, 0)) == NULL) {
	  iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
	  LOG_PRINT(LOG_ERR, "could not connect to database server");
	  delete_all_lilimem(&sentinel);
	  return 1;
	}
      }
      strcpy(prev_db, the_db); /* save current db for next cycle */
    }

    LOG_PRINT(LOG_DEBUG, sql_command);

    rendinfo.dbires = dbi_conn_query(conn, sql_command);
    if (!rendinfo.dbires) {
      tiwrite(ptr_clrequest->fd, "select failed\n", TERM_NO);
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      return 1;
    }

    /* create a "header" if necessary */
    /* prepare_render resets sql_command */

    /* fill in variable elements of structure */
    rendinfo.ptr_ref = &sql_command;
    rendinfo.ptr_ref_len = &nsql_command_len;

    prepare_render_bibtex(&rendinfo);

    /* fetch all articles we're interested in */
    while (dbi_result_next_row(rendinfo.dbires)) {
      /* fill in variable elements of structure */
      rendinfo.ptr_ref = &sql_command;
      rendinfo.ptr_ref_len = &nsql_command_len;
      rendinfo.nref_counter = -1;
      rendinfo.database = the_db;
      rendinfo.dbname = dbstring;

      /* recycle sql_command */
      if ((new_sql_command = render_bibtex(&rendinfo)) == NULL) {
	iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
	dbi_result_free(rendinfo.dbires);
	dbi_conn_close(conn);
	delete_all_lilimem(&sentinel);
	ptr_addresult->failure++;
	return 1;
      }
      else {
	sql_command = new_sql_command;
      }
      ptr_addresult->success++;
      iwrite(ptr_clrequest->fd, sql_command, strlen(sql_command));
      sql_command[0] = '\0';
    }

    /* create a "footer" if necessary */
    /* fill in variable elements of structure */
    rendinfo.ptr_ref = &sql_command;
    rendinfo.ptr_ref_len = &nsql_command_len;

    if ((new_sql_command = finish_render_bibtex(&rendinfo)) == NULL) {
      dbi_result_free(rendinfo.dbires);
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }
    iwrite(ptr_clrequest->fd, sql_command, strlen(sql_command));
/*     dbi_result_free(dbires); */
  }

  dbi_conn_close(conn);

  tiwrite(ptr_clrequest->fd, "", TERM_YES); /* finish sending data */
  
  /* wait for receipt from client, reuse sql_command */
  result = tread(ptr_clrequest->fd, sql_command, nsql_command_len);
  
  if (result == -1) {
    LOG_PRINT(LOG_WARNING, "could not read ack from client");
  }
  
  if (strncmp(sql_command, positive.text, positive.length) != 0) {
    LOG_PRINT(LOG_WARNING, "could not read ack from client");
  }

  delete_all_lilimem(&sentinel);
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  bibtexify_author(): output authorname in BibTeX-compatible form as
                      F. M. Last lineage

  static char* bibtexify_author returns a ptr to the converted string
                      or NULL if an error occurs

  char** ptr_bibauthor ptr to a ptr to a string buffer. This fn will
                      allocate the memory needed for the converted
		      string and assign it to *ptr_bibauthor. The
		      calling function is responsible to free this
		      memory unless the return value is NULL.

  char* author ptr to the authorname to convert

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* bibtexify_author(char** ptr_bibauthor, char* author) {
  int nauthor_len;
  struct AUTHORTOKENS atoken;
  Lilistring* ptr_middle;

  if (author == NULL) {
    return NULL;
  }

  nauthor_len = strlen(author);

  if (tokenize_author(author, &atoken) == NULL) {
    return NULL;
  }

  *ptr_bibauthor = (char*)malloc(nauthor_len+11);
  if (*ptr_bibauthor == NULL) {
    return NULL;
  }

  (*ptr_bibauthor)[0] = '\0'; /* terminate string */

  strcat(*ptr_bibauthor, atoken.sur);

  if (*atoken.lineage) {
    strcat(*ptr_bibauthor, ", ");
    strcat(*ptr_bibauthor, atoken.lineage);
  }

  if (*atoken.first) {
    strcat(*ptr_bibauthor, ", ");
    strcat(*ptr_bibauthor, atoken.first);
    if (*(atoken.first+1) == '\0') {
      strcat(*ptr_bibauthor, ".");
    }
  }

  if (atoken.ptr_middlelist) {
    ptr_middle = atoken.ptr_middlelist;
    while ((ptr_middle = get_next_lilistring(ptr_middle)) != NULL) {
      strcat(*ptr_bibauthor, " ");
      strcat(*ptr_bibauthor, ptr_middle->token);
      if (*(ptr_middle->token+1) == '\0') {
	strcat(*ptr_bibauthor, ".");
      }
    }
  }

  clean_authortokens(&atoken);
  return *ptr_bibauthor;
}




