/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend.c defines the functions to access the database search output
  markus@mhoenicka.de 6-19-00
  $Id: backend.c,v 1.27 2004/05/01 23:40:13 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <stdlib.h>
#include <string.h>
#include <dbi/dbi.h>
#include <syslog.h>

#include "refdb.h"
#include "linklist.h"
#include "refdbd.h"
#include "backend.h"
#include "strfncs.h"
#include "dbfncs.h"
#include "risdb.h"
#include "authorinfo.h"

/* globals */

extern int n_log_level;
extern dbi_result dbi_style_res;
extern struct BSTRING outomem;

/* this string array is used for author queries. MySQL uses an enum field,
   pgsql uses an int field to store the author type */
const char author_type_string[2][3][12] = {
  {"\'primary\'", "\'secondary\'", "\'tertiary\'"},
  {"1", "2", "3"}
};

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  assemble_header(): assembles various html, xhtml, and XML headers

  struct renderinfo* ptr_rendinfo ptr to a structure containing 
                     information how the stuff is to be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* assemble_header(struct renderinfo* ptr_rendinfo) {
  /* see whether the client specified a character encoding */
  char enc_string[PREFS_BUF_LEN+80];
  char* header;
  size_t header_len = 1024;

  if ((header = (char*)malloc(header_len)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }

  *header = '\0';

  switch (ptr_rendinfo->ref_format) {
  case REFHTML:
    /* fall through */
  case REFXHTML:
    /* html/xhtml output */
    if (*(ptr_rendinfo->ptr_biblio_info->encoding)) {
      if (ptr_rendinfo->ref_format == REFHTML) {
	snprintf(enc_string, PREFS_BUF_LEN+80, "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=%s\">\n", ptr_rendinfo->ptr_biblio_info->encoding);
      }
      else { /* XHTML, need to close meta tag for XML */
	snprintf(enc_string, PREFS_BUF_LEN+80, "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=%s\"/>\n", ptr_rendinfo->ptr_biblio_info->encoding);
      }
    }
    else {
      /* should never happen */
      *enc_string = '\0';
    }

    if (*(ptr_rendinfo->cgi_url)) {
      if (ptr_rendinfo->ref_format == REFHTML) {
	snprintf(header, header_len, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n<html>\n<head>\n%s<title>%s reference list</title>\n<meta name=\"generator\" content=\"%s %s\">\n<link rel=\"stylesheet\" type=\"text/css\" href=\"%s\">\n</head>\n<body>\n<h1 class='h1'>refdb reference list</h1>\n", enc_string, PACKAGE, PACKAGE, VERSION, ptr_rendinfo->cgi_url);
      }
      else { /* REFXHTML */
	snprintf(header, header_len, "<?xml version=\"1.0\" encoding=\"%s\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\"><head>%s<title>%s reference list</title><meta name=\"generator\" content=\"%s %s\" id=\"generator\" /><link rel=\"stylesheet\" type=\"text/css\" href=\"%s\" />\n</head>\n<body>\n<h1 class='h1'>refdb reference list</h1>\n", ptr_rendinfo->ptr_biblio_info->encoding, enc_string, PACKAGE, PACKAGE, VERSION, ptr_rendinfo->cgi_url);
      }
    }
    else { /* don't use css */
      if (ptr_rendinfo->ref_format == REFHTML) {
	snprintf(header, header_len, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n<html>\n<head>\n%s<title>%s reference list</title>\n<meta name=\"generator\" content=\"%s %s\">\n</head>\n<body>\n<h1 class='h1'>refdb reference list</h1>\n", enc_string, PACKAGE, PACKAGE, VERSION);
      }
      else { /* REFXHTML */
	snprintf(header, header_len, "<?xml version=\"1.0\" encoding=\"%s\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\"><head>%s<title>%s reference list</title><meta name=\"generator\" content=\"%s %s\" id=\"generator\" /></head>\n<body>\n<h1 class='h1'>refdb reference list</h1>\n", ptr_rendinfo->ptr_biblio_info->encoding, enc_string, PACKAGE, PACKAGE, VERSION);
      }
    }
    break;
  case RISX:
    /* risx output */
    snprintf(header, header_len, "<?xml version=\"1.0\" encoding=\"%s\"?>\n%s", ptr_rendinfo->ptr_biblio_info->encoding, RISX_PUBID);
    break;
  case XNOTE:
    /* xnote output */
    snprintf(header, header_len, "<?xml version=\"1.0\" encoding=\"%s\"?>\n%s", ptr_rendinfo->ptr_biblio_info->encoding, XNOTE_PUBID);
    break;
/*   default: */
    /* empty string */
  }


  return header;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_id(): retrieves the note_id element from a note query

  char* get_refdb_note_id returns a pointer to a string containing the 
                     note_id element

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* id ptr to string that will receive the id. Must be long
                    enough for the representation of an unsigned
		    long long number

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_note_id(dbi_result dbires, char* id) {
  unsigned long long n_id;
  dbi_error_flag errflag;

  n_id = my_dbi_result_get_idval(dbires, "note_id");
  errflag = dbi_conn_error_flag(dbi_result_get_conn(dbires));

  if (n_id == 0 || errflag) { /* all IDs > 0 so this must be an error */
    return NULL;
  }
  else {
    sprintf(id, ULLSPEC, (unsigned long long)n_id);
    return id;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_key(): retrieves the note_key element from a note query

  char* get_refdb_note_key returns a pointer to a string containing the 
                     note_key element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_note_key(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "note_key");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_user_name(): retrieves the user_name element from a note query

  char* get_refdb_note_user_name returns a pointer to a string containing the 
                     user_name element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_note_user_name(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "user_name");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_title_copy(): retrieves the title element from a note query

  char* get_refdb_note_title_copy returns a pointer to a string containing the 
                     note_title element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_note_title_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "note_title");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_content_copy(): retrieves the content element from a note query

  char* get_refdb_note_content_copy returns a pointer to a string containing the 
                     note_content element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_note_content_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "note_content");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_content_type(): retrieves the content_type element
                                 from a note query

  char* get_refdb_content_type returns a pointer to a string containing the 
                     content_type element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_note_content_type(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "note_content_type");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_content_xmllang(): retrieves the content_xmllang element
                                 from a note query

  char* get_refdb_content_xmllang returns a pointer to a string containing the 
                     content_xmllang element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_note_content_xmllang(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "note_content_xmllang");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_note_date(): retrieves the note_date element from a note query

  char* get_refdb_note_date returns a pointer to a string containing the 
                     note_date element

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* date string that receives the date. Must be >= 256 byte

  int mode 0 = string 1 = notex representation
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_note_date(dbi_result dbires, char* date, int mode) {
  time_t the_time;

  the_time = dbi_result_get_datetime(dbires, "note_date");

  if (dbi_conn_error_flag(dbi_result_get_conn(dbires)) == 0) {
    if (the_time) {
      if (!mode) {
	strftime(date, 256, "%a %b %d %Y", gmtime(&the_time));
      }
      else {
	strftime(date, 256, "%Y-%m-%d", gmtime(&the_time));
      }
    }
    else {
      strcpy(date, "nd");
    }
    return date;
  }
  else {
    return NULL; /* have no date */
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  request_notes_by_ref(): requests all notes that are associated to
                          a given reference

  dbi_result request_notes_by_ref returns a pointer to a dbi result structure
                     which contains the notes

  dbi_conn conn the connection

  int mode 1 = reference entry 2 = keyword entry 4 = author entry
           8 = periodical entry

  const char* user name of current user

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result request_notes_by_ref(dbi_conn conn, unsigned long long n_id, int mode, const char* user) {
  int nis_first = 1;
  char* sql_command;
  char* sql_chunk;
  const char* drivername;
  dbi_result dbires;
  dbi_driver driver;

  driver = dbi_conn_get_driver(conn);
  drivername = dbi_driver_get_name(driver);

  sql_command = (char*)malloc(2048); 

  if (sql_command == NULL) {
    return NULL;
  }

  sql_chunk = (char*)malloc(2048); 

  if (sql_chunk == NULL) {
    free(sql_command);
    return NULL;
  }

  *sql_command = '\0';

  if (!strcmp(my_dbi_driver_get_cap(driver, "union"), "t")) {
    if (mode & REFERENCE) {
      sprintf(sql_chunk,
	      "SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xnote.xref_id=t_refdb.refdb_id AND t_xnote.xnote_type=\'REFERENCE\'",
	      user,
	      (unsigned long long)n_id);
      
      strcat(sql_command, sql_chunk);
      nis_first = 0;
    }
    
    if (mode & KEYWORD) {
      sprintf(sql_chunk,
	      "SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb, t_xkeyword, t_keyword WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xkeyword.keyword_id=t_keyword.keyword_id AND t_xkeyword.xref_id=t_refdb.refdb_id AND t_xnote.xref_id=t_keyword.keyword_id AND t_xnote.xnote_type=\'KEYWORD\'",
	      user,
	      (unsigned long long)n_id);

      if (!nis_first) {
	strcat(sql_command, " UNION ");
      }
      strcat(sql_command, sql_chunk);
      nis_first = 0;
    }      

    if (mode & AUTHOR) {
      sprintf(sql_chunk,
	      "SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb, t_xauthor, t_author WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xauthor.author_id=t_author.author_id AND t_xauthor.refdb_id=t_refdb.refdb_id AND t_xnote.xref_id=t_author.author_id AND t_xnote.xnote_type=\'AUTHOR\'",
	      user,
	      (unsigned long long)n_id);
      
      if (!nis_first) {
	strcat(sql_command, " UNION ");
      }
      strcat(sql_command, sql_chunk);
      nis_first = 0;
    }      

    if (mode & PERIODICAL) {
      sprintf(sql_chunk,
	      "SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb, t_periodical WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_refdb.refdb_periodical_id=t_periodical.periodical_id AND t_xnote.xref_id=t_periodical.periodical_id AND t_xnote.xnote_type=\'PERIODICAL\'",
	      user,
	      (unsigned long long)n_id);
      
      if (!nis_first) {
	strcat(sql_command, " UNION ");
      }
      strcat(sql_command, sql_chunk);
    }
  }
  else {
    /* if the db engine does not support UNION, emulate UNION with
       SELECT INTO temptable */
    if (!strcmp(drivername, "mysql")) {
      /* mysql */
      strcpy(sql_command, "CREATE TEMPORARY TABLE t_noteunion \
                                    (note_id INT, \
                                    note_key VARCHAR(255), \
                                    note_title VARCHAR(255), \
                                    note_content BLOB, \
                                    note_content_type VARCHAR(255), \
                                    note_content_xmllang VARCHAR(255), \
                                    note_user_id INT, \
                                    note_date DATETIME)");
    }
    else if (!strcmp(drivername, "pgsql")) {
      /* pgsql */
      strcpy(sql_command, "CREATE TEMPORARY TABLE t_noteunion \
                                   (note_id BIGSERIAL, \
                                    note_key VARCHAR(255) UNIQUE, \
                                    note_title VARCHAR(255), \
                                    note_content TEXT, \
                                    note_content_type VARCHAR(255), \
                                    note_content_xmllang VARCHAR(255), \
                                    note_user_id BIGINT, \
                                    note_date DATE)");
    }
    else if (!strcmp(drivername, "sqlite")) {
      /* sqlite */
      strcpy(sql_command, "CREATE TEMPORARY TABLE t_noteunion \
                                   (note_id INTEGER, \
                                    note_key TEXT, \
                                    note_title TEXT, \
                                    note_content_type TEXT, \
                                    note_content_xmllang TEXT, \
                                    note_user_id INTEGER, \
                                    note_date DATE, \
                                    note_content TEXT)");
    }
    else {
      /* should never happen */
      LOG_PRINT(LOG_WARNING, "unsupported driver");
      free(sql_command);
      free(sql_chunk);
      return NULL;
    }

    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires = dbi_conn_query(conn, sql_command);

    if (!dbires) {
      LOG_PRINT(LOG_WARNING, "create temporary table failed");
      free(sql_command);
      free(sql_chunk);
      return NULL;
    }


    /* now collect the data in the temporary table */
    if (mode & REFERENCE) {
      sprintf(sql_chunk,
	      "INSERT INTO t_noteunion (note_id, note_key, note_title, note_user_id, note_date, note_content, note_content_type, note_content_xmllang) SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xnote.xref_id=t_refdb.refdb_id AND t_xnote.xnote_type=\'REFERENCE\'",
	      user,
	      (unsigned long long)n_id);
      
      LOG_PRINT(LOG_DEBUG, sql_chunk);
      dbires = dbi_conn_query(conn, sql_chunk);

      if (!dbires) {
	LOG_PRINT(LOG_WARNING, "insert into temporary table failed");
/* 	free(sql_command); */
/* 	free(sql_chunk); */
/* 	return NULL; */
      }
    }
    
    if (mode & KEYWORD) {
      sprintf(sql_chunk,
	      "INSERT INTO t_noteunion (note_id, note_key, note_title, note_user_id, note_date, note_content, note_content_type, note_content_xmllang) SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content FROM t_note,t_xnote, t_user, t_refdb, t_xkeyword, t_keyword WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xkeyword.keyword_id=t_keyword.keyword_id AND t_xkeyword.xref_id=t_refdb.refdb_id AND t_xnote.xref_id=t_keyword.keyword_id AND t_xnote.xnote_type=\'KEYWORD\'",
	      user,
	      (unsigned long long)n_id);

      LOG_PRINT(LOG_DEBUG, sql_chunk);
      dbires = dbi_conn_query(conn, sql_chunk);

      if (!dbires) {
	LOG_PRINT(LOG_WARNING, "insert into temporary table failed");
/* 	free(sql_command); */
/* 	free(sql_chunk); */
/* 	return NULL; */
      }
    }      

    if (mode & AUTHOR) {
      sprintf(sql_chunk,
	      "INSERT INTO t_noteunion (note_id, note_key, note_title, note_user_id, note_date, note_content, note_content_type, note_content_xmllang) SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb, t_xauthor, t_author WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_xauthor.author_id=t_author.author_id AND t_xauthor.refdb_id=t_refdb.refdb_id AND t_xnote.xref_id=t_author.author_id AND t_xnote.xnote_type=\'AUTHOR\'",
	      user,
	      (unsigned long long)n_id);
      
      LOG_PRINT(LOG_DEBUG, sql_chunk);
      dbires = dbi_conn_query(conn, sql_chunk);

      if (!dbires) {
	LOG_PRINT(LOG_WARNING, "insert into temporary table failed");
/* 	free(sql_command); */
/* 	free(sql_chunk); */
/* 	return NULL; */
      }
    }      

    if (mode & PERIODICAL) {
      sprintf(sql_chunk,
	      "INSERT INTO t_noteunion (note_id, note_key, note_title, note_user_id, note_date, note_content, note_content_type, note_content_xmllang) SELECT DISTINCT t_note.note_id, t_note.note_key, t_note.note_title, t_note.note_user_id, t_note.note_date, t_note.note_content, t_note.note_content_type, t_note.note_content_xmllang FROM t_note,t_xnote, t_user, t_refdb, t_periodical WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_refdb.refdb_id="ULLSPEC" AND t_note.note_id=t_xnote.note_id AND t_refdb.refdb_periodical_id=t_periodical.periodical_id AND t_xnote.xref_id=t_periodical.periodical_id AND t_xnote.xnote_type=\'PERIODICAL\'",
	      user,
	      (unsigned long long)n_id);
      
      LOG_PRINT(LOG_DEBUG, sql_chunk);
      dbires = dbi_conn_query(conn, sql_chunk);

      if (!dbires) {
	LOG_PRINT(LOG_WARNING, "insert into temporary table failed");
/* 	free(sql_command); */
/* 	free(sql_chunk); */
/* 	return NULL; */
      }
    }

    /* now retrieve the datasets again */
    sprintf(sql_command, "SELECT DISTINCT note_id, note_key, note_title, note_user_id, note_date, note_content, note_content_type, note_content_xmllang FROM t_noteunion");

  }

  free(sql_chunk);
	 
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);

 /*  if (!dbires) { */
    /* begin weird hack */
/*     sprintf(sql_command, "SELECT refdb_id FROM t_refdb where refdb_id=-1"); */
/*     LOG_PRINT(LOG_DEBUG, sql_command); */
/*     dbires = dbi_conn_query(conn, sql_command); */
    /* end weird hack */
/*     return NULL; */
/*   } */

  free(sql_command);

  return dbires;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  free_request_notes_by_ref(): cleans up the resources used by a previous
                               call to request_notes_by_ref

  int free_request_notes_by_ref returns 0 on success and 1 if an error
                               occurs

  dbi_conn conn the connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int free_request_notes_by_ref(dbi_conn conn) {
  dbi_result dbires;
  char sql_command[23];

  /* check whether we used a temporary table. If yes, drop it */
  if (strcmp(my_dbi_driver_get_cap(dbi_conn_get_driver(conn), "union"), "t")) {
    /* get rid of the temporary table */
    strcpy(sql_command, "DROP TABLE t_noteunion");
    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires = dbi_conn_query(conn, sql_command);
    
    if (!dbires) {
      LOG_PRINT(LOG_WARNING, "drop temporary table failed");
      return 1;
    }
  }

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  request_links(): prepares the retrieval of links

  dbi_result request_links returns a pointer to a dbi result structure
                     which contains the links

  dbi_conn conn the connection

  int mode 0 = reference entry 1 = keyword entry 2 = author entry
           3 = periodical entry

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result request_links(dbi_conn conn, unsigned long long n_id, int mode) {
  char* sql_command;
  char linktype[4][11] = {"REFERENCE", "KEYWORD", "AUTHOR", "PERIODICAL"};
  int error;
  dbi_result dbires;

  sql_command = (char*)malloc(512); 

  if (sql_command == NULL) {
    return NULL;
  }

  if (!mode) {
    /* link to refererence. Need citation key */
    sprintf(sql_command, "SELECT t_refdb.refdb_citekey FROM t_refdb, t_xnote , t_note WHERE t_xnote.xnote_type='\%s\' AND t_refdb.refdb_id=t_xnote.xref_id AND t_xnote.note_id= t_note.note_id AND t_note.note_id="ULLSPEC, linktype[mode], (unsigned long long)n_id);
  }
  else if (mode == 1) {
    /* link to keyword. Need keyword name */
    sprintf(sql_command, "SELECT t_keyword.keyword_name FROM t_keyword, t_xnote , t_note WHERE t_xnote.xnote_type='\%s\' AND t_keyword.keyword_id=t_xnote.xref_id AND t_xnote.note_id= t_note.note_id AND t_note.note_id="ULLSPEC, linktype[mode], (unsigned long long)n_id);
  }
  else if (mode == 2) {
    /* link to author. Need author name */
    sprintf(sql_command, "SELECT t_author.author_name FROM t_author, t_xnote , t_note WHERE t_xnote.xnote_type='\%s\' AND t_author.author_id=t_xnote.xref_id AND t_xnote.note_id= t_note.note_id AND t_note.note_id="ULLSPEC, linktype[mode], (unsigned long long)n_id);
  }
  else /* if (mode == 3) */ {
    /* link to periodical. Need one of the periodical names */
    sprintf(sql_command, "SELECT t_periodical.periodical_name,t_periodical.periodical_abbrev,t_periodical.periodical_custabbrev1,t_periodical.periodical_custabbrev2 FROM t_periodical, t_xnote , t_note WHERE t_xnote.xnote_type='\%s\' AND t_periodical.periodical_id=t_xnote.xref_id AND t_xnote.note_id= t_note.note_id AND t_note.note_id="ULLSPEC, linktype[mode], (unsigned long long)n_id);
  }

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);
  free(sql_command);
  if (!dbires) {
    return NULL;
  }

  return dbires;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_link(): retrieves a link

  char* get_link returns a pointer to a string containing the 
                     link

  dbi_result dbires database query result, pointing to the current
                    dataset

  int *ptr_mode pointer to an integer to request the link type. This
                    argument is currently ignored for all but periodical
		    links. Initialize *ptr_mode to one of: 0 = reference entry
		    1 = keyword entry 2 = author entry 3 = periodical entry
		    For successfully retrieved periodical entries,
		    *ptr_mode will be updated to reflect the type of the
		    periodical name used: 0 = full name, 1 = abbreviated
		    name, 2 = custom abbreviation 1, 3 = custom abbreviation 2

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_link(dbi_result dbires, int *ptr_mode) {
  const char* result = NULL;
  int i = 1; /* index into a dbi result, starts at 1 */

  if (dbi_result_next_row(dbires)) {
    if (*ptr_mode == 3) {
      /* periodical links require special treatment as we don't know
	 in advance which periodical names are available */
      while (i < 5 && (!result || !*result)) {
	result = dbi_result_get_string_idx(dbires, i);
	if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
	  return NULL;
	}
	else if (result && *result) {
	  *ptr_mode = i-1;
	  return result;
	}
	i++;
      }
      return NULL;
    }
    else {
      result = dbi_result_get_string_idx(dbires, 1);
      if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
	return NULL;
      }
      else {
	return result;
      }
    }
  }
  else {
    return NULL;
  }
}

  /* 0  refdb_id             ID  - (if numeric)
     1  refdb_type           TY  - 
     2  refdb_pubyear        PY  - (partial)
     3  refdb_startpage      SP  - 
     4  refdb_endpage        EP  - 
     5  refdb_abstract       N2  - 
     6  refdb_title          TI  - 
     7  refdb_volume         VL  - 
     8  refdb_issue          CP  - 
     9  refdb_booktitle      BT  - 
    10  refdb_city           CY  - 
    11  refdb_publisher      PB  - 
    12  refdb_title_series   T3  - 
    13  refdb_address        AD  - 
    14  refdb_url            UR  - 
    15  refdb_issn           SN  - 
    16  refdb_periodical_id  JO  - (indirect)
    17  refdb_pyother_info   PY  - (partial)
    18  refdb_secyear        Y2  - (partial)
    19  refdb_secother_info  Y2  - (partial)
    20  refdb_user1          U1  - 
    21  refdb_user2          U2  - 
    22  refdb_user3          U3  - 
    23  refdb_user4          U4  - 
    24  refdb_user5          U5  - 
    25  refdb_misc1          M1  - 
    26  refdb_misc2          M2  - 
    27  refdb_misc3          M3  - 
    28  refdb_citekey        ID  - (if non-numeric)
*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_id(): retrieves the refdb_id element from a query result

  char* get_refdb_id returns a pointer to a string containing the 
                     refdb_id element

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* id ptr to string that will receive the id. Must be long
                    enough for the representation of an unsigned
		    long long number

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_id(dbi_result dbires, char* id) {
  unsigned long long n_id;
  dbi_error_flag errflag;

  n_id = my_dbi_result_get_idval(dbires, "refdb_id");
  errflag = dbi_conn_error_flag(dbi_result_get_conn(dbires));

  if (n_id == 0 || errflag) { /* all IDs > 0 so this must be an error */
    return NULL;
  }
  else {
    sprintf(id, ULLSPEC, (unsigned long long)n_id);
    return id;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_citekey(): retrieves the refdb_citekey element from a query result

  char* get_refdb_citekey returns a pointer to a string containing the 
                     refdb_citekey element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_citekey(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_citekey");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_type(): retrieves the refdb_type element from a query result

  char* get_refdb_type returns a pointer to a string containing the 
                     refdb_type element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_type(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_type");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_pubyear(): retrieves the refdb_pubyear element from a query result

  char* get_refdb_pubyear returns a pointer to a string containing the 
                     refdb_pubyear element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_pubyear(dbi_result dbires, char* year) {
  int n_year;

  n_year = dbi_result_get_short(dbires, "refdb_pubyear");

  if (dbi_conn_error_flag(dbi_result_get_conn(dbires)) == 0) {
    if (n_year != 0) {
      sprintf(year, "%hd", n_year);
    }
    else {
      strcpy(year, "nd");
    }
    return year;
  }
  else {
    return NULL; /* have no year */
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_startpage(): retrieves the refdb_startpage element from a query result

  char* get_refdb_startpage returns a pointer to a string containing the 
                     refdb_startpage element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_startpage(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_startpage");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_endpage(): retrieves the refdb_endpage element from a query result

  char* get_refdb_endpage returns a pointer to a string containing the 
                     refdb_endpage element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_endpage(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_endpage");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_abstract_copy(): retrieves the refdb_abstract element from
                             a query result

  char* get_refdb_abstract_copy returns a pointer to a string containing the 
                     refdb_abstract element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_abstract_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_abstract");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_title_copy(): retrieves the refdb_title element from a query result

  char* get_refdb_title_copy returns a pointer to a string containing the 
                     refdb_title element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_title_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_title");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_volume(): retrieves the refdb_volume element from a query result

  char* get_refdb_volume returns a pointer to a string containing the 
                     refdb_volume element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_volume(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_volume");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_issue(): retrieves the refdb_issue element from a query result

  char* get_refdb_issue  returns a pointer to a string containing the 
                     refdb_issue element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_issue(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_issue");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_booktitle_copy(): retrieves the refdb_booktitle element from a query result

  char* get_refdb_booktitle_copy returns a pointer to a string containing the 
                     refdb_booktitle element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_booktitle_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_booktitle");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_city(): retrieves the refdb_city element from a query result

  char* get_refdb_city returns a pointer to a string containing the 
                     refdb_city element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_city(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_city");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_publisher(): retrieves the refdb_publisher element from a query result

  char* get_refdb_publisher returns a pointer to a string containing the 
                     refdb_publisher element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_publisher(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_publisher");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_title_series_copy(): retrieves the refdb_title_series
                                 element from a query result

  char* get_refdb_title_series_copy returns a pointer to a string
                     containing the refdb_title_series element. 
                     The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_title_series_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_title_series");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_address_copy(): retrieves the refdb_address element from a
                            query result

  char* get_refdb_address_copy returns a pointer to a string containing the 
                     refdb_address element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_address_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_address");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_url_copy(): retrieves the refdb_url element from a query result

  char* get_refdb_url_copy returns a pointer to a string containing the 
                     refdb_url element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_url_copy(dbi_result dbires) {
  return my_dbi_result_get_string_copy(dbires, "refdb_url");
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_issn(): retrieves the refdb_issn element from a query result

  char* get_refdb_issn returns a pointer to a string containing the 
                     refdb_issn element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_issn(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_issn");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_pyother_info(): retrieves the refdb_pyother_info element from a query result

  char* get_refdb_pyother_info returns a pointer to a string containing the 
                     refdb_pyother_info element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_pyother_info(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_pyother_info");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_secyear(): retrieves the refdb_secyear element from a query result

  char* get_refdb_secyear returns a pointer to a string containing the 
                     refdb_secyear element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_refdb_secyear(dbi_result dbires, char* year) {
  int n_year;

  n_year = dbi_result_get_short(dbires, "refdb_secyear");

  if (n_year != 0 && dbi_conn_error_flag(dbi_result_get_conn(dbires)) == 0) {
    sprintf(year, "%hd", n_year);
    return year;
  }
  else {
    return NULL; /* have no year */
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_secother_info(): retrieves the refdb_secother_info element from a query result

  char* get_refdb_secother_info returns a pointer to a string containing the 
                     refdb_secother_info element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_secother_info(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_secother_info");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_user1(): retrieves the refdb_user1 element from a query result

  char* get_refdb_user1 returns a pointer to a string containing the 
                     refdb_user1 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_user1(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_user1");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_user2(): retrieves the refdb_user2 element from a query result

  char* get_refdb_user1 returns a pointer to a string containing the 
                     refdb_user2 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_user2(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_user2");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_user3(): retrieves the refdb_user3 element from a query result

  char* get_refdb_user3 returns a pointer to a string containing the 
                     refdb_user3 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_user3(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_user3");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_user4(): retrieves the refdb_user4 element from a query result

  char* get_refdb_user4 returns a pointer to a string containing the 
                     refdb_user1 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_user4(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_user4");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_user5(): retrieves the refdb_user5 element from a query result

  char* get_refdb_user5 returns a pointer to a string containing the 
                     refdb_user5 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_user5(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_user5");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_misc1(): retrieves the refdb_misc1 element from a query result

  char* get_refdb_misc1 returns a pointer to a string containing the 
                     refdb_misc1 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_misc1(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_misc1");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_misc2(): retrieves the refdb_misc2 element from a query result

  char* get_refdb_misc2 returns a pointer to a string containing the 
                     refdb_misc2 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_misc2(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_misc2");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_refdb_misc3(): retrieves the refdb_misc3 element from a query result

  char* get_refdb_misc3 returns a pointer to a string containing the 
                     refdb_misc3 element

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_refdb_misc3(dbi_result dbires) {
  const char* result;

  result = dbi_result_get_string(dbires, "refdb_misc3");
  if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    return NULL;
  }
  else {
    return result;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_notes_copy(): retrieves the xuser_notes element from a query result

  char* get_notes_copy returns a pointer to a string containing the 
                     xuser_notes element. The string is allocated and
		     must be freed by the caller

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* username ptr to string with the current username, or NULL if
                    the search should not be restricted to the 
                    current user

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_notes_copy(dbi_result dbires, char* username) {
  char* sql_command;
  const char* notes = NULL;
  int error;
  unsigned long long n_id;
  unsigned int notes_length;
  dbi_result dbires1;

  n_id = my_dbi_result_get_idval(dbires, "refdb_id");
  if (n_id == 0) { /* all IDs >=1 so this must be an error */
    return NULL;
  }

  sql_command = (char*)malloc(256); 

  if (sql_command == NULL) {
    return NULL;
  }

  if (username != NULL) {
    sprintf(sql_command, "SELECT t_xuser.xuser_notes FROM t_xuser, t_user WHERE t_xuser.user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_xuser.refdb_id="ULLSPEC, username, (unsigned long long)n_id);
  }
  else { /* use the first entry regardless of which user */
    sprintf(sql_command, "SELECT t_xuser.xuser_notes FROM t_xuser, t_user WHERE t_xuser.user_id=t_user.user_id AND t_xuser.refdb_id="ULLSPEC, (unsigned long long)n_id);
  }

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires1 = dbi_conn_query(dbi_result_get_conn(dbires), sql_command);
  free(sql_command);
  
  if (!dbires1) {
    return NULL;
  }

  if (dbi_result_next_row(dbires1)) {
    /* the result string will be an allocated copy that we have to free */
    notes = my_dbi_result_get_string_copy(dbires1, "xuser_notes");
  }
  dbi_result_free(dbires1);

  /* notes will still be NULL if result contains zero rows */
  return (char*)notes;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_reprint(): retrieves the xuser_reprint element and related
                 elements from a query result

  struct REPRINT get_reprint returns a structure which points to
                     strings containing the xuser_reprint and
                     xuser_date elements. If the previous search
                     with request_reprint() was limited to the current
                     user, this function will return the reprint status
                     of this user. Otherwise it will search for the
                     first user that has the reference in file. If no
                     one has it in file, it will search for the first
                     user that has it on request. If there is still
                     no match, it will return any "NOT IN FILE" field.
		     In case of an error, one or both REPRINT structure
                     members are NULL

  dbi_result dbires database query result, pointing to the current
                    dataset

  struct REPRINT* ptr_reprint ptr to structure which will receive the result

  char* username ptr to string with username or NULL if search should not
                 be limited to a particular user

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
struct REPRINT* get_reprint(dbi_result dbires, struct REPRINT* ptr_reprint, char* username, int yeardigits) {
  char* sql_command;
  const char* xuser_reprint;
  const char* xuser_avail;
  int error;
  int num_rows;
  unsigned long long n_id;
  time_t reprint_date;
  dbi_conn conn;
  dbi_result dbires1;

  sql_command = (char*)malloc(256); 

  if (sql_command == NULL) {
    return NULL;
  }

  n_id = my_dbi_result_get_idval(dbires, "refdb_id");

  if (n_id == 0) { /* all IDs > 0, so this must be an error */
    free(sql_command);
    return NULL;
  }

  conn = dbi_result_get_conn(dbires);

  if (username != NULL) {
    sprintf(sql_command, "SELECT t_xuser.xuser_reprint, t_xuser.xuser_date, t_xuser.xuser_avail FROM t_xuser, t_user WHERE t_xuser.user_id=t_user.user_id AND t_user.user_name=\'%s\' AND t_xuser.refdb_id="ULLSPEC, username, (unsigned long long)n_id);
  }
  else {
    sprintf(sql_command, "SELECT t_xuser.xuser_reprint, t_xuser.xuser_date, t_xuser.xuser_avail FROM t_xuser WHERE t_xuser.refdb_id="ULLSPEC, (unsigned long long)n_id);
  }

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires1 = dbi_conn_query(conn, sql_command);
  free(sql_command);
  
  if (!dbires1) {
    return NULL;
  }

  /* check how many rows. If > 1, search for first entry IN FILE, or
     ON REQUEST if the former is not found. Assemble string from the
     user_reprint and user_date columns */

  /* set defaults */
  (ptr_reprint->reprint)[0] = '\0';
  (ptr_reprint->date)[0] = '\0';
  (ptr_reprint->avail)[0] = '\0';

  if ((num_rows = dbi_result_get_numrows(dbires1)) > 1) {
    /* find first entry IN FILE */
    while (dbi_result_next_row(dbires1)) {
      xuser_reprint = dbi_result_get_string_idx(dbires1, 1);
      if (!xuser_reprint || dbi_conn_error_flag(conn)) {
	continue;
      }
      if (strncmp(xuser_reprint, "IN FILE", 7) == 0) {
	strcpy(ptr_reprint->reprint, xuser_reprint);
	reprint_date = dbi_result_get_datetime_idx(dbires1, 2);
	if (reprint_date && !dbi_conn_error_flag(conn)) {
	  print_risdate(&reprint_date, ptr_reprint->date, yeardigits);
	}
	xuser_avail = dbi_result_get_string_idx(dbires1, 3);
	if (xuser_avail && !dbi_conn_error_flag(conn)) {
	  strcpy(ptr_reprint->avail, xuser_avail);
	}
	return ptr_reprint; /* the first IN FILE entry */
      }
    } /* end while */

    /* rewind... */
    dbi_result_first_row(dbires1);

    while (dbi_result_next_row(dbires1)) {
      xuser_reprint = dbi_result_get_string_idx(dbires1, 1);
      if (!xuser_reprint || dbi_conn_error_flag(conn)) {
	continue;
      }
      if (strncmp(xuser_reprint, "ON REQUEST", 10) == 0) {
	strcpy(ptr_reprint->reprint, xuser_reprint);
	reprint_date = dbi_result_get_datetime_idx(dbires1, 2);
	if (reprint_date && !dbi_conn_error_flag(conn)) {
	  print_risdate(&reprint_date, ptr_reprint->date, yeardigits);
	}
	xuser_avail = dbi_result_get_string_idx(dbires1, 3);
	if (xuser_avail && !dbi_conn_error_flag(conn)) {
	  strcpy(ptr_reprint->avail, xuser_avail);
	}
	return ptr_reprint; /* the first ON REQUEST entry */
      }
    } /* end while */

    /* rewind... */
    dbi_result_first_row(dbires1);
    
    if (dbi_result_next_row(dbires1)) {
      xuser_reprint = dbi_result_get_string_idx(dbires1, 1);
      if (xuser_reprint && !dbi_conn_error_flag(conn)) {
	strcpy(ptr_reprint->reprint, xuser_reprint);
      }
      reprint_date = dbi_result_get_datetime_idx(dbires1, 2);
      if (reprint_date && !dbi_conn_error_flag(conn)) {
	print_risdate(&reprint_date, ptr_reprint->date, yeardigits);
      }
      xuser_avail = dbi_result_get_string_idx(dbires1, 3);
      if (xuser_avail && !dbi_conn_error_flag(conn)) {
	strcpy(ptr_reprint->avail, xuser_avail);
      }
      return ptr_reprint; /* the first NOT IN FILE entry */
    }
    else {
      return ptr_reprint; /* have nothing */
    }
  }
  else if (num_rows == 1) {
    if (dbi_result_next_row(dbires1)) {
      xuser_reprint = dbi_result_get_string_idx(dbires1, 1);
      if (xuser_reprint && !dbi_conn_error_flag(conn)) {
	strcpy(ptr_reprint->reprint, xuser_reprint);
      }
      reprint_date = dbi_result_get_datetime_idx(dbires1, 2);
      if (reprint_date && !dbi_conn_error_flag(conn)) {
	print_risdate(&reprint_date, ptr_reprint->date, yeardigits);
      }
      xuser_avail = dbi_result_get_string_idx(dbires1, 3);
      if (xuser_avail && !dbi_conn_error_flag(conn)) {
	strcpy(ptr_reprint->avail, xuser_avail);
      }
      return ptr_reprint; /* the first NOT IN FILE entry */
    }
    else {
      return ptr_reprint; /* have nothing */
    }
  }
  else {
    return ptr_reprint; /* have nothing */
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_periodical(): prepares the retrieval of periodical from
                           a given database

  char* get_periodical returns the periodical string or NULL if an
                    error occurred

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* periodical ptr to string that will receive the result. Must
                    hold at least 256 characters.

  const char* db string holding the database to search in, or NULL if the
                  current database should be used

  int type indicates which periodical name is to be retrieved. 0=name, 
                     1=custabbrev1, 2=custabbrev2, 3=abbrev, 4=the first
                     existing name in the order name, abbrev, custabbrev1,
                     custabbrev2

  int* errcode ptr to an int that will receive the error code
               (0 = no error, 1 = nothing found, 2 = memory error,
	        3 = database server error)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* get_periodical(dbi_conn conn, char* periodical, const char* db, int type, int* errcode, unsigned long long n_id) {
  char *sql_command;
  char *dbstring = NULL;
  const char *item;
  const char *drivername;
  int error;
  int i;
  dbi_result dbires1;

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  sql_command = (char*)malloc(1024); 

  if (sql_command == NULL) {
    *errcode = 2;
    return NULL;
  }

  if (!n_id) {
    *errcode = 1;
    return NULL;
  }

  /* never use db argument with pgsql driver */
  if (db && *db && !strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
    dbstring = (char*)malloc(strlen(db)+2); 

    if (dbstring == NULL) {
      *errcode = 2;
      return NULL;
    }
    sprintf(dbstring, "%s.", db);
  }
  else {
    /* this is weird but simplifies the free() calls further down */
    dbstring = (char*)malloc(1); 

    if (dbstring == NULL) {
      *errcode = 2;
      return NULL;
    }
    dbstring[0] = '\0';
  }

  if (type == 4) {
    /* we try all periodical names in a fixed order and return the first
       one we can find, or NULL if we don't find anything */
    for (i = 0; i < 4; i++) {
      switch (i) {
      case 0:
	/* first try name */
	sprintf(sql_command, "SELECT %st_periodical.periodical_name FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
	break;
      case 1:
	/* then try abbrev */
	sprintf(sql_command, "SELECT %st_periodical.periodical_abbrev FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
	break;
      case 2:
	/* then try custabbrev1 */
	sprintf(sql_command, "SELECT %st_periodical.periodical_custabbrev1 FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
	break;
      case 3:
	/* finally try custabbrev2 */
	sprintf(sql_command, "SELECT %st_periodical.periodical_custabbrev2 FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
	break;
      default:
	LOG_PRINT(LOG_CRIT, "hope you never see this");
	return NULL;
      }

      LOG_PRINT(LOG_DEBUG, sql_command);
      dbires1 = dbi_conn_query(conn, sql_command);
      if (!dbires1) {
	*errcode = 3;
	free(sql_command);
	free(dbstring);
	return NULL;
      }

      item = real_get_periodical(dbires1);

      if (item && *item) {
	/* we're done, copy string into provided buffer and return it */
	*errcode = 0;
	strcpy(periodical, item);
	dbi_result_free(dbires1);
	free(sql_command);
	free(dbstring);
	return periodical;
      }
    } /* end for */


    if (i == 4) { /* nothing found */
      *errcode = 1;
      free(sql_command);
      free(dbstring);
      return NULL;
    }
  }
  else {
    if (!type) {
      sprintf(sql_command, "SELECT %st_periodical.periodical_name FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
    }
    else if (type == 1) {
      sprintf(sql_command, "SELECT %st_periodical.periodical_custabbrev1 FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
    }
    else if (type == 2) {
      sprintf(sql_command, "SELECT %st_periodical.periodical_custabbrev2 FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
    }
    else if (type == 3) {
      sprintf(sql_command, "SELECT %st_periodical.periodical_abbrev FROM %st_periodical, %st_refdb WHERE %st_refdb.refdb_periodical_id=%st_periodical.periodical_id AND %st_refdb.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);
    }

    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires1 = dbi_conn_query(conn, sql_command);
  
    free(sql_command);
    free(dbstring);

    if (!dbires1) {
      *errcode = 3;
      return NULL;
    }

    item = real_get_periodical(dbires1);
    if (!item) {
      *errcode = 1;
      dbi_result_free(dbires1);
      return NULL;
    }
    else {
      *errcode = 0;
      strcpy(periodical, item);
      dbi_result_free(dbires1);
      return periodical;
    }
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  real_get_periodical(): retrieves the periodical from a query result

  char* real_get_periodical returns a pointer to a string containing the 
                     periodical

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* real_get_periodical(dbi_result dbires) {
  const char* result;

  if (dbi_result_next_row(dbires)) {
    result = dbi_result_get_string_idx(dbires, 1);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  request_authors(): prepares the retrieval of authors from the
                     currently selected database

  dbi_result request_authors returns a query result
                     which contains the authors

  dbi_conn conn connection to the database server

  int type 1 = primary author, 2 = secondary author, 3 = tertiary author

  const char* db ptr to a string with the database name or NULL if the
                    current database should be used

  unsigned long long n_id id of the dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result request_authors(dbi_conn conn, int type, const char* db, unsigned long long n_id) {
  char *sql_command;
  char *dbstring;
  const char *drivername;
  char idstring[32];
  int error;
  dbi_result dbires;
  dbi_driver driver;

  driver = dbi_conn_get_driver(conn);
  drivername = dbi_driver_get_name(driver);

  sql_command = (char*)malloc(1024); 

  if (sql_command == NULL) {
    return NULL;
  }

  /* never use db argument with pgsql */
  if (db && *db && !strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
    dbstring = (char*)malloc(strlen(db)+2); 

    if (dbstring == NULL) {
      return NULL;
    }
    sprintf(dbstring, "%s.", db);
  }
  else {
    /* this is weird but simplifies the free() calls further down */
    dbstring = (char*)malloc(1); 

    if (dbstring == NULL) {
      return NULL;
    }
    dbstring[0] = '\0';
  }

  sprintf(sql_command, "SELECT %st_author.author_name,%st_author.author_lastname,%st_author.author_firstname,%st_author.author_middlename,%st_author.author_suffix FROM %st_author, %st_xauthor WHERE %st_author.author_id=%st_xauthor.author_id AND %st_xauthor.xauthor_type=%s AND %st_xauthor.refdb_id="ULLSPEC" ORDER BY t_xauthor.xauthor_position", dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, get_author_type_string(driver, type), dbstring, (unsigned long long)n_id);

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  free(sql_command);
  free(dbstring);

  return dbires;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_author(): retrieves an author

  char* get_author returns a pointer to a string containing the 
                     author

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_author(dbi_result dbires) {
  const char* result;

  if (dbi_result_next_row(dbires)) {
    result = dbi_result_get_string_idx(dbires, 1);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_author_parts(): retrieves an author's name parts

  char* get_author_parts returns a pointer to a struct containing the 
                     author name parts

  dbi_result dbires database query result, pointing to the current
                    dataset

  struct AUTHOR_INFO* ptr_ainfo ptr to a structure that will receive
                    the result

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
struct AUTHOR_INFO* get_author_parts(dbi_result dbires, struct AUTHOR_INFO* ptr_ainfo) {
  const char* result;

  /* start with a clean struct */
  ptr_ainfo->name[0] = '\0';
  ptr_ainfo->lastname[0] = '\0';
  ptr_ainfo->firstname[0] = '\0';
  ptr_ainfo->middlename[0] = '\0';
  ptr_ainfo->suffix[0] = '\0';

  if (dbi_result_next_row(dbires)) {
    /* full name */
    result = dbi_result_get_string_idx(dbires, 1);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else if (result) {
      strcpy(ptr_ainfo->name, result);
    }

    /* last name */
    result = dbi_result_get_string_idx(dbires, 2);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else if (result) {
      strcpy(ptr_ainfo->lastname, result);
    }

    /* first name */
    result = dbi_result_get_string_idx(dbires, 3);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else if (result) {
      strcpy(ptr_ainfo->firstname, result);
    }

    /* middle name */
    result = dbi_result_get_string_idx(dbires, 4);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else if (result) {
      strcpy(ptr_ainfo->middlename, result);
    }

    /* suffix */
    result = dbi_result_get_string_idx(dbires, 5);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else if (result) {
      strcpy(ptr_ainfo->suffix, result);
    }
  }
  else {
    return NULL;
  }

  return ptr_ainfo;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  request_keywords(): prepares the retrieval of keywords

  dbi_result request_keywords returns a pointer to a dbi result structure
                     which contains the keywords

  dbi_result dbires database query result, pointing to the current
                    dataset (used only to retrieve the connection)

  int mode 0 = reference entry 1 = note entry

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result request_keywords(dbi_conn conn, unsigned long long n_id, int mode) {
  char* sql_command;
  char keytype[2][10] = {"REFERENCE", "NOTE"};
  int error;
  dbi_result dbires;

  sql_command = (char*)malloc(512); 

  if (sql_command == NULL) {
    return NULL;
  }

  sprintf(sql_command, "SELECT t_keyword.keyword_name FROM t_keyword, t_xkeyword WHERE t_keyword.keyword_id=t_xkeyword.keyword_id AND t_xkeyword.xkeyword_type=\'%s\' AND t_xkeyword.xref_id="ULLSPEC, keytype[mode], (unsigned long long)n_id);

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);
  free(sql_command);
  if (!dbires) {
    return NULL;
  }

  return dbires;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_keyword(): retrieves a keyword

  char* get_author returns a pointer to a string containing the 
                     keyword

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_keyword(dbi_result dbires) {
  const char* result;

  if (dbi_result_next_row(dbires)) {
    result = dbi_result_get_string_idx(dbires, 1);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  request_users(): prepares the retrieval of users of the
                     currently selected database

  dbi_result request_users returns a query result
                     which contains the users

  dbi_conn conn connection to database server

  const char* db ptr to a string with the database name or NULL if the
                    current database should be used

  unsigned long long n_id id of the dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result request_users(dbi_conn conn, const char* db, unsigned long long n_id) {
  char *sql_command;
  char *dbstring;
  const char *drivername;
  char idstring[32];
  int error;
  dbi_result dbires;
  dbi_driver driver;

  driver = dbi_conn_get_driver(conn);
  drivername = dbi_driver_get_name(driver);

  sql_command = (char*)malloc(1024); 

  if (sql_command == NULL) {
    return NULL;
  }

  /* never use db argument with pgsql */
  if (db && *db && !strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
    dbstring = (char*)malloc(strlen(db)+2); 

    if (dbstring == NULL) {
      return NULL;
    }
    sprintf(dbstring, "%s.", db);
  }
  else {
    /* this is weird but simplifies the free() calls further down */
    dbstring = (char*)malloc(1); 

    if (dbstring == NULL) {
      return NULL;
    }
    dbstring[0] = '\0';
  }

  sprintf(sql_command, "SELECT %st_user.user_name FROM %st_user, %st_xuser WHERE %st_user.user_id=%st_xuser.user_id AND %st_xuser.refdb_id="ULLSPEC, dbstring, dbstring, dbstring, dbstring, dbstring, dbstring, (unsigned long long)n_id);

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  free(sql_command);
  free(dbstring);

  return dbires;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_user(): retrieves an user

  char* get_user returns a pointer to a string containing the 
                     user

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_user(dbi_result dbires) {
  const char* result;

  if (dbi_result_next_row(dbires)) {
    result = dbi_result_get_string_idx(dbires, 1);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  clean_request() frees the dbi result data structure after a request is
                  finished

  void clean_request

  dbi_result dbires database query result, pointing to the current
                    dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void clean_request(dbi_result dbires) {
  dbi_result_free(dbires);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_string_copy(): retrieves a given field which is either a
                            string or a binary string from a query result

  char* my_dbi_result_get_string_copy returns a pointer to a string containing the 
                     requested field element, or NULL in case something
		     went wrong. The returned string must be freed by the
		     calling function

  dbi_result dbires database query result, pointing to the current
                    dataset

  const char* fieldname name of the field to retrieve

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* my_dbi_result_get_string_copy(dbi_result dbires, const char* fieldname) {
  const char* result;
  char* myresult;
  unsigned short n_field_type;
  unsigned int n_stringsize;

  n_field_type = dbi_result_get_field_type(dbires, fieldname);

  if (n_field_type == DBI_TYPE_BINARY) {
    /* binary strings are returned without a terminating \0 so we have
       to bend over backwards to treat it as a normal string */
    result = dbi_result_get_binary(dbires, fieldname);
    if (!result || dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      n_stringsize = dbi_result_get_field_size(dbires, fieldname);

      /* need one extra byte for a terminating \0 */
      myresult = (char*)malloc(n_stringsize+1);
      if (myresult == NULL) {
	return NULL;
      }
      memcpy(myresult, result, n_stringsize);
      myresult[n_stringsize] = '\0'; /* terminate string */
      return myresult;
    }
  }
  else if (n_field_type == DBI_TYPE_STRING) {
    result = dbi_result_get_string_copy(dbires, fieldname);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return (char*)result;
    }
  }
  else {
    /* should never happen */
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_string_copy_idx(): retrieves a given field which is either a
                            string or a binary string from a query result

  char* my_dbi_result_get_string_copy_idx returns a pointer to a string containing the 
                     requested field element, or NULL in case something
		     went wrong. The returned string must be freed by the
		     calling function

  dbi_result dbires database query result, pointing to the current
                    dataset

  unsigned int idx index of field to retrieve (1-base)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* my_dbi_result_get_string_copy_idx(dbi_result dbires, unsigned int idx) {
  const char* result;
  char* myresult;
  unsigned short n_field_type;
  unsigned int n_stringsize;

  n_field_type = dbi_result_get_field_type_idx(dbires, idx);

  if (n_field_type == DBI_TYPE_BINARY) {
    /* binary strings are returned without a terminating \0 so we have
       to bend over backwards to treat it as a normal string */
    result = dbi_result_get_binary_idx(dbires, idx);
    if (!result || dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      n_stringsize = dbi_result_get_field_size_idx(dbires, idx);

      /* need one extra byte for a terminating \0 */
      myresult = (char*)malloc(n_stringsize+1);
      if (myresult == NULL) {
	return NULL;
      }
      memcpy(myresult, result, n_stringsize);
      myresult[n_stringsize] = '\0'; /* terminate string */
      return myresult;
    }
  }
  else if (n_field_type == DBI_TYPE_STRING) {
    result = dbi_result_get_string_copy_idx(dbires, idx);
    if (dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
      return NULL;
    }
    else {
      return (char*)result;
    }
  }
  else {
    /* should never happen */
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_binary_copy(): retrieves a given field which is a
                             binary string from a query result

  char* my_dbi_result_get_binary_copy returns a pointer to a string 
                     containing the requested field element, or NULL
		     in case something went wrong. The returned string
		     must be freed by the calling function

  dbi_result Result database query result, pointing to the current
                    dataset

  const char* fieldname name of the field to retrieve

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* my_dbi_result_get_binary_copy(dbi_result Result, const char* fieldname) {
  /* this is a helper function that should be dropped if libdbi accepts
     the patch for dbi_result_get_binary. The issue is whether or not
     a terminating \0 is returned by the dbi function or not. This
     implementation is necessary if dbi does *not* return the \0 */

  unsigned int n_stringsize;
  char *myresult;
  const char *result;
  const char *errstring;
  dbi_conn conn;

  conn = dbi_result_get_conn(Result);

  result = dbi_result_get_binary(Result, fieldname);

  if (dbi_conn_error_flag(conn)) {
    dbi_conn_error(conn, &errstring);
    LOG_PRINT(LOG_WARNING, errstring);
    return NULL;
  }

  n_stringsize = dbi_result_get_field_size(Result, fieldname);

  /* need one extra byte for a terminating \0 */
  myresult = (char*)malloc(n_stringsize+1);
  if (myresult == NULL) {
    LOG_PRINT(LOG_WARNING, "out of memory");
    return NULL;
  }
  memcpy(myresult, result, n_stringsize);
  myresult[n_stringsize] = '\0'; /* terminate string */
  return myresult;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_string(): retrieves a given TEXT field

  const char* my_dbi_result_get_string returns a pointer to a string 
                     containing the requested field element, or NULL
		     in case something went wrong.

  dbi_result Result database query result, pointing to the current
                    dataset

  const char* fieldname name of the field to retrieve

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* my_dbi_result_get_string(dbi_result Result, const char* fieldname) {
  const char* result;
  char* myresult;
  unsigned short n_field_type;
  unsigned int n_stringsize;

  /* this kludgy function is necessary because MySQL treats TEXT fields
     as binary data */
  n_field_type = dbi_result_get_field_type(Result, fieldname);

  if (n_field_type == DBI_TYPE_BINARY) {
    /* this implementation assumes that binary strings are
       returned by the mysql driver including a trailing \0 */
    result = dbi_result_get_binary(Result, fieldname);
    if (!result || dbi_conn_error_flag(dbi_result_get_conn(Result))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else if (n_field_type == DBI_TYPE_STRING) {
    result = dbi_result_get_string(Result, fieldname);
    if (dbi_conn_error_flag(dbi_result_get_conn(Result))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    /* should never happen */
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_string_idx(): retrieves a given TEXT field

  const char* my_dbi_result_get_string_idx returns a pointer to a string 
                     containing the requested field element, or NULL
		     in case something went wrong.

  dbi_result Result database query result, pointing to the current
                    dataset

  unsigned int idx index of the field to retrieve (1-based)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* my_dbi_result_get_string_idx(dbi_result Result, unsigned int idx) {
  const char* result;
  char* myresult;
  unsigned short n_field_type;
  unsigned int n_stringsize;

  /* this kludgy function is necessary because MySQL treats TEXT fields
     as binary data */
  n_field_type = dbi_result_get_field_type_idx(Result, idx);

  if (n_field_type == DBI_TYPE_BINARY) {
    /* this implementation assumes that binary strings are
       returned by the mysql driver including a trailing \0 */
    result = dbi_result_get_binary_idx(Result, idx);
    if (!result || dbi_conn_error_flag(dbi_result_get_conn(Result))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else if (n_field_type == DBI_TYPE_STRING) {
    result = dbi_result_get_string_idx(Result, idx);
    if (dbi_conn_error_flag(dbi_result_get_conn(Result))) {
      return NULL;
    }
    else {
      return result;
    }
  }
  else {
    /* should never happen */
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  get_author_type_string(): returns a string that denotes the
                author type depending on the driver

  const char* get_author_type_string returns a ptr to the result string
                 or NULL in case of an unsupported driver

  dbi_driver driver ptr to a database driver structure

  int type 1 = primary author, 2 = secondary author, 3 = tertiary author

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* get_author_type_string(dbi_driver driver, int type) {
  const char *drivername;

  if (!strcmp(my_dbi_driver_get_cap(driver, "enum"), "t")) {
    return author_type_string[0][type-1];
  }
  else {
    return author_type_string[1][type-1];
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_in_list(): checks from a query result whether a given reference is
                in the personal reference list of a given user

  int is_in_list returns 1 if the reference is in the list, 0 if not,
                 -1 if some error occurs

  dbi_result dbires database query result, pointing to the current
                    dataset

  char* username ptr to string with a username

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int is_in_list(dbi_result dbires, char* username) {
  char *sql_command;
  char *quoted_username;
  int error;
  unsigned long long n_id;
  dbi_result dbires1;
  dbi_driver driver;

  sql_command = (char*)malloc(256); 

  if (sql_command == NULL) {
    return -1;
  }

  driver = dbi_conn_get_driver(dbi_result_get_conn(dbires));

  quoted_username = mstrdup(username);
  if (!quoted_username) {
    free(sql_command);
    return -1;
  }
  if (dbi_driver_quote_string(driver, &quoted_username) == -1) {
    free(quoted_username);
    free(sql_command);
    return -1;
  }

  n_id = my_dbi_result_get_idval(dbires, "refdb_id");

  if (username != NULL && !dbi_conn_error_flag(dbi_result_get_conn(dbires))) {
    sprintf(sql_command, "SELECT t_xuser.xuser_id FROM t_xuser, t_user WHERE t_xuser.user_id=t_user.user_id AND t_user.user_name=%s AND t_xuser.refdb_id="ULLSPEC, quoted_username, (unsigned long long)n_id);
  }
  else {
    free(quoted_username);
    free(sql_command);
    return -1; /* should never happen */
  }

  free(quoted_username);
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires1 = dbi_conn_query(dbi_result_get_conn(dbires), sql_command);
  free(sql_command);
  
  if (!dbires1) {
    return -1;
  }

  if (dbi_result_next_row(dbires1)) {
    dbi_result_free(dbires1);
    return 1;
  }
  else {
    dbi_result_free(dbires1);
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  load_style(): loads the style information for a particular publication
                type

  dbi_result load_style returns query result if ok, NULL if an error occurred

  const char* pubtype ptr to string with the publication type

  char* citestyle_id ptr to string with the ID of the citation style

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_result load_style(const char* pubtype, unsigned long long citstyle_id, dbi_conn conn) {
  int error;
  char* sql_command;
  const char *drivername;

  /* get memory for sql command buffer */
  sql_command = (char*)malloc(8192);
  if (!sql_command) {
    LOG_PRINT(LOG_WARNING, "out of memory");
    return NULL;
  }

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  /*
     0  ID
     1	AUTHORLISTSTYLE
     2	EDITORLISTSTYLE
     3	SEDITORLISTSTYLE
     4	ALLALISTSTYLE
     5	AUTHORLISTABBREVIATEFIRST
     6	EDITORLISTABBREVIATEFIRST
     7	SEDITORLISTABBREVIATEFIRST
     8	ALLALISTABBREVIATEFIRST
     9	AUTHORLISTABBREVIATESUBSEQ
    10	EDITORLISTABBREVIATESUBSEQ
    11	SEDITORLISTABBREVIATESUBSEQ
    12	ALLALISTABBREVIATESUBSEQ
    13	AUTHORLISTABBREVIATEFIRSTMAXAUTHOR
    14	EDITORLISTABBREVIATEFIRSTMAXAUTHOR
    15	SEDITORLISTABBREVIATEFIRSTMAXAUTHOR
    16	ALLALISTABBREVIATEFIRSTMAXAUTHOR
    17	AUTHORLISTABBREVIATESUBSEQMAXAUTHOR
    18	EDITORLISTABBREVIATESUBSEQMAXAUTHOR
    19	SEDITORLISTABBREVIATESUBSEQMAXAUTHOR
    20	ALLALISTABBREVIATESUBSEQMAXAUTHOR
    21	AUTHORLISTABBREVIATEFIRSTDISPLAYAUTHOR
    22	EDITORLISTABBREVIATEFIRSTDISPLAYAUTHOR
    23	SEDITORLISTABBREVIATEFIRSTDISPLAYAUTHOR
    24	ALLALISTABBREVIATEFIRSTDISPLAYAUTHOR
    25	AUTHORLISTABBREVIATESUBSEQDISPLAYAUTHOR
    26	EDITORLISTABBREVIATESUBSEQDISPLAYAUTHOR
    27	SEDITORLISTABBREVIATESUBSEQDISPLAYAUTHOR
    28	ALLALISTABBREVIATESUBSEQDISPLAYAUTHOR
    29	AUTHORLISTAEMPTY
    30	EDITORLISTAEMPTY
    31	SEDITORLISTAEMPTY
    32	ALLALISTAEMPTY
    33	AUTHORLISTASAME
    34	EDITORLISTASAME
    35	SEDITORLISTASAME
    36	ALLALISTASAME
    37	AUTHORLISTPRECEEDING
    38	EDITORLISTPRECEEDING
    39	SEDITORLISTPRECEEDING
    40	ALLALISTPRECEEDING
    41	AUTHORLISTFOLLOWING
    42	EDITORLISTFOLLOWING
    43	SEDITORLISTFOLLOWING
    44	ALLALISTFOLLOWING
    45	AUTHORLISTAUTHORSEPSTWOSEPS
    46	EDITORLISTAUTHORSEPSTWOSEPS
    47	SEDITORLISTAUTHORSEPSTWOSEPS
    48	ALLALISTAUTHORSEPSTWOSEPS
    49	AUTHORLISTAUTHORSEPSTHREESEPSTHREESEPSEACH
    50	EDITORLISTAUTHORSEPSTHREESEPSTHREESEPSEACH
    51	SEDITORLISTAUTHORSEPSTHREESEPSTHREESEPSEACH
    52	ALLALISTAUTHORSEPSTHREESEPSTHREESEPSEACH
    53	AUTHORLISTAUTHORSEPSTHREESEPSTHREESEPSLAST
    54	EDITORLISTAUTHORSEPSTHREESEPSTHREESEPSLAST
    55	SEDITORLISTAUTHORSEPSTHREESEPSTHREESEPSLAST
    56	ALLALISTAUTHORSEPSTHREESEPSTHREESEPSLAST
    57	AUTHORLISTAUTHORNAMESNAMEFIRSTNAMEORDER
    58	EDITORLISTAUTHORNAMESNAMEFIRSTNAMEORDER
    59	SEDITORLISTAUTHORNAMESNAMEFIRSTNAMEORDER
    60	ALLALISTAUTHORNAMESNAMEFIRSTNAMEORDER
    61	AUTHORLISTAUTHORNAMESNAMEFIRSTINITIALSTYLE
    62	EDITORLISTAUTHORNAMESNAMEFIRSTINITIALSTYLE
    63	SEDITORLISTAUTHORNAMESNAMEFIRSTINITIALSTYLE
    64	ALLALISTAUTHORNAMESNAMEFIRSTINITIALSTYLE
    65	AUTHORLISTAUTHORNAMESNAMEFIRSTUPPERCASE
    66	EDITORLISTAUTHORNAMESNAMEFIRSTUPPERCASE
    67	SEDITORLISTAUTHORNAMESNAMEFIRSTUPPERCASE
    68	ALLALISTAUTHORNAMESNAMEFIRSTUPPERCASE
    69	AUTHORLISTAUTHORNAMESNAMEOTHERNAMEORDER
    70	EDITORLISTAUTHORNAMESNAMEOTHERNAMEORDER
    71	SEDITORLISTAUTHORNAMESNAMEOTHERNAMEORDER
    72	ALLALISTAUTHORNAMESNAMEOTHERNAMEORDER
    73	AUTHORLISTAUTHORNAMESNAMEOTHERINITIALSTYLE
    74	EDITORLISTAUTHORNAMESNAMEOTHERINITIALSTYLE
    75	SEDITORLISTAUTHORNAMESNAMEOTHERINITIALSTYLE
    76	ALLALISTAUTHORNAMESNAMEOTHERINITIALSTYLE
    77	AUTHORLISTAUTHORNAMESNAMEOTHERUPPERCASE
    78	EDITORLISTAUTHORNAMESNAMEOTHERUPPERCASE
    79	SEDITORLISTAUTHORNAMESNAMEOTHERUPPERCASE
    80	ALLALISTAUTHORNAMESNAMEOTHERUPPERCASE
    81	AUTHORLISTTEXTTEXTSINGLEPRECEEDING
    82	EDITORLISTTEXTTEXTSINGLEPRECEEDING
    83	SEDITORLISTTEXTTEXTSINGLEPRECEEDING
    84	ALLALISTTEXTTEXTSINGLEPRECEEDING
    85	AUTHORLISTTEXTTEXTSINGLEFOLLOWING
    86	EDITORLISTTEXTTEXTSINGLEFOLLOWING
    87	SEDITORLISTTEXTTEXTSINGLEFOLLOWING
    88	ALLALISTTEXTTEXTSINGLEFOLLOWING
    89	AUTHORLISTTEXTTEXTMULTIPLEPRECEEDING
    90	EDITORLISTTEXTTEXTMULTIPLEPRECEEDING
    91	SEDITORLISTTEXTTEXTMULTIPLEPRECEEDING
    92	ALLALISTTEXTTEXTMULTIPLEPRECEEDING
    93	AUTHORLISTTEXTTEXTMULTIPLEFOLLOWING
    94	EDITORLISTTEXTTEXTMULTIPLEFOLLOWING
    95	SEDITORLISTTEXTTEXTMULTIPLEFOLLOWING
    96	ALLALISTTEXTTEXTMULTIPLEFOLLOWING
    97	PUBDATEPRECEEDING
    98	PUBDATEFOLLOWING
    99	PUBDATEFORMAT
   100	PUBDATESEQUENCE
   101	PUBDATEMONTHFORMAT
   102	PUBDATEDAYFORMAT
   103	PUBDATEYEARFORMAT
   104	PUBDATEPADLEADINGZERO
   105	PUBDATESTYLE
   106	PUBDATESECPRECEEDING
   107	PUBDATESECFOLLOWING
   108	PUBDATESECFORMAT
   109	PUBDATESECSEQUENCE
   110	PUBDATESECMONTHFORMAT
   111	PUBDATESECDAYFORMAT
   112	PUBDATESECYEARFORMAT
   113	PUBDATESECPADLEADINGZERO
   114	PUBDATESECSTYLE
   115	PUBDATEALLPRECEEDING
   116	PUBDATEALLFOLLOWING
   117	PUBDATEALLFORMAT
   118	PUBDATEALLSEQUENCE
   119	PUBDATEALLMONTHFORMAT
   120	PUBDATEALLDAYFORMAT
   121	PUBDATEALLYEARFORMAT
   122	PUBDATEALLPADLEADINGZERO
   123	PUBDATEALLSTYLE
   124	TITLEPRECEEDING
   125	TITLEFOLLOWING
   126	TITLECASE
   127	TITLESTYLE
   128	BOOKTITLEPRECEEDING
   129	BOOKTITLEFOLLOWING
   130	BOOKTITLECASE
   131	BOOKTITLESTYLE
   132	SERIESTITLEPRECEEDING
   133	SERIESTITLEFOLLOWING
   134	SERIESTITLECASE
   135	SERIESTITLESTYLE
   136	ALLTITLEPRECEEDING
   137	ALLTITLEFOLLOWING
   138	ALLTITLECASE
   139	ALLTITLESTYLE
   140	JOURNALNAMEPRECEEDING
   141	JOURNALNAMEFOLLOWING
   142	JOURNALNAMECASE
   143	JOURNALNAMEDEFAULTTEXT
   144	JOURNALNAMEALTERNATETEXT
   145	JOURNALNAMEPUNCTUATION
   146	JOURNALNAMESTYLE
   147	VOLUMEPRECEEDING
   148	VOLUMEFOLLOWING
   149	VOLUMESTYLE
   150	ISSUEPRECEEDING
   151	ISSUEFOLLOWING
   152	ISSUESTYLE
   153	PAGESSTYLE
   154	PAGESSINGLEPAGEPRECEEDING
   155	PAGESSINGLEPAGEFOLLOWING
   156	PAGESPAGERANGEPRECEEDING
   157	PAGESPAGERANGEFOLLOWING
   158	PAGESPAGERANGETYPE
   159	PUBLISHERPRECEEDING
   160	PUBLISHERFOLLOWING
   161	PUBLISHERSTYLE
   162	PUBPLACEPRECEEDING
   163	PUBPLACEFOLLOWING
   164	PUBPLACESTYLE
   165	PAGESPRECEEDING
   166	PAGESFOLLOWING
   167	REFNUMBERPRECEEDING
   168	REFNUMBERFOLLOWING
   169	REFNUMBERSTYLE
   170	PUBDATEFIRSTSEP
   171	PUBDATESECONDSEP
   172	PUBDATESECFIRSTSEP
   173	PUBDATESECSECONDSEP
   174	PUBDATEALLFIRSTSEP
   175	PUBDATEALLSECONDSEP
   176  PAGESPAGERANGERANGESEPARATOR
   177  SERIALPRECEEDING
   178  SERIALFOLLOWING
   179  SERIALSTYLE
   180  ADDRESSPRECEEDING
   181  ADDRESSFOLLOWING
   182  ADDRESSSTYLE
   183  URLPRECEEDING
   184  URLFOLLOWING
   185  URLSTYLE
   186  USERDEF1PRECEEDING
   187  USERDEF1FOLLOWING
   188  USERDEF1STYLE
   189  USERDEF2PRECEEDING
   190  USERDEF2FOLLOWING
   191  USERDEF2STYLE
   192  USERDEF3PRECEEDING
   193  USERDEF3FOLLOWING
   194  USERDEF3STYLE
   195  USERDEF4PRECEEDING
   196  USERDEF4FOLLOWING
   197  USERDEF4STYLE
   198  USERDEF5PRECEEDING
   199  USERDEF5FOLLOWING
   200  USERDEF5STYLE
   201  MISC1PRECEEDING
   202  MISC1FOLLOWING
   203  MISC1STYLE
   204  MISC2PRECEEDING
   205  MISC2FOLLOWING
   206  MISC2STYLE
   207  MISC3PRECEEDING
   208  MISC3FOLLOWING
   209  MISC3STYLE
   210  NOTESPRECEEDING
   211  NOTESFOLLOWING
   212  NOTESSTYLE
   213  ABSTRACTPRECEEDING
   214  ABSTRACTFOLLOWING
   215  ABSTRACTSTYLE
   216  LINK1PRECEEDING
   217  LINK1FOLLOWING
   218  LINK1STYLE
   219  LINK2PRECEEDING
   220  LINK2FOLLOWING
   221  LINK2STYLE
   222  LINK3PRECEEDING
   223  LINK3FOLLOWING
   224  LINK3STYLE
   225  LINK4PRECEEDING
   226  LINK4FOLLOWING
   227  LINK4STYLE
   */

  /* now retrieve matching style information from REFSTYLE into*/
  if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
    sprintf(sql_command, "SELECT ID, QSTYLE, XSTYLE, YSTYLE, ZSTYLE, QABBREVIATEFIRST, XABBREVIATEFIRST, YABBREVIATEFIRST, ZABBREVIATEFIRST, QABBREVIATESUBSEQ, XABBREVIATESUBSEQ, YABBREVIATESUBSEQ, ZABBREVIATESUBSEQ, QABBREVIATEFIRSTMAXAUTHOR, XABBREVIATEFIRSTMAXAUTHOR, YABBREVIATEFIRSTMAXAUTHOR, ZABBREVIATEFIRSTMAXAUTHOR, QABBREVIATESUBSEQMAXAUTHOR, XABBREVIATESUBSEQMAXAUTHOR, YABBREVIATESUBSEQMAXAUTHOR, ZABBREVIATESUBSEQMAXAUTHOR, QABBREVIATEFIRSTDISPLAYAUTHOR, XABBREVIATEFIRSTDISPLAYAUTHOR, YABBREVIATEFIRSTDISPLAYAUTHOR, ZABBREVIATEFIRSTDISPLAYAUTHOR, QABBREVIATESUBSEQDISPLAYAUTHOR, XABBREVIATESUBSEQDISPLAYAUTHOR, YABBREVIATESUBSEQDISPLAYAUTHOR, ZABBREVIATESUBSEQDISPLAYAUTHOR, QAEMPTY, XAEMPTY, YAEMPTY, ZAEMPTY, QASAME, XASAME, YASAME, ZASAME, QPRECEEDING, XPRECEEDING, YPRECEEDING, ZPRECEEDING, QFOLLOWING, XFOLLOWING, YFOLLOWING, ZFOLLOWING, QAUTHORSEPSTWOSEPS, XAUTHORSEPSTWOSEPS, YAUTHORSEPSTWOSEPS, ZAUTHORSEPSTWOSEPS, QAUTHORSEPSTHREESEPSTHREESEPSEACH, XAUTHORSEPSTHREESEPSTHREESEPSEACH, YAUTHORSEPSTHREESEPSTHREESEPSEACH, ZAUTHORSEPSTHREESEPSTHREESEPSEACH, QAUTHORSEPSTHREESEPSTHREESEPSLAST, XAUTHORSEPSTHREESEPSTHREESEPSLAST, YAUTHORSEPSTHREESEPSTHREESEPSLAST, ZAUTHORSEPSTHREESEPSTHREESEPSLAST, QAUTHORNAMESNAMEFIRSTNAMEORDER, XAUTHORNAMESNAMEFIRSTNAMEORDER, YAUTHORNAMESNAMEFIRSTNAMEORDER, ZAUTHORNAMESNAMEFIRSTNAMEORDER, QAUTHORNAMESNAMEFIRSTINITIALSTYLE, XAUTHORNAMESNAMEFIRSTINITIALSTYLE, YAUTHORNAMESNAMEFIRSTINITIALSTYLE, ZAUTHORNAMESNAMEFIRSTINITIALSTYLE, QAUTHORNAMESNAMEFIRSTUPPERCASE, XAUTHORNAMESNAMEFIRSTUPPERCASE, YAUTHORNAMESNAMEFIRSTUPPERCASE, ZAUTHORNAMESNAMEFIRSTUPPERCASE, QAUTHORNAMESNAMEOTHERNAMEORDER, XAUTHORNAMESNAMEOTHERNAMEORDER, YAUTHORNAMESNAMEOTHERNAMEORDER, ZAUTHORNAMESNAMEOTHERNAMEORDER, QAUTHORNAMESNAMEOTHERINITIALSTYLE, XAUTHORNAMESNAMEOTHERINITIALSTYLE, YAUTHORNAMESNAMEOTHERINITIALSTYLE, ZAUTHORNAMESNAMEOTHERINITIALSTYLE, QAUTHORNAMESNAMEOTHERUPPERCASE, XAUTHORNAMESNAMEOTHERUPPERCASE, YAUTHORNAMESNAMEOTHERUPPERCASE, ZAUTHORNAMESNAMEOTHERUPPERCASE, QTEXTTEXTSINGLEPRECEEDING, XTEXTTEXTSINGLEPRECEEDING, YTEXTTEXTSINGLEPRECEEDING, ZTEXTTEXTSINGLEPRECEEDING, QTEXTTEXTSINGLEFOLLOWING, XTEXTTEXTSINGLEFOLLOWING, YTEXTTEXTSINGLEFOLLOWING, ZTEXTTEXTSINGLEFOLLOWING, QTEXTTEXTMULTIPLEPRECEEDING, XTEXTTEXTMULTIPLEPRECEEDING, YTEXTTEXTMULTIPLEPRECEEDING, ZTEXTTEXTMULTIPLEPRECEEDING, QTEXTTEXTMULTIPLEFOLLOWING, XTEXTTEXTMULTIPLEFOLLOWING, YTEXTTEXTMULTIPLEFOLLOWING, ZTEXTTEXTMULTIPLEFOLLOWING, PUBDATEPRECEEDING, PUBDATEFOLLOWING, PUBDATEFORMAT, PUBDATESEQUENCE, PUBDATEMONTHFORMAT, PUBDATEDAYFORMAT, PUBDATEYEARFORMAT, PUBDATEPADLEADINGZERO, PUBDATESTYLE, PUBDATESECPRECEEDING, PUBDATESECFOLLOWING, PUBDATESECFORMAT, PUBDATESECSEQUENCE, PUBDATESECMONTHFORMAT, PUBDATESECDAYFORMAT, PUBDATESECYEARFORMAT, PUBDATESECPADLEADINGZERO, PUBDATESECSTYLE, PUBDATEALLPRECEEDING, PUBDATEALLFOLLOWING, PUBDATEALLFORMAT, PUBDATEALLSEQUENCE, PUBDATEALLMONTHFORMAT, PUBDATEALLDAYFORMAT, PUBDATEALLYEARFORMAT, PUBDATEALLPADLEADINGZERO, PUBDATEALLSTYLE, TITLEPRECEEDING, TITLEFOLLOWING, TITLECASE, TITLESTYLE, BOOKTITLEPRECEEDING, BOOKTITLEFOLLOWING, BOOKTITLECASE, BOOKTITLESTYLE, SERIESTITLEPRECEEDING, SERIESTITLEFOLLOWING, SERIESTITLECASE, SERIESTITLESTYLE, ALLTITLEPRECEEDING, ALLTITLEFOLLOWING, ALLTITLECASE, ALLTITLESTYLE, JOURNALNAMEPRECEEDING, JOURNALNAMEFOLLOWING, JOURNALNAMECASE, JOURNALNAMEDEFAULTTEXT, JOURNALNAMEALTERNATETEXT, JOURNALNAMEPUNCTUATION, JOURNALNAMESTYLE, VOLUMEPRECEEDING, VOLUMEFOLLOWING, VOLUMESTYLE, ISSUEPRECEEDING, ISSUEFOLLOWING, ISSUESTYLE, PAGESSTYLE, PAGESSINGLEPAGEPRECEEDING, PAGESSINGLEPAGEFOLLOWING, PAGESPAGERANGEPRECEEDING, PAGESPAGERANGEFOLLOWING, PAGESPAGERANGETYPE, PUBLISHERPRECEEDING, PUBLISHERFOLLOWING, PUBLISHERSTYLE, PUBPLACEPRECEEDING, PUBPLACEFOLLOWING, PUBPLACESTYLE, PAGESPRECEEDING, PAGESFOLLOWING, REFNUMBERPRECEEDING, REFNUMBERFOLLOWING, REFNUMBERSTYLE, PUBDATEFIRSTSEP, PUBDATESECONDSEP, PUBDATESECFIRSTSEP, PUBDATESECSECONDSEP, PUBDATEALLFIRSTSEP, PUBDATEALLSECONDSEP, PAGESPAGERANGERANGESEPARATOR, SERIALPRECEEDING, SERIALFOLLOWING, SERIALSTYLE, ADDRESSPRECEEDING, ADDRESSFOLLOWING, ADDRESSSTYLE, URLPRECEEDING, URLFOLLOWING, URLSTYLE, USERDEF1PRECEEDING, USERDEF1FOLLOWING, USERDEF1STYLE, USERDEF2PRECEEDING, USERDEF2FOLLOWING, USERDEF2STYLE, USERDEF3PRECEEDING, USERDEF3FOLLOWING, USERDEF3STYLE, USERDEF4PRECEEDING, USERDEF4FOLLOWING, USERDEF4STYLE, USERDEF5PRECEEDING, USERDEF5FOLLOWING, USERDEF5STYLE, MISC1PRECEEDING, MISC1FOLLOWING, MISC1STYLE, MISC2PRECEEDING, MISC2FOLLOWING, MISC2STYLE, MISC3PRECEEDING, MISC3FOLLOWING, MISC3STYLE, NOTESPRECEEDING, NOTESFOLLOWING, NOTESSTYLE, ABSTRACTPRECEEDING, ABSTRACTFOLLOWING, ABSTRACTSTYLE, LINK1PRECEEDING, LINK1FOLLOWING, LINK1STYLE, LINK2PRECEEDING, LINK2FOLLOWING, LINK2STYLE, LINK3PRECEEDING, LINK3FOLLOWING, LINK3STYLE, LINK4PRECEEDING, LINK4FOLLOWING, LINK4STYLE from %s.REFSTYLE where PUBTYPE='%s' and CITSTYLEID="ULLSPEC, MAIN_DB, pubtype, (unsigned long long)citstyle_id);
  }
  else {
    sprintf(sql_command, "SELECT ID, QSTYLE, XSTYLE, YSTYLE, ZSTYLE, QABBREVIATEFIRST, XABBREVIATEFIRST, YABBREVIATEFIRST, ZABBREVIATEFIRST, QABBREVIATESUBSEQ, XABBREVIATESUBSEQ, YABBREVIATESUBSEQ, ZABBREVIATESUBSEQ, QABBREVIATEFIRSTMAXAUTHOR, XABBREVIATEFIRSTMAXAUTHOR, YABBREVIATEFIRSTMAXAUTHOR, ZABBREVIATEFIRSTMAXAUTHOR, QABBREVIATESUBSEQMAXAUTHOR, XABBREVIATESUBSEQMAXAUTHOR, YABBREVIATESUBSEQMAXAUTHOR, ZABBREVIATESUBSEQMAXAUTHOR, QABBREVIATEFIRSTDISPLAYAUTHOR, XABBREVIATEFIRSTDISPLAYAUTHOR, YABBREVIATEFIRSTDISPLAYAUTHOR, ZABBREVIATEFIRSTDISPLAYAUTHOR, QABBREVIATESUBSEQDISPLAYAUTHOR, XABBREVIATESUBSEQDISPLAYAUTHOR, YABBREVIATESUBSEQDISPLAYAUTHOR, ZABBREVIATESUBSEQDISPLAYAUTHOR, QAEMPTY, XAEMPTY, YAEMPTY, ZAEMPTY, QASAME, XASAME, YASAME, ZASAME, QPRECEEDING, XPRECEEDING, YPRECEEDING, ZPRECEEDING, QFOLLOWING, XFOLLOWING, YFOLLOWING, ZFOLLOWING, QAUTHORSEPSTWOSEPS, XAUTHORSEPSTWOSEPS, YAUTHORSEPSTWOSEPS, ZAUTHORSEPSTWOSEPS, QAUTHORSEPSTHREESEPSTHREESEPSEACH, XAUTHORSEPSTHREESEPSTHREESEPSEACH, YAUTHORSEPSTHREESEPSTHREESEPSEACH, ZAUTHORSEPSTHREESEPSTHREESEPSEACH, QAUTHORSEPSTHREESEPSTHREESEPSLAST, XAUTHORSEPSTHREESEPSTHREESEPSLAST, YAUTHORSEPSTHREESEPSTHREESEPSLAST, ZAUTHORSEPSTHREESEPSTHREESEPSLAST, QAUTHORNAMESNAMEFIRSTNAMEORDER, XAUTHORNAMESNAMEFIRSTNAMEORDER, YAUTHORNAMESNAMEFIRSTNAMEORDER, ZAUTHORNAMESNAMEFIRSTNAMEORDER, QAUTHORNAMESNAMEFIRSTINITIALSTYLE, XAUTHORNAMESNAMEFIRSTINITIALSTYLE, YAUTHORNAMESNAMEFIRSTINITIALSTYLE, ZAUTHORNAMESNAMEFIRSTINITIALSTYLE, QAUTHORNAMESNAMEFIRSTUPPERCASE, XAUTHORNAMESNAMEFIRSTUPPERCASE, YAUTHORNAMESNAMEFIRSTUPPERCASE, ZAUTHORNAMESNAMEFIRSTUPPERCASE, QAUTHORNAMESNAMEOTHERNAMEORDER, XAUTHORNAMESNAMEOTHERNAMEORDER, YAUTHORNAMESNAMEOTHERNAMEORDER, ZAUTHORNAMESNAMEOTHERNAMEORDER, QAUTHORNAMESNAMEOTHERINITIALSTYLE, XAUTHORNAMESNAMEOTHERINITIALSTYLE, YAUTHORNAMESNAMEOTHERINITIALSTYLE, ZAUTHORNAMESNAMEOTHERINITIALSTYLE, QAUTHORNAMESNAMEOTHERUPPERCASE, XAUTHORNAMESNAMEOTHERUPPERCASE, YAUTHORNAMESNAMEOTHERUPPERCASE, ZAUTHORNAMESNAMEOTHERUPPERCASE, QTEXTTEXTSINGLEPRECEEDING, XTEXTTEXTSINGLEPRECEEDING, YTEXTTEXTSINGLEPRECEEDING, ZTEXTTEXTSINGLEPRECEEDING, QTEXTTEXTSINGLEFOLLOWING, XTEXTTEXTSINGLEFOLLOWING, YTEXTTEXTSINGLEFOLLOWING, ZTEXTTEXTSINGLEFOLLOWING, QTEXTTEXTMULTIPLEPRECEEDING, XTEXTTEXTMULTIPLEPRECEEDING, YTEXTTEXTMULTIPLEPRECEEDING, ZTEXTTEXTMULTIPLEPRECEEDING, QTEXTTEXTMULTIPLEFOLLOWING, XTEXTTEXTMULTIPLEFOLLOWING, YTEXTTEXTMULTIPLEFOLLOWING, ZTEXTTEXTMULTIPLEFOLLOWING, PUBDATEPRECEEDING, PUBDATEFOLLOWING, PUBDATEFORMAT, PUBDATESEQUENCE, PUBDATEMONTHFORMAT, PUBDATEDAYFORMAT, PUBDATEYEARFORMAT, PUBDATEPADLEADINGZERO, PUBDATESTYLE, PUBDATESECPRECEEDING, PUBDATESECFOLLOWING, PUBDATESECFORMAT, PUBDATESECSEQUENCE, PUBDATESECMONTHFORMAT, PUBDATESECDAYFORMAT, PUBDATESECYEARFORMAT, PUBDATESECPADLEADINGZERO, PUBDATESECSTYLE, PUBDATEALLPRECEEDING, PUBDATEALLFOLLOWING, PUBDATEALLFORMAT, PUBDATEALLSEQUENCE, PUBDATEALLMONTHFORMAT, PUBDATEALLDAYFORMAT, PUBDATEALLYEARFORMAT, PUBDATEALLPADLEADINGZERO, PUBDATEALLSTYLE, TITLEPRECEEDING, TITLEFOLLOWING, TITLECASE, TITLESTYLE, BOOKTITLEPRECEEDING, BOOKTITLEFOLLOWING, BOOKTITLECASE, BOOKTITLESTYLE, SERIESTITLEPRECEEDING, SERIESTITLEFOLLOWING, SERIESTITLECASE, SERIESTITLESTYLE, ALLTITLEPRECEEDING, ALLTITLEFOLLOWING, ALLTITLECASE, ALLTITLESTYLE, JOURNALNAMEPRECEEDING, JOURNALNAMEFOLLOWING, JOURNALNAMECASE, JOURNALNAMEDEFAULTTEXT, JOURNALNAMEALTERNATETEXT, JOURNALNAMEPUNCTUATION, JOURNALNAMESTYLE, VOLUMEPRECEEDING, VOLUMEFOLLOWING, VOLUMESTYLE, ISSUEPRECEEDING, ISSUEFOLLOWING, ISSUESTYLE, PAGESSTYLE, PAGESSINGLEPAGEPRECEEDING, PAGESSINGLEPAGEFOLLOWING, PAGESPAGERANGEPRECEEDING, PAGESPAGERANGEFOLLOWING, PAGESPAGERANGETYPE, PUBLISHERPRECEEDING, PUBLISHERFOLLOWING, PUBLISHERSTYLE, PUBPLACEPRECEEDING, PUBPLACEFOLLOWING, PUBPLACESTYLE, PAGESPRECEEDING, PAGESFOLLOWING, REFNUMBERPRECEEDING, REFNUMBERFOLLOWING, REFNUMBERSTYLE, PUBDATEFIRSTSEP, PUBDATESECONDSEP, PUBDATESECFIRSTSEP, PUBDATESECSECONDSEP, PUBDATEALLFIRSTSEP, PUBDATEALLSECONDSEP, PAGESPAGERANGERANGESEPARATOR, SERIALPRECEEDING, SERIALFOLLOWING, SERIALSTYLE, ADDRESSPRECEEDING, ADDRESSFOLLOWING, ADDRESSSTYLE, URLPRECEEDING, URLFOLLOWING, URLSTYLE, USERDEF1PRECEEDING, USERDEF1FOLLOWING, USERDEF1STYLE, USERDEF2PRECEEDING, USERDEF2FOLLOWING, USERDEF2STYLE, USERDEF3PRECEEDING, USERDEF3FOLLOWING, USERDEF3STYLE, USERDEF4PRECEEDING, USERDEF4FOLLOWING, USERDEF4STYLE, USERDEF5PRECEEDING, USERDEF5FOLLOWING, USERDEF5STYLE, MISC1PRECEEDING, MISC1FOLLOWING, MISC1STYLE, MISC2PRECEEDING, MISC2FOLLOWING, MISC2STYLE, MISC3PRECEEDING, MISC3FOLLOWING, MISC3STYLE, NOTESPRECEEDING, NOTESFOLLOWING, NOTESSTYLE, ABSTRACTPRECEEDING, ABSTRACTFOLLOWING, ABSTRACTSTYLE, LINK1PRECEEDING, LINK1FOLLOWING, LINK1STYLE, LINK2PRECEEDING, LINK2FOLLOWING, LINK2STYLE, LINK3PRECEEDING, LINK3FOLLOWING, LINK3STYLE, LINK4PRECEEDING, LINK4FOLLOWING, LINK4STYLE from REFSTYLE where PUBTYPE='%s' and CITSTYLEID="ULLSPEC, pubtype, (unsigned long long)citstyle_id);
  }

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbi_style_res = dbi_conn_query(conn, sql_command);
  if (!dbi_style_res) {
    dbi_conn_error(conn, (const char**)&sql_command);
    LOG_PRINT(LOG_WARNING, "CITSTYLE select failed");
    LOG_PRINT(LOG_WARNING, sql_command);
    free(sql_command);
    return NULL;
  }

  if (dbi_result_next_row(dbi_style_res) == 0) {
    /* there is no style for this citation type. Try the generic type instead */
    dbi_result_free(dbi_style_res);
    if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
      sprintf(sql_command, "SELECT ID, QSTYLE, XSTYLE, YSTYLE, ZSTYLE, QABBREVIATEFIRST, XABBREVIATEFIRST, YABBREVIATEFIRST, ZABBREVIATEFIRST, QABBREVIATESUBSEQ, XABBREVIATESUBSEQ, YABBREVIATESUBSEQ, ZABBREVIATESUBSEQ, QABBREVIATEFIRSTMAXAUTHOR, XABBREVIATEFIRSTMAXAUTHOR, YABBREVIATEFIRSTMAXAUTHOR, ZABBREVIATEFIRSTMAXAUTHOR, QABBREVIATESUBSEQMAXAUTHOR, XABBREVIATESUBSEQMAXAUTHOR, YABBREVIATESUBSEQMAXAUTHOR, ZABBREVIATESUBSEQMAXAUTHOR, QABBREVIATEFIRSTDISPLAYAUTHOR, XABBREVIATEFIRSTDISPLAYAUTHOR, YABBREVIATEFIRSTDISPLAYAUTHOR, ZABBREVIATEFIRSTDISPLAYAUTHOR, QABBREVIATESUBSEQDISPLAYAUTHOR, XABBREVIATESUBSEQDISPLAYAUTHOR, YABBREVIATESUBSEQDISPLAYAUTHOR, ZABBREVIATESUBSEQDISPLAYAUTHOR, QAEMPTY, XAEMPTY, YAEMPTY, ZAEMPTY, QASAME, XASAME, YASAME, ZASAME, QPRECEEDING, XPRECEEDING, YPRECEEDING, ZPRECEEDING, QFOLLOWING, XFOLLOWING, YFOLLOWING, ZFOLLOWING, QAUTHORSEPSTWOSEPS, XAUTHORSEPSTWOSEPS, YAUTHORSEPSTWOSEPS, ZAUTHORSEPSTWOSEPS, QAUTHORSEPSTHREESEPSTHREESEPSEACH, XAUTHORSEPSTHREESEPSTHREESEPSEACH, YAUTHORSEPSTHREESEPSTHREESEPSEACH, ZAUTHORSEPSTHREESEPSTHREESEPSEACH, QAUTHORSEPSTHREESEPSTHREESEPSLAST, XAUTHORSEPSTHREESEPSTHREESEPSLAST, YAUTHORSEPSTHREESEPSTHREESEPSLAST, ZAUTHORSEPSTHREESEPSTHREESEPSLAST, QAUTHORNAMESNAMEFIRSTNAMEORDER, XAUTHORNAMESNAMEFIRSTNAMEORDER, YAUTHORNAMESNAMEFIRSTNAMEORDER, ZAUTHORNAMESNAMEFIRSTNAMEORDER, QAUTHORNAMESNAMEFIRSTINITIALSTYLE, XAUTHORNAMESNAMEFIRSTINITIALSTYLE, YAUTHORNAMESNAMEFIRSTINITIALSTYLE, ZAUTHORNAMESNAMEFIRSTINITIALSTYLE, QAUTHORNAMESNAMEFIRSTUPPERCASE, XAUTHORNAMESNAMEFIRSTUPPERCASE, YAUTHORNAMESNAMEFIRSTUPPERCASE, ZAUTHORNAMESNAMEFIRSTUPPERCASE, QAUTHORNAMESNAMEOTHERNAMEORDER, XAUTHORNAMESNAMEOTHERNAMEORDER, YAUTHORNAMESNAMEOTHERNAMEORDER, ZAUTHORNAMESNAMEOTHERNAMEORDER, QAUTHORNAMESNAMEOTHERINITIALSTYLE, XAUTHORNAMESNAMEOTHERINITIALSTYLE, YAUTHORNAMESNAMEOTHERINITIALSTYLE, ZAUTHORNAMESNAMEOTHERINITIALSTYLE, QAUTHORNAMESNAMEOTHERUPPERCASE, XAUTHORNAMESNAMEOTHERUPPERCASE, YAUTHORNAMESNAMEOTHERUPPERCASE, ZAUTHORNAMESNAMEOTHERUPPERCASE, QTEXTTEXTSINGLEPRECEEDING, XTEXTTEXTSINGLEPRECEEDING, YTEXTTEXTSINGLEPRECEEDING, ZTEXTTEXTSINGLEPRECEEDING, QTEXTTEXTSINGLEFOLLOWING, XTEXTTEXTSINGLEFOLLOWING, YTEXTTEXTSINGLEFOLLOWING, ZTEXTTEXTSINGLEFOLLOWING, QTEXTTEXTMULTIPLEPRECEEDING, XTEXTTEXTMULTIPLEPRECEEDING, YTEXTTEXTMULTIPLEPRECEEDING, ZTEXTTEXTMULTIPLEPRECEEDING, QTEXTTEXTMULTIPLEFOLLOWING, XTEXTTEXTMULTIPLEFOLLOWING, YTEXTTEXTMULTIPLEFOLLOWING, ZTEXTTEXTMULTIPLEFOLLOWING, PUBDATEPRECEEDING, PUBDATEFOLLOWING, PUBDATEFORMAT, PUBDATESEQUENCE, PUBDATEMONTHFORMAT, PUBDATEDAYFORMAT, PUBDATEYEARFORMAT, PUBDATEPADLEADINGZERO, PUBDATESTYLE, PUBDATESECPRECEEDING, PUBDATESECFOLLOWING, PUBDATESECFORMAT, PUBDATESECSEQUENCE, PUBDATESECMONTHFORMAT, PUBDATESECDAYFORMAT, PUBDATESECYEARFORMAT, PUBDATESECPADLEADINGZERO, PUBDATESECSTYLE, PUBDATEALLPRECEEDING, PUBDATEALLFOLLOWING, PUBDATEALLFORMAT, PUBDATEALLSEQUENCE, PUBDATEALLMONTHFORMAT, PUBDATEALLDAYFORMAT, PUBDATEALLYEARFORMAT, PUBDATEALLPADLEADINGZERO, PUBDATEALLSTYLE, TITLEPRECEEDING, TITLEFOLLOWING, TITLECASE, TITLESTYLE, BOOKTITLEPRECEEDING, BOOKTITLEFOLLOWING, BOOKTITLECASE, BOOKTITLESTYLE, SERIESTITLEPRECEEDING, SERIESTITLEFOLLOWING, SERIESTITLECASE, SERIESTITLESTYLE, ALLTITLEPRECEEDING, ALLTITLEFOLLOWING, ALLTITLECASE, ALLTITLESTYLE, JOURNALNAMEPRECEEDING, JOURNALNAMEFOLLOWING, JOURNALNAMECASE, JOURNALNAMEDEFAULTTEXT, JOURNALNAMEALTERNATETEXT, JOURNALNAMEPUNCTUATION, JOURNALNAMESTYLE, VOLUMEPRECEEDING, VOLUMEFOLLOWING, VOLUMESTYLE, ISSUEPRECEEDING, ISSUEFOLLOWING, ISSUESTYLE, PAGESSTYLE, PAGESSINGLEPAGEPRECEEDING, PAGESSINGLEPAGEFOLLOWING, PAGESPAGERANGEPRECEEDING, PAGESPAGERANGEFOLLOWING, PAGESPAGERANGETYPE, PUBLISHERPRECEEDING, PUBLISHERFOLLOWING, PUBLISHERSTYLE, PUBPLACEPRECEEDING, PUBPLACEFOLLOWING, PUBPLACESTYLE, PAGESPRECEEDING, PAGESFOLLOWING, REFNUMBERPRECEEDING, REFNUMBERFOLLOWING, REFNUMBERSTYLE, PUBDATEFIRSTSEP, PUBDATESECONDSEP, PUBDATESECFIRSTSEP, PUBDATESECSECONDSEP, PUBDATEALLFIRSTSEP, PUBDATEALLSECONDSEP, PAGESPAGERANGERANGESEPARATOR, SERIALPRECEEDING, SERIALFOLLOWING, SERIALSTYLE, ADDRESSPRECEEDING, ADDRESSFOLLOWING, ADDRESSSTYLE, URLPRECEEDING, URLFOLLOWING, URLSTYLE, USERDEF1PRECEEDING, USERDEF1FOLLOWING, USERDEF1STYLE, USERDEF2PRECEEDING, USERDEF2FOLLOWING, USERDEF2STYLE, USERDEF3PRECEEDING, USERDEF3FOLLOWING, USERDEF3STYLE, USERDEF4PRECEEDING, USERDEF4FOLLOWING, USERDEF4STYLE, USERDEF5PRECEEDING, USERDEF5FOLLOWING, USERDEF5STYLE, MISC1PRECEEDING, MISC1FOLLOWING, MISC1STYLE, MISC2PRECEEDING, MISC2FOLLOWING, MISC2STYLE, MISC3PRECEEDING, MISC3FOLLOWING, MISC3STYLE, NOTESPRECEEDING, NOTESFOLLOWING, NOTESSTYLE, ABSTRACTPRECEEDING, ABSTRACTFOLLOWING, ABSTRACTSTYLE, LINK1PRECEEDING, LINK1FOLLOWING, LINK1STYLE, LINK2PRECEEDING, LINK2FOLLOWING, LINK2STYLE, LINK3PRECEEDING, LINK3FOLLOWING, LINK3STYLE, LINK4PRECEEDING, LINK4FOLLOWING, LINK4STYLE from %s.REFSTYLE where PUBTYPE='GEN' and CITSTYLEID="ULLSPEC, MAIN_DB, (unsigned long long)citstyle_id);
    }
    else {
      sprintf(sql_command, "SELECT ID, QSTYLE, XSTYLE, YSTYLE, ZSTYLE, QABBREVIATEFIRST, XABBREVIATEFIRST, YABBREVIATEFIRST, ZABBREVIATEFIRST, QABBREVIATESUBSEQ, XABBREVIATESUBSEQ, YABBREVIATESUBSEQ, ZABBREVIATESUBSEQ, QABBREVIATEFIRSTMAXAUTHOR, XABBREVIATEFIRSTMAXAUTHOR, YABBREVIATEFIRSTMAXAUTHOR, ZABBREVIATEFIRSTMAXAUTHOR, QABBREVIATESUBSEQMAXAUTHOR, XABBREVIATESUBSEQMAXAUTHOR, YABBREVIATESUBSEQMAXAUTHOR, ZABBREVIATESUBSEQMAXAUTHOR, QABBREVIATEFIRSTDISPLAYAUTHOR, XABBREVIATEFIRSTDISPLAYAUTHOR, YABBREVIATEFIRSTDISPLAYAUTHOR, ZABBREVIATEFIRSTDISPLAYAUTHOR, QABBREVIATESUBSEQDISPLAYAUTHOR, XABBREVIATESUBSEQDISPLAYAUTHOR, YABBREVIATESUBSEQDISPLAYAUTHOR, ZABBREVIATESUBSEQDISPLAYAUTHOR, QAEMPTY, XAEMPTY, YAEMPTY, ZAEMPTY, QASAME, XASAME, YASAME, ZASAME, QPRECEEDING, XPRECEEDING, YPRECEEDING, ZPRECEEDING, QFOLLOWING, XFOLLOWING, YFOLLOWING, ZFOLLOWING, QAUTHORSEPSTWOSEPS, XAUTHORSEPSTWOSEPS, YAUTHORSEPSTWOSEPS, ZAUTHORSEPSTWOSEPS, QAUTHORSEPSTHREESEPSTHREESEPSEACH, XAUTHORSEPSTHREESEPSTHREESEPSEACH, YAUTHORSEPSTHREESEPSTHREESEPSEACH, ZAUTHORSEPSTHREESEPSTHREESEPSEACH, QAUTHORSEPSTHREESEPSTHREESEPSLAST, XAUTHORSEPSTHREESEPSTHREESEPSLAST, YAUTHORSEPSTHREESEPSTHREESEPSLAST, ZAUTHORSEPSTHREESEPSTHREESEPSLAST, QAUTHORNAMESNAMEFIRSTNAMEORDER, XAUTHORNAMESNAMEFIRSTNAMEORDER, YAUTHORNAMESNAMEFIRSTNAMEORDER, ZAUTHORNAMESNAMEFIRSTNAMEORDER, QAUTHORNAMESNAMEFIRSTINITIALSTYLE, XAUTHORNAMESNAMEFIRSTINITIALSTYLE, YAUTHORNAMESNAMEFIRSTINITIALSTYLE, ZAUTHORNAMESNAMEFIRSTINITIALSTYLE, QAUTHORNAMESNAMEFIRSTUPPERCASE, XAUTHORNAMESNAMEFIRSTUPPERCASE, YAUTHORNAMESNAMEFIRSTUPPERCASE, ZAUTHORNAMESNAMEFIRSTUPPERCASE, QAUTHORNAMESNAMEOTHERNAMEORDER, XAUTHORNAMESNAMEOTHERNAMEORDER, YAUTHORNAMESNAMEOTHERNAMEORDER, ZAUTHORNAMESNAMEOTHERNAMEORDER, QAUTHORNAMESNAMEOTHERINITIALSTYLE, XAUTHORNAMESNAMEOTHERINITIALSTYLE, YAUTHORNAMESNAMEOTHERINITIALSTYLE, ZAUTHORNAMESNAMEOTHERINITIALSTYLE, QAUTHORNAMESNAMEOTHERUPPERCASE, XAUTHORNAMESNAMEOTHERUPPERCASE, YAUTHORNAMESNAMEOTHERUPPERCASE, ZAUTHORNAMESNAMEOTHERUPPERCASE, QTEXTTEXTSINGLEPRECEEDING, XTEXTTEXTSINGLEPRECEEDING, YTEXTTEXTSINGLEPRECEEDING, ZTEXTTEXTSINGLEPRECEEDING, QTEXTTEXTSINGLEFOLLOWING, XTEXTTEXTSINGLEFOLLOWING, YTEXTTEXTSINGLEFOLLOWING, ZTEXTTEXTSINGLEFOLLOWING, QTEXTTEXTMULTIPLEPRECEEDING, XTEXTTEXTMULTIPLEPRECEEDING, YTEXTTEXTMULTIPLEPRECEEDING, ZTEXTTEXTMULTIPLEPRECEEDING, QTEXTTEXTMULTIPLEFOLLOWING, XTEXTTEXTMULTIPLEFOLLOWING, YTEXTTEXTMULTIPLEFOLLOWING, ZTEXTTEXTMULTIPLEFOLLOWING, PUBDATEPRECEEDING, PUBDATEFOLLOWING, PUBDATEFORMAT, PUBDATESEQUENCE, PUBDATEMONTHFORMAT, PUBDATEDAYFORMAT, PUBDATEYEARFORMAT, PUBDATEPADLEADINGZERO, PUBDATESTYLE, PUBDATESECPRECEEDING, PUBDATESECFOLLOWING, PUBDATESECFORMAT, PUBDATESECSEQUENCE, PUBDATESECMONTHFORMAT, PUBDATESECDAYFORMAT, PUBDATESECYEARFORMAT, PUBDATESECPADLEADINGZERO, PUBDATESECSTYLE, PUBDATEALLPRECEEDING, PUBDATEALLFOLLOWING, PUBDATEALLFORMAT, PUBDATEALLSEQUENCE, PUBDATEALLMONTHFORMAT, PUBDATEALLDAYFORMAT, PUBDATEALLYEARFORMAT, PUBDATEALLPADLEADINGZERO, PUBDATEALLSTYLE, TITLEPRECEEDING, TITLEFOLLOWING, TITLECASE, TITLESTYLE, BOOKTITLEPRECEEDING, BOOKTITLEFOLLOWING, BOOKTITLECASE, BOOKTITLESTYLE, SERIESTITLEPRECEEDING, SERIESTITLEFOLLOWING, SERIESTITLECASE, SERIESTITLESTYLE, ALLTITLEPRECEEDING, ALLTITLEFOLLOWING, ALLTITLECASE, ALLTITLESTYLE, JOURNALNAMEPRECEEDING, JOURNALNAMEFOLLOWING, JOURNALNAMECASE, JOURNALNAMEDEFAULTTEXT, JOURNALNAMEALTERNATETEXT, JOURNALNAMEPUNCTUATION, JOURNALNAMESTYLE, VOLUMEPRECEEDING, VOLUMEFOLLOWING, VOLUMESTYLE, ISSUEPRECEEDING, ISSUEFOLLOWING, ISSUESTYLE, PAGESSTYLE, PAGESSINGLEPAGEPRECEEDING, PAGESSINGLEPAGEFOLLOWING, PAGESPAGERANGEPRECEEDING, PAGESPAGERANGEFOLLOWING, PAGESPAGERANGETYPE, PUBLISHERPRECEEDING, PUBLISHERFOLLOWING, PUBLISHERSTYLE, PUBPLACEPRECEEDING, PUBPLACEFOLLOWING, PUBPLACESTYLE, PAGESPRECEEDING, PAGESFOLLOWING, REFNUMBERPRECEEDING, REFNUMBERFOLLOWING, REFNUMBERSTYLE, PUBDATEFIRSTSEP, PUBDATESECONDSEP, PUBDATESECFIRSTSEP, PUBDATESECSECONDSEP, PUBDATEALLFIRSTSEP, PUBDATEALLSECONDSEP, PAGESPAGERANGERANGESEPARATOR, SERIALPRECEEDING, SERIALFOLLOWING, SERIALSTYLE, ADDRESSPRECEEDING, ADDRESSFOLLOWING, ADDRESSSTYLE, URLPRECEEDING, URLFOLLOWING, URLSTYLE, USERDEF1PRECEEDING, USERDEF1FOLLOWING, USERDEF1STYLE, USERDEF2PRECEEDING, USERDEF2FOLLOWING, USERDEF2STYLE, USERDEF3PRECEEDING, USERDEF3FOLLOWING, USERDEF3STYLE, USERDEF4PRECEEDING, USERDEF4FOLLOWING, USERDEF4STYLE, USERDEF5PRECEEDING, USERDEF5FOLLOWING, USERDEF5STYLE, MISC1PRECEEDING, MISC1FOLLOWING, MISC1STYLE, MISC2PRECEEDING, MISC2FOLLOWING, MISC2STYLE, MISC3PRECEEDING, MISC3FOLLOWING, MISC3STYLE, NOTESPRECEEDING, NOTESFOLLOWING, NOTESSTYLE, ABSTRACTPRECEEDING, ABSTRACTFOLLOWING, ABSTRACTSTYLE, LINK1PRECEEDING, LINK1FOLLOWING, LINK1STYLE, LINK2PRECEEDING, LINK2FOLLOWING, LINK2STYLE, LINK3PRECEEDING, LINK3FOLLOWING, LINK3STYLE, LINK4PRECEEDING, LINK4FOLLOWING, LINK4STYLE from REFSTYLE where PUBTYPE='GEN' and CITSTYLEID="ULLSPEC, (unsigned long long)citstyle_id);
    }

    LOG_PRINT(LOG_DEBUG, sql_command);
    dbi_style_res = dbi_conn_query(conn, sql_command);
    if (!dbi_style_res) {
      LOG_PRINT(LOG_WARNING, "CITSTYLE select failed");
      free(sql_command);
      return NULL;
    }

    if (dbi_result_next_row(dbi_style_res) == 0) {
      LOG_PRINT(LOG_WARNING, "no CITSTYLE found");
      dbi_result_free(dbi_style_res);
      dbi_style_res = NULL;
      free(sql_command);
      return NULL;
    }
  }

  return dbi_style_res;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  unload_style(): frees the memory used by load_style(). If no memory
                  was allocated previously this function will do
		  nothing instead of crash

  void unload_style has no return value

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void unload_style() {
  if (dbi_style_res != NULL) {
    dbi_result_free(dbi_style_res);
    dbi_style_res = NULL;
  }
}












