/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backendn-notex.c defines the notes xnote output backend of refdbd
  markus@mhoenicka.de 2003-10-15
  $Id: backendn-notex.c,v 1.8 2004/01/25 13:20:10 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <time.h>
#include <dbi/dbi.h>

#include "refdb.h"
#include "linklist.h"
#include "refdbd.h"
#include "backend.h"
#include "backendn-notex.h"
#include "strfncs.h"
#include "dbfncs.h"
#include "xmlhelper.h"


/* some globals */
extern struct BSTRING outomem;
extern int n_log_level; /* numeric version of log_level */

/* forward declaration of local functions */
static int indent_notbelow_xnote(const char* name);
static int is_entry_xnote(const char* name);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_note_xnote(): writes a header for the xnote output of a
                         notes query

  char* prepare_render_note_xnote returns a ptr to the result string or
                         NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the note should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* prepare_render_note_xnote(struct renderinfo* ptr_rendinfo) {
  char* new_ref;
  char* header;

  if ((header = assemble_header(ptr_rendinfo)) == NULL) {
    return NULL;
  }

  new_ref = mstrcpy(*(ptr_rendinfo->ptr_ref), header, ptr_rendinfo->ptr_ref_len);
  free(header);

  if (new_ref == NULL) {
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<xnoteset>\n", ptr_rendinfo->ptr_ref_len, 0);

  if (new_ref == NULL) {
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return *(ptr_rendinfo->ptr_ref);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_note_xnote(): writes a footer for the xnote output of
                             a notes query

  char* finish_render_note_xnote returns ptr to the buffer if successful,
                        NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* finish_render_note_xnote(struct renderinfo* ptr_rendinfo) {
  char* new_ref;

  new_ref = mstrcpy(*(ptr_rendinfo->ptr_ref), "</xnoteset>\n", ptr_rendinfo->ptr_ref_len);

  if (new_ref == NULL) {
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return *(ptr_rendinfo->ptr_ref);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_note_xnote() renders a note for xnote output

  char* render_note_xnote returns ptr to the buffer if successful,
                     NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* render_note_xnote(struct renderinfo* ptr_rendinfo) {
  int i;
  char id[32] = "";
  char linktype[4][11] = {"reference", "keyword", "author", "periodical"};
  char date_buffer[256];
  const char* citem;
  const char* owner;
  char* item;
  char* new_ref;
  dbi_conn conn;
  dbi_result dbires;
  struct xmlindent xindent;

  /*initialize xindent */
  xindent.indent_current = 2;
  xindent.maybe_in_leaf = 0;
  xindent.indent_toodeep = 0;
  xindent.notbelow = indent_notbelow_xnote;
  xindent.isentry = is_entry_xnote;
  memset (xindent.indent_string, ' ', INDENT_MAX);
  xindent.indent_string[INDENT_MAX] = '<';
  xindent.indent_string[INDENT_MAX+1] = 0;

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /*----------------------------------------------------------------*/
  /* ID */
  get_refdb_note_id(ptr_rendinfo->dbires, id);

  citem = get_refdb_note_key(ptr_rendinfo->dbires);
  if (citem && !*citem) {
    citem = NULL;
  }
  
  get_refdb_note_date(ptr_rendinfo->dbires, date_buffer, 1);

  owner = get_refdb_note_user_name(ptr_rendinfo->dbires);

  if (!owner && !*owner) {
    owner = ptr_rendinfo->username;
  }

  if (*id) {
    if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "xnote", "id", id, "key", citem, "user", owner, "date", date_buffer, 0, &xindent) == NULL) {
      return NULL;
    }
  }
  else {
    LOG_PRINT(LOG_WARNING, "note has no ID");
    return NULL; /* this is bad and will hopefully never happen */
  }

  /*----------------------------------------------------------------*/
  /* note title */
  item = get_refdb_note_title_copy(ptr_rendinfo->dbires);
  if (item != NULL) {
    if (print_element_x(item, ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "title", NULL, NULL, &xindent) == NULL) {
      free((char*)item);
      return NULL;
    }
    free((char*)item);

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* the content of the note */
  item = get_refdb_note_content_copy(ptr_rendinfo->dbires);
  if (item != NULL) {
    if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "content", "type", get_refdb_note_content_type(ptr_rendinfo->dbires), "xml:lang", get_refdb_note_content_xmllang(ptr_rendinfo->dbires), NULL, NULL, NULL, NULL, 0, &xindent) == NULL) {
      free(item);
      return NULL;
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free(item);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(item);

    if (print_elend_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "content", &xindent) == NULL) {
      return NULL;
    }
  }

  /*----------------------------------------------------------------*/
  /* keywords */
  dbires = request_keywords(conn, my_dbi_result_get_idval(ptr_rendinfo->dbires, "note_id"), 1);
  if (dbires == NULL) {
    return NULL;
  }

  /* fetch all keywords of this article */
  while ((citem = get_keyword(dbires)) != NULL) {

    if (print_element_x(citem, ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "keyword", NULL, NULL, &xindent) == NULL) {
      clean_request(dbires);
      return NULL;
    }
  }
  clean_request(dbires);

  /*----------------------------------------------------------------*/
  /* links */

  /* loop over reference, keyword, author, periodical links */
  for (i = 0; i < 4; i++) {
    int mode = i;
    dbires = request_links(conn, my_dbi_result_get_idval(ptr_rendinfo->dbires, "note_id"), i);
    if (dbires == NULL) {
      return NULL;
    }

    /* fetch all links of this type and note */
    while ((citem = get_link(dbires, &mode)) != NULL) {
      if (i == 3) { /* periodical */
	if (!mode) { /* full name */
	  if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "link", "type", "journalfull", "target", citem, NULL, NULL, NULL, NULL, 1, &xindent) == NULL) {
	    return NULL;
	  }
	}
	else if (mode == 1) { /* abbrev */
	  if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "link", "type", "journalabbrev", "target", citem, NULL, NULL, NULL, NULL, 1, &xindent) == NULL) {
	    return NULL;
	  }
	}
	else if (mode == 2) { /* custabbrev 1 */
	  if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "link", "type", "journalcustabbrev1", "target", citem, NULL, NULL, NULL, NULL, 1, &xindent) == NULL) {
	    return NULL;
	  }
	}
	else if (mode == 3) { /* custabbrev 2 */
	  if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "link", "type", "journalcustabbrev2", "target", citem, NULL, NULL, NULL, NULL, 1, &xindent) == NULL) {
	    return NULL;
	  }
	}
      }
      else {
	if (print_elstart_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "link", "type", linktype[i], "target", citem, NULL, NULL, NULL, NULL, 1, &xindent) == NULL) {
	  return NULL;
	}
      }
    }
    clean_request(dbires);
  } /* end for */


  if (print_elend_x(ptr_rendinfo->ptr_ref, ptr_rendinfo->ptr_ref_len, "xnote", &xindent) == NULL) {
    return NULL;
  }

  /* finish with an empty line */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }
/*      printf("%s\n", ref); */
  return new_ref;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  indent_notbelow_xnote(): checks whether or not to indent below the current
                     element

  static int indent_notbelow returns 1 if not to indent, 0 if to indent

  const char* name ptr to element name

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int indent_notbelow_xnote(const char* name)
{
  if (!strcmp(name, "date")) {
    return 1;
  }
  else {
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_entry_xnote(): checks whether an element starts a new entry

  static int is_entry_risx returns 1 if entry starts, 0 if not

  const char* name ptr to element name

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int is_entry_xnote(const char* name) {
  if (!strcmp(name, "refdbnote")) {
    return 1;
  }
  else {
    return 0;
  }
}
