/*++++++++++++++++++++
dbfncs.c: refdb database-specific support functions
markus@mhoenicka.de 2002-08-06
$Id: dbfncs.c,v 1.15.2.2 2004/11/25 21:34:16 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   +++++++++++++++++++++++++*/

#include <dbi/dbi.h>
#include <syslog.h>

#include "linklist.h"
#include "refdb.h"
#include "refdbd.h"
#include "strfncs.h"
#include "dbfncs.h"
#include "connect.h"

/* global variables */
extern int n_log_level;
extern struct db_caps* ptr_dbcaps;

/* forward declarations of static functions */
static int my_dbi_conn_transaction(dbi_conn conn, int trans_type);
static int set_cap_versioninfo(dbi_conn conn, const char* drivername);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  new_db_caps(): creates and initializes a new db_caps struct

  struct db_caps* new_db_caps returns a ptr to the struct if successful
                             NULL if out of memory

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
struct db_caps* new_db_caps(struct CLIENT_REQUEST* ptr_clrequest) {
  struct db_caps* ptr_caps;

  if ((ptr_caps = (struct db_caps*)malloc(sizeof(struct db_caps))) == NULL) {
    return NULL;
  }

  if (!strcmp(ptr_clrequest->dbserver, "mysql")) {
    strcpy(ptr_caps->multiple_db, "t");
    strcpy(ptr_caps->sql_enum, "t");
    strcpy(ptr_caps->rlike, "RLIKE");
    strcpy(ptr_caps->not_rlike, "NOT RLIKE");
    strcpy(ptr_caps->transaction, "f");
    strcpy(ptr_caps->localhost, "localhost");
    strcpy(ptr_caps->encoding, "f");
    strcpy(ptr_caps->groups, "f");
    strcpy(ptr_caps->admin_systable, "mysql.user");
    strcpy(ptr_caps->listall, ".*");
    strcpy(ptr_caps->bigint, "t");
    strcpy(ptr_caps->sql_union, "f");
    ptr_caps->has_versioninfo = 0;
  }
  else if (!strcmp(ptr_clrequest->dbserver, "pgsql")) {
    strcpy(ptr_caps->multiple_db, "f");
    strcpy(ptr_caps->sql_enum, "f");
    strcpy(ptr_caps->rlike, "~");
    strcpy(ptr_caps->not_rlike, "!~");
    strcpy(ptr_caps->transaction, "t");
    strcpy(ptr_caps->localhost, "");
    strcpy(ptr_caps->encoding, "t");
    strcpy(ptr_caps->groups, "t");
    strcpy(ptr_caps->admin_systable, "pg_shadow");
    strcpy(ptr_caps->listall, ".*");
    strcpy(ptr_caps->bigint, "t");
    strcpy(ptr_caps->sql_union, "t");
    ptr_caps->has_versioninfo = 1;
  }
  else if (!strcmp(ptr_clrequest->dbserver, "sqlite")) {
    strcpy(ptr_caps->multiple_db, "f");
    strcpy(ptr_caps->sql_enum, "f");
    strcpy(ptr_caps->rlike, "LIKE");
    strcpy(ptr_caps->not_rlike, "NOT LIKE");
    strcpy(ptr_caps->transaction, "t");
    strcpy(ptr_caps->localhost, "NA");
    strcpy(ptr_caps->encoding, "f");
    strcpy(ptr_caps->groups, "NA");
    strcpy(ptr_caps->admin_systable, "NA");
    strcpy(ptr_caps->listall, "%");
    strcpy(ptr_caps->bigint, "f");
    strcpy(ptr_caps->sql_union, "t");
    ptr_caps->has_versioninfo = 1;
  }
  else {
    LOG_PRINT(LOG_WARNING, "unknown libdbi driver requested");
    free(ptr_caps);
    return NULL;
  }

  return ptr_caps;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  free_db_caps(): frees the memory of a db_caps struct

  free_db_caps returns nothing

  struct db_caps* ptr_caps ptr to a db_caps structure

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void free_db_caps(struct db_caps* ptr_caps) {
  if (ptr_caps) {
    free(ptr_caps);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  create_tables_mysql(): creates MySQL-specific reference data tables

  int create_tables_mysql returns 1 if error, 0 if successful 

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int create_tables_mysql(dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_result dbires;

  /* create the metadata table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_meta (meta_app VARCHAR(20), meta_type VARCHAR(20), meta_version VARCHAR(20), meta_create_date DATETIME, meta_modify_date DATETIME)");

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_meta failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the main table */
  /* Version issue: later versions support something like:
     refdb_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT 
     without a separate PRIMARY KEY statement */
  /* using a DATE field for refdb_pubyear is not possible as MySQL < 3.23
     does not allow to store dates like 2000-00-00, i.e. when only the
     year is known */
  /* refdb_periodical_id is actually a FOREIGN KEY, but MySQL does not
     support this so a simple INT has to do the trick */

  dbires = dbi_conn_query(conn, "CREATE TABLE t_refdb \
                                   (refdb_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    refdb_citekey VARCHAR(255), \
                                    refdb_type VARCHAR(6), \
                                    refdb_pubyear SMALLINT NOT NULL, \
                                    refdb_secyear SMALLINT NOT NULL, \
                                    refdb_startpage VARCHAR(255), \
                                    refdb_endpage VARCHAR(255), \
                                    refdb_abstract MEDIUMBLOB, \
                                    refdb_title BLOB, \
                                    refdb_volume VARCHAR(255), \
                                    refdb_issue VARCHAR(255), \
                                    refdb_booktitle BLOB, \
                                    refdb_city VARCHAR(255), \
                                    refdb_publisher VARCHAR(255), \
                                    refdb_title_series BLOB, \
                                    refdb_address BLOB, \
                                    refdb_url BLOB, \
                                    refdb_issn VARCHAR(255), \
                                    refdb_pyother_info VARCHAR(255), \
                                    refdb_secother_info VARCHAR(255), \
                                    refdb_periodical_id BIGINT, \
                                    refdb_user1 VARCHAR(255), \
                                    refdb_user2 VARCHAR(255), \
                                    refdb_user3 VARCHAR(255), \
                                    refdb_user4 VARCHAR(255), \
                                    refdb_user5 VARCHAR(255), \
                                    refdb_misc1 VARCHAR(255), \
                                    refdb_misc2 VARCHAR(255), \
                                    refdb_misc3 VARCHAR(255), \
                                    refdb_linkpdf BLOB, \
                                    refdb_linkfull BLOB, \
                                    refdb_linkrel BLOB, \
                                    refdb_linkimg BLOB, \
                                    KEY (refdb_pubyear), \
                                    UNIQUE (refdb_citekey), \
                                    PRIMARY KEY (refdb_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_refdb failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_author \
                                   (author_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    author_name VARCHAR(255) NOT NULL, \
                                    author_lastname VARCHAR(255), \
                                    author_firstname VARCHAR(255), \
                                    author_middlename VARCHAR(255), \
                                    author_suffix VARCHAR(255), \
                                    PRIMARY KEY (author_id), \
                                    KEY (author_name(7)))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_author failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_keyword \
                                   (keyword_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    keyword_name VARCHAR(255) NOT NULL, \
                                    PRIMARY KEY (keyword_id), \
                                    KEY (keyword_name(10)))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_keyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the periodical table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_periodical \
                                   (periodical_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    periodical_name VARCHAR(255) NOT NULL, \
                                    periodical_abbrev VARCHAR(255) NOT NULL, \
                                    periodical_custabbrev1 VARCHAR(255) NOT NULL, \
                                    periodical_custabbrev2 VARCHAR(255) NOT NULL, \
                                    PRIMARY KEY (periodical_id), \
                                    KEY (periodical_name(20)), \
                                    KEY (periodical_abbrev(5)), \
                                    KEY (periodical_custabbrev1(5)), \
                                    KEY (periodical_custabbrev2(5)))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_periodical failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_note \
                                   (note_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    note_key VARCHAR(255), \
                                    note_title VARCHAR(255), \
                                    note_content BLOB, \
                                    note_content_type VARCHAR(255), \
                                    note_content_xmllang VARCHAR(255), \
                                    note_user_id BIGINT, \
                                    note_date DATE, \
                                    KEY (note_title), \
                                    KEY (note_user_id), \
                                    KEY (note_date), \
                                    UNIQUE (note_key), \
                                    PRIMARY KEY (note_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_note failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user table */
  /* the user_name length limit is imposed by MySQL, don't know about PostgreSQL */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_user \
                                   (user_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    user_name VARCHAR(16) NOT NULL, \
                                    PRIMARY KEY (user_id), \
                                    KEY (user_name(8)))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_user failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xauthor \
                                   (xauthor_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    author_id BIGINT NOT NULL, \
                                    refdb_id BIGINT NOT NULL, \
                                    xauthor_type ENUM('primary', 'secondary', 'tertiary'), \
                                    xauthor_role VARCHAR(64), \
                                    xauthor_position INT, \
                                    PRIMARY KEY (xauthor_id), \
                                    KEY (author_id), \
                                    KEY (refdb_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xauthor failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xkeyword \
                                   (xkeyword_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    keyword_id BIGINT NOT NULL, \
                                    xref_id BIGINT NOT NULL, \
                                    xkeyword_type ENUM(\"REFERENCE\",\"NOTE\"), \
                                    PRIMARY KEY (xkeyword_id), \
                                    KEY (keyword_id), \
                                    KEY (xref_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xkeyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xuser \
                                   (xuser_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    user_id BIGINT NOT NULL, \
                                    refdb_id BIGINT NOT NULL, \
                                    xuser_reprint ENUM(\"IN FILE\",\"NOT IN FILE\",\"ON REQUEST\"), \
                                    xuser_date DATE, \
                                    xuser_avail VARCHAR(255), \
                                    xuser_notes MEDIUMBLOB, \
                                    PRIMARY KEY (xuser_id), \
                                    KEY (user_id), \
                                    KEY (refdb_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xuser failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xnote \
                                   (xnote_id BIGINT NOT NULL AUTO_INCREMENT, \
                                    note_id BIGINT NOT NULL, \
                                    xref_id BIGINT NOT NULL, \
                                    xnote_type ENUM(\"REFERENCE\",\"KEYWORD\",\"AUTHOR\",\"PERIODICAL\"), \
                                    PRIMARY KEY (xnote_id), \
                                    KEY (note_id), \
                                    KEY (xref_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xnote failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  create_tables_pgsql(): creates PostgreSQL-specific reference data tables

  int create_tables_pgsql returns 1 if error, 0 if successful 

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int create_tables_pgsql(dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_result dbires;
  char buffer[512];

  /* create the metadata table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_meta (meta_app VARCHAR(20), meta_type VARCHAR(20), meta_version VARCHAR(20), meta_create_date TIMESTAMP, meta_modify_date TIMESTAMP)");

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_meta failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the main table */
  /* using a DATE field for refdb_pubyear is not possible as MySQL < 3.23
     does not allow to store dates like 2000-00-00, i.e. when only the
     year is known */
  /* refdb_periodical_id is actually a FOREIGN KEY, but MySQL does not
     support this so a simple INT has to do the trick */

  /* this will implicitly create a SEQUENCE t_refdb_refdb_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_refdb \
                                   (refdb_id BIGSERIAL, \
                                    refdb_citekey VARCHAR(255) UNIQUE, \
                                    refdb_type VARCHAR(6), \
                                    refdb_pubyear SMALLINT, \
                                    refdb_secyear SMALLINT, \
                                    refdb_startpage VARCHAR(255), \
                                    refdb_endpage VARCHAR(255), \
                                    refdb_abstract TEXT, \
                                    refdb_title TEXT, \
                                    refdb_volume VARCHAR(255), \
                                    refdb_issue VARCHAR(255), \
                                    refdb_booktitle TEXT, \
                                    refdb_city VARCHAR(255), \
                                    refdb_publisher VARCHAR(255), \
                                    refdb_title_series TEXT, \
                                    refdb_address TEXT, \
                                    refdb_url TEXT, \
                                    refdb_issn VARCHAR(255), \
                                    refdb_pyother_info VARCHAR(255), \
                                    refdb_secother_info VARCHAR(255), \
                                    refdb_periodical_id BIGINT, \
                                    refdb_user1 VARCHAR(255), \
                                    refdb_user2 VARCHAR(255), \
                                    refdb_user3 VARCHAR(255), \
                                    refdb_user4 VARCHAR(255), \
                                    refdb_user5 VARCHAR(255), \
                                    refdb_misc1 VARCHAR(255), \
                                    refdb_misc2 VARCHAR(255), \
                                    refdb_misc3 VARCHAR(255), \
                                    refdb_linkpdf TEXT, \
                                    refdb_linkfull TEXT, \
                                    refdb_linkrel TEXT, \
                                    refdb_linkimg TEXT, \
                                    PRIMARY KEY (refdb_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_refdb failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);
  
  /* create the periodical index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_pubyear ON t_refdb (refdb_pubyear)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author table */
  /* this will implicitly create a SEQUENCE t_author_author_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_author \
                                   (author_id BIGSERIAL, \
                                    author_name VARCHAR(255) NOT NULL, \
                                    author_lastname VARCHAR(255), \
                                    author_firstname VARCHAR(255), \
                                    author_middlename VARCHAR(255), \
                                    author_suffix VARCHAR(255), \
                                    PRIMARY KEY (author_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_author failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author_name index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_author_name ON t_author (author_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_author_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword table */
  /* this will implicitly create a SEQUENCE t_keyword_keyword_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_keyword \
                                   (keyword_id BIGSERIAL, \
                                    keyword_name VARCHAR(255) NOT NULL, \
                                    PRIMARY KEY (keyword_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_keyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword_name index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_keyword_name ON t_keyword (keyword_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_keyword_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the periodical table */
  /* this will implicitly create a SEQUENCE t_periodical_periodical_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_periodical \
                                   (periodical_id BIGSERIAL, \
                                    periodical_name VARCHAR(255), \
                                    periodical_abbrev VARCHAR(255), \
                                    periodical_custabbrev1 VARCHAR(255), \
                                    periodical_custabbrev2 VARCHAR(255), \
                                    PRIMARY KEY (periodical_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_periodical failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the t_periodical indices */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_name ON t_periodical (periodical_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_abbrev ON t_periodical (periodical_abbrev)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_abbrev failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_custabbrev1 ON t_periodical (periodical_custabbrev1)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_custabbrev1 failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_custabbrev2 ON t_periodical (periodical_custabbrev2)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_custabbrev2 failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes table */
  /* this will implicitly create a SEQUENCE t_note_note_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_note \
                                   (note_id BIGSERIAL, \
                                    note_key VARCHAR(255) UNIQUE, \
                                    note_title VARCHAR(255), \
                                    note_content TEXT, \
                                    note_content_type VARCHAR(255), \
                                    note_content_xmllang VARCHAR(255), \
                                    note_user_id BIGINT, \
                                    note_date DATE, \
                                    PRIMARY KEY (note_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_note failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the t_note index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_title ON t_note (note_title)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_title failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_note_user_id ON t_note (note_user_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_user_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_date ON t_note (note_date)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_date failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user table */
  /* the user_name length limit is imposed by MySQL, don't know about PostgreSQL */
  /* this will implicitly create a SEQUENCE t_user_user_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_user \
                                   (user_id BIGSERIAL, \
                                    user_name VARCHAR(16) NOT NULL, \
                                    PRIMARY KEY (user_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_user failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create user_name index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_user_name ON t_user (user_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_user_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author xtable */
  /* this will implicitly create a SEQUENCE t_xauthor_xauthor_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xauthor \
                                   (xauthor_id BIGSERIAL, \
                                    author_id BIGINT NOT NULL, \
                                    refdb_id BIGINT NOT NULL, \
                                    xauthor_type SMALLINT DEFAULT 1, \
                                    xauthor_role VARCHAR(64), \
                                    xauthor_position INTEGER, \
                                    PRIMARY KEY (xauthor_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xauthor failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create x_author indices */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xauthor_author_id ON t_xauthor (author_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xauthor_author_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xauthor_refdb_id ON t_xauthor (refdb_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xauthor_refdb_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword xtable */
  /* this will implicitly create a SEQUENCE t_xkeyword_xkeyword_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xkeyword \
                                   (xkeyword_id BIGSERIAL, \
                                    keyword_id BIGINT NOT NULL, \
                                    xkeyword_type VARCHAR(11) DEFAULT \'REFERENCE\', \
                                    xref_id BIGINT NOT NULL, \
                                    PRIMARY KEY (xkeyword_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xkeyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create x_keyword indices */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xkeyword_keyword_id ON t_xkeyword (keyword_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xkeyword_keyword_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xkeyword_xref_id ON t_xkeyword (xref_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xkeyword_xref_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user xtable */
  /* this will implicitly create a SEQUENCE t_xuser_xuser_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xuser \
                                   (xuser_id BIGSERIAL, \
                                    user_id BIGINT NOT NULL, \
                                    refdb_id BIGINT NOT NULL, \
                                    xuser_reprint VARCHAR(11) DEFAULT \'NOT IN FILE\', \
                                    xuser_date DATE, \
                                    xuser_avail VARCHAR(255), \
                                    xuser_notes TEXT, \
                                    PRIMARY KEY (xuser_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xuser failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create x_user indices */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xuser_user_id ON t_xuser (user_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xuser_user_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xuser_refdb_id ON t_xuser (refdb_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xuser_refdb_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes xtable */
  /* this will implicitly create a SEQUENCE t_xnote_xnote_id_seq */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xnote \
                                   (xnote_id BIGSERIAL, \
                                    note_id BIGINT NOT NULL, \
                                    xref_id BIGINT NOT NULL, \
                                    xnote_type VARCHAR(11) DEFAULT \'REFERENCE\', \
                                    PRIMARY KEY (xnote_id))");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xnote failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create x_note indices */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xnote_note_id ON t_xnote (note_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xnote_note_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xnote_xref_id ON t_xnote (xref_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xnote_xref_id failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xnote_xnote_type ON t_xnote (xnote_type)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xnote_xnote_type failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create user permissions */
  /* PostgreSQL uses table-based permissions, so it is easier to create
     them here than in adduser. We grant permissions to a group "<dbname>user"
     here, whereas adduser just adds users to/removes from this group */

  /* ptr_clrequest->argument contains the database name */
  if (is_group_pgsql(conn, ptr_clrequest->argument) != 1) {
    sprintf(buffer, "CREATE GROUP %suser", ptr_clrequest->argument);
    LOG_PRINT(LOG_DEBUG, buffer);
    
    dbires = dbi_conn_query(conn, buffer);

    if (!dbires) {
      tiwrite(ptr_clrequest->fd, "CREATE GROUP failed\n", TERM_NO);
      return 1;
    }
  }

  sprintf(buffer, "GRANT SELECT, INSERT, UPDATE, DELETE ON t_meta, t_refdb, t_author, t_keyword, t_periodical, t_user, t_note, t_xauthor, t_xkeyword, t_xuser, t_xnote, t_refdb_refdb_id_seq, t_author_author_id_seq, t_keyword_keyword_id_seq, t_periodical_periodical_id_seq, t_user_user_id_seq, t_note_note_id_seq, t_xauthor_xauthor_id_seq, t_xkeyword_xkeyword_id_seq, t_xuser_xuser_id_seq, t_xnote_xnote_id_seq TO GROUP %suser", ptr_clrequest->argument);
  LOG_PRINT(LOG_DEBUG, buffer);

  dbires = dbi_conn_query(conn, buffer);

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "GRANT failed for new tables\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  create_tables_sqlite(): creates SQLite-specific reference data tables

  int create_tables_sqlite returns 1 if error, 0 if successful 

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int create_tables_sqlite(dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_result dbires;

  /* create the metadata table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_meta (meta_app VARCHAR(20), meta_type VARCHAR(20), meta_version VARCHAR(20), meta_create_date DATETIME, meta_modify_date DATETIME)");

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_meta failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the main table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_refdb \
                                   (refdb_id INTEGER PRIMARY KEY, \
                                    refdb_citekey TEXT UNIQUE, \
                                    refdb_type TEXT, \
                                    refdb_pubyear SMALLINT, \
                                    refdb_secyear SMALLINT, \
                                    refdb_startpage TEXT, \
                                    refdb_endpage TEXT, \
                                    refdb_abstract TEXT, \
                                    refdb_title TEXT, \
                                    refdb_volume TEXT, \
                                    refdb_issue TEXT, \
                                    refdb_booktitle TEXT, \
                                    refdb_city TEXT, \
                                    refdb_publisher TEXT, \
                                    refdb_title_series TEXT, \
                                    refdb_address TEXT, \
                                    refdb_url TEXT, \
                                    refdb_issn TEXT, \
                                    refdb_pyother_info TEXT, \
                                    refdb_secother_info TEXT, \
                                    refdb_periodical_id INTEGER, \
                                    refdb_user1 TEXT, \
                                    refdb_user2 TEXT, \
                                    refdb_user3 TEXT, \
                                    refdb_user4 TEXT, \
                                    refdb_user5 TEXT, \
                                    refdb_misc1 TEXT, \
                                    refdb_misc2 TEXT, \
                                    refdb_misc3 TEXT, \
                                    refdb_linkpdf TEXT, \
                                    refdb_linkfull TEXT, \
                                    refdb_linkrel TEXT, \
                                    refdb_linkimg TEXT)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_refdb failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the main table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_refdb_pubyear ON t_refdb (refdb_pubyear)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_refdb_pubyear failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE UNIQUE INDEX i_refdb_citekey ON t_refdb (refdb_citekey)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_refdb_citekey failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_author \
                                   (author_id INTEGER PRIMARY KEY, \
                                    author_name TEXT NOT NULL, \
                                    author_lastname TEXT, \
                                    author_firstname TEXT, \
                                    author_middlename TEXT, \
                                    author_suffix TEXT)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_author failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_author table index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_author_name ON t_author (author_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_author_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_keyword \
                                   (keyword_id INTEGER PRIMARY KEY, \
                                    keyword_name TEXT NOT NULL)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_keyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_keyword table index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_keyword_name ON t_keyword (keyword_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_keyword_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the periodical table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_periodical \
                                   (periodical_id INTEGER PRIMARY KEY, \
                                    periodical_name TEXT, \
                                    periodical_abbrev TEXT, \
                                    periodical_custabbrev1 TEXT, \
                                    periodical_custabbrev2 TEXT)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_periodical failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_periodical table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_name ON t_periodical (periodical_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_abbrev ON t_periodical (periodical_abbrev)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_abbrev failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_custabbrev1 ON t_periodical (periodical_custabbrev1)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_custabbrev failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_periodical_custabbrev2 ON t_periodical (periodical_custabbrev2)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_periodical_custabbrev failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_note \
                                   (note_id INTEGER PRIMARY KEY, \
                                    note_key TEXT UNIQUE, \
                                    note_title TEXT, \
                                    note_content_type TEXT, \
                                    note_content_xmllang TEXT, \
                                    note_user_id INTEGER, \
                                    note_date DATE, \
                                    note_content TEXT)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_note failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_note table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_title ON t_note (note_title)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_title failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_note_userid ON t_note (note_user_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_note_userid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_note_date ON t_note (note_date)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_note_date failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user table */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_user \
                                   (user_id INTEGER PRIMARY KEY, \
                                    user_name TEXT NOT NULL)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_user failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_user table index */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_user_name ON t_user (user_name)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_user_name failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the author xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xauthor \
                                   (xauthor_id INTEGER PRIMARY KEY, \
                                    author_id INTEGER NOT NULL, \
                                    refdb_id INTEGER NOT NULL, \
                                    xauthor_type SMALLINT DEFAULT 1, \
                                    xauthor_role TEXT, \
                                    xauthor_position INTEGER)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xauthor failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_xauthor table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xauthor_autid ON t_xauthor (author_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xauthor_autid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xauthor_refid ON t_xauthor (refdb_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xauthor_refid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the keyword xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xkeyword \
                                   (xkeyword_id INTEGER PRIMARY KEY, \
                                    xkeyword_type TEXT, \
                                    keyword_id INTEGER NOT NULL, \
                                    xref_id INTEGER NOT NULL)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xkeyword failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_xkeyword table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xkeyword_kwid ON t_xkeyword (keyword_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xkeyword_kwid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xkeyword_xrefid ON t_xkeyword (xref_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xkeyword_xrefid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the user xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xuser \
                                   (xuser_id INTEGER PRIMARY KEY, \
                                    user_id INTEGER NOT NULL, \
                                    refdb_id INTEGER NOT NULL, \
                                    xuser_reprint TEXT DEFAULT \'NOT IN FILE\', \
                                    xuser_date DATE, \
                                    xuser_avail TEXT, \
                                    xuser_notes TEXT)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xuser failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_xuser table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xuser_userid ON t_xuser (user_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xuser_userid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xuser_refid ON t_xuser (refdb_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xuser_refid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create the notes xtable */
  dbires = dbi_conn_query(conn, "CREATE TABLE t_xnote \
                                   (xnote_id INTEGER PRIMARY KEY, \
                                    note_id INTEGER NOT NULL, \
                                    xref_id INTEGER NOT NULL, \
                                    xnote_type TEXT DEFAULT \'REFERENCE\')");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create t_xnote failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  /* create t_xnote table indexes */
  dbires = dbi_conn_query(conn, "CREATE INDEX i_xnote_noteid ON t_xnote (note_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xnote_noteid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "CREATE INDEX i_xnote_xrefid ON t_xnote (xref_id)");


  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "create i_xnote_xrefid failed\n", TERM_NO);
    return 1;
  }
   
  dbi_result_free(dbires);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_user_pgsql(): checks whether a user exists

  int is_user_pgsql returns 1 if the user exists, 0 if the user doesn't
                   exist, -1 if an error occurs 

  dbi_conn conn connection to the database to query

  const char* username ptr to string with the name of the user

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int is_user_pgsql(dbi_conn conn, const char* username) {
  char sql_command[256];
  dbi_result dbires;

  /* The PostgreSQL system table pg_user contains a list of users */
  sprintf(sql_command, "SELECT usename FROM pg_user WHERE usename=\'%s\'", username);
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);


  if (!dbires) {
    return -1;
  }
   
  if (dbi_result_get_numrows(dbires)) {
    dbi_result_free(dbires);
    return 1;
  }
  else {
    dbi_result_free(dbires);
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_group_pgsql(): checks whether a group exists

  int is_group_pgsql returns 1 if the group exists, 0 if the group doesn't
                   exist, -1 if an error occurs 

  dbi_conn conn connection to the database to query

  const char* dbname ptr to string with the name of the database
                     related to the group

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int is_group_pgsql(dbi_conn conn, const char* dbname) {
  char sql_command[256];
  dbi_result dbires;

  /* The PostgreSQL system table pg_group contains a list of groups */
  sprintf(sql_command, "SELECT groname FROM pg_group WHERE groname=\'%suser\'", dbname);
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);


  if (!dbires) {
    return -1;
  }
   
  if (dbi_result_get_numrows(dbires)) {
    dbi_result_free(dbires);
    return 1;
  }
  else {
    dbi_result_free(dbires);
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_transaction(): deals with starting/ending transactions

  static int my_dbi_conn_transaction returns 0 if ok, 1 if error

  dbi_conn conn database connection

  int trans_type 0 is "BEGIN", 1 is "COMMIT", 2 is "ROLLBACK"

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int my_dbi_conn_transaction(dbi_conn conn, int trans_type) {
  const char *drivername;
  char sql_command[2][3][14] = {
    {"BEGIN WORK", "COMMIT WORK", "ROLLBACK WORK"},
    {"BEGIN", "COMMIT", "ROLLBACK"}
  };

  dbi_result dbires;

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  if (!strcmp(drivername, "pgsql")) {
    dbires = dbi_conn_query(conn, sql_command[0][trans_type]);
    LOG_PRINT(LOG_DEBUG, sql_command[0][trans_type]);
    if (!dbires) {
      return 1;
    }
    dbi_result_free(dbires);
  }
  else if (!strcmp(drivername, "sqlite")) {
    dbires = dbi_conn_query(conn, sql_command[1][trans_type]);
    LOG_PRINT(LOG_DEBUG, sql_command[1][trans_type]);
    if (!dbires) {
      return 1;
    }
    dbi_result_free(dbires);
  }
  /* else: transactions not supported by other drivers */
  /* ToDo: MySQL > 4.0 knows transactions, but how do we know the server
     version? And the table type? */
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_begin(): begin a transaction

  int my_dbi_conn_begin returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_begin(dbi_conn conn) {
  return my_dbi_conn_transaction(conn, 0);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_commit(): commits a transaction

  int my_dbi_conn_commit returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_commit(dbi_conn conn) {
  return my_dbi_conn_transaction(conn, 1);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_rollback(): ends a transaction and discards all changes

  int my_dbi_conn_rollback returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_rollback(dbi_conn conn) {
  return my_dbi_conn_transaction(conn, 2);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_lock(): locks reference database tables

  int my_dbi_conn_lock returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_lock(dbi_conn conn) {
  const char *drivername;
  dbi_result dbires;

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  if (!strcmp(drivername, "mysql")) {
    char* sql_command;

    if ((sql_command = (char*)malloc(256)) == NULL) {
      return 1;
    }
    sprintf(sql_command, "LOCK TABLES t_refdb WRITE, t_author WRITE, t_keyword WRITE, t_periodical WRITE, t_user WRITE, t_xauthor WRITE, t_xkeyword WRITE, t_xuser WRITE, t_note WRITE, t_xnote WRITE, %s.t_journal_words WRITE", MAIN_DB);
    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires = dbi_conn_query(conn, sql_command);
    free(sql_command);
    if (!dbires) {
      return 1;
    }
    dbi_result_free(dbires);
  }
  else if (!strcmp(drivername, "pgsql")) {
    int i;
    char sql_command[64];
    const char tables[10][13] = {
      "t_refdb",
      "t_author",
      "t_keyword",
      "t_periodical",
      "t_user",
      "t_xauthor",
      "t_xkeyword",
      "t_xuser",
      "t_note",
      "t_xnote"
    };

    for (i = 0; i < 10; i++) {
      sprintf(sql_command, "LOCK TABLE %s IN SHARE MODE", tables[i]);
      dbires = dbi_conn_query(conn, sql_command);
      LOG_PRINT(LOG_DEBUG, sql_command);
      if (!dbires) {
	return 1;
      }
      dbi_result_free(dbires);
    }
  }
  /* else: sqlite does not support table locking */

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_lock_note(): locks notes database tables

  int my_dbi_conn_lock_note returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_lock_note(dbi_conn conn) {
  const char *drivername;
  dbi_result dbires;

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  if (!strcmp(drivername, "mysql")) {
    char* sql_command;

    if ((sql_command = (char*)malloc(256)) == NULL) {
      return 1;
    }
    strcpy(sql_command, "LOCK TABLES t_keyword WRITE, t_user WRITE, t_xkeyword WRITE, t_xuser WRITE, t_note WRITE, t_xnote WRITE, t_periodical WRITE, t_author WRITE, t_refdb WRITE");
    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires = dbi_conn_query(conn, sql_command);
    free(sql_command);
    if (!dbires) {
      return 1;
    }
    dbi_result_free(dbires);
  }
  else if (!strcmp(drivername, "pgsql")) {
    int i;
    char sql_command[64];
    const char tables[6][13] = {
      "t_keyword",
      "t_user",
      "t_xkeyword",
      "t_xuser",
      "t_note",
      "t_xnote"
    };

    for (i = 0; i < 6; i++) {
      sprintf(sql_command, "LOCK TABLE %s IN SHARE MODE", tables[i]);
      dbires = dbi_conn_query(conn, sql_command);
      LOG_PRINT(LOG_DEBUG, sql_command);
      if (!dbires) {
	return 1;
      }
      dbi_result_free(dbires);
    }
  }
  /* else: sqlite does not support table locking */

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_unlock(): unlocks locked tables

  int my_dbi_conn_unlock returns 0 if ok, 1 if error

  dbi_conn conn database connection

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_unlock(dbi_conn conn) {
  dbi_result dbires;
  const char *drivername;

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  if (!strcmp(drivername, "mysql")) {
    dbires = dbi_conn_query(conn, "UNLOCK TABLES");
    LOG_PRINT(LOG_DEBUG, "UNLOCK TABLES");
    if (!dbires) {
      return 1;
    }
    dbi_result_free(dbires);
  }
  /* else: pgsql unlocks when a transaction is finished, */
  /*       sqlite does not support lock/unlock */

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_get_cap(): returns a driver capability

  const char* my_dbi_conn_get_cap returns a capability string or NULL on error

  dbi_conn conn database connection

  const char* cap string containing the name of the requested
                  capability

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* my_dbi_conn_get_cap(dbi_conn conn, const char* cap) {
  dbi_driver driver;
  const char *drivername;

  driver = dbi_conn_get_driver(conn);

  drivername = dbi_driver_get_name(driver);

  /* check whether we already have the version-specific information */
  if (!ptr_dbcaps->has_versioninfo) {
    set_cap_versioninfo(conn, drivername);
  }

  return my_dbi_driver_get_cap(driver, cap);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_driver_get_cap(): returns a driver capability

  const char* my_dbi_driver_get_cap returns a capability string or NULL on error

  dbi_driver database driver

  const char* cap string containing the name of the requested
                  capability

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char* my_dbi_driver_get_cap(dbi_driver driver, const char* cap) {
  if (!strcmp(cap, "multiple_db")) {
    return ptr_dbcaps->multiple_db;
  }
  else if (!strcmp(cap, "enum")) {
    return ptr_dbcaps->sql_enum;
  }
  else if (!strcmp(cap, "rlike")) {
    return ptr_dbcaps->rlike;
  }
  else if (!strcmp(cap, "not_rlike")) {
    return ptr_dbcaps->not_rlike;
  }
  else if (!strcmp(cap, "transaction")) {
    return ptr_dbcaps->transaction;
  }
  else if (!strcmp(cap, "localhost")) {
    return ptr_dbcaps->localhost;
  }
  else if (!strcmp(cap, "encoding")) {
    return ptr_dbcaps->encoding;
  }
  else if (!strcmp(cap, "groups")) {
    return ptr_dbcaps->groups;
  }
  else if (!strcmp(cap, "admin_systable")) {
    return ptr_dbcaps->admin_systable;
  }
  else if (!strcmp(cap, "listall")) {
    return ptr_dbcaps->listall;
  }
  else if (!strcmp(cap, "bigint")) {
    return ptr_dbcaps->bigint;
  }
  else if (!strcmp(cap, "union")) {
    return ptr_dbcaps->sql_union;
  }
  /* else: should never happen */

  /* unknown cap */
  return NULL;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_idval(): returns an id value by field name. Takes
                             care of the fact that some db servers do
			     not support ulonglongs

  unsigned long long my_dbi_result_get_idval returns an id value 

  dbi_result dbires result of a previous query

  const char* fieldname name of the field to retrieve

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
unsigned long long my_dbi_result_get_idval(dbi_result dbires, const char* fieldname) {
  dbi_conn conn;
  char buffer[512];
  unsigned long the_ulong;
  unsigned long long the_ulonglong;

  conn = dbi_result_get_conn(dbires);

  if (!strcmp(my_dbi_conn_get_cap(conn, "bigint"), "f")) {
    /* database server does not support long long ints, retrieve as 
       long int instead and cast */
    the_ulong = dbi_result_get_ulong(dbires, fieldname);
/*     sprintf(buffer, "ulong: u %u<< lu %lu<< llu %llu", the_ulong, the_ulong, the_ulong); */
/*     LOG_PRINT(LOG_DEBUG, buffer); */
    return (unsigned long long)the_ulong;
  }
  else {
    the_ulonglong = dbi_result_get_ulonglong(dbires, fieldname);
/*     sprintf(buffer, "ulonglong: u %u<< lu %lu<< llu %llu", the_ulonglong, the_ulonglong, the_ulonglong); */
/*     LOG_PRINT(LOG_DEBUG, buffer); */
    return the_ulonglong;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_result_get_idval_idx(): returns an id value by index. Takes
                                 care of the fact that some db servers
				 do not support ulonglongs

  unsigned long long my_dbi_result_get_idval_idx returns an id value 

  dbi_result dbires result of a previous query

  unsigned long long idx index of field to query

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
unsigned long long my_dbi_result_get_idval_idx(dbi_result dbires, unsigned long long idx) {
  dbi_conn conn;

  conn = dbi_result_get_conn(dbires);

  if (!strcmp(my_dbi_conn_get_cap(conn, "bigint"), "f")) {
    /* database server does not support long long ints, retrieve as 
       long int instead and cast */
    return (unsigned long long)dbi_result_get_ulong_idx(dbires, idx);
  }
  else {
    return dbi_result_get_ulonglong_idx(dbires, idx);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  my_dbi_conn_get_versioninfo(): retrieves the version info of a db engine

  int my_dbi_conn_get_versioninfo returns 0 if ok, 1 if error

  dbi_conn conn connection to a database engine

  struct VERSIONINFO* ptr_ver ptr to  struct to receive version info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int my_dbi_conn_get_versioninfo(dbi_conn conn, struct VERSIONINFO* ptr_ver) {
  char* versioninfo = NULL;
  char sql_command[] = "SELECT VERSION()";
  dbi_result dbires;

  ptr_ver->major = 0;
  ptr_ver->minor = 0;
  ptr_ver->minuscule = 0;

  dbires = dbi_conn_query(conn, sql_command);
  LOG_PRINT(LOG_DEBUG, sql_command);
  if (dbires) {
    if (dbi_result_next_row(dbires)) {
      versioninfo = dbi_result_get_string_copy_idx(dbires, 1); /* 1-base index */
    }
    dbi_result_free(dbires);
  }

  if (versioninfo) {
    if (parse_versioninfo(versioninfo, ptr_ver)) {
      return 1;
    }
    free(versioninfo);
    /* 	printf("major: %d minor: %d minuscule: %d\n", ver.major, ver.minor, ver.minuscule); */
  }

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  set_cap_versioninfo(): sets version-specific info in dbcaps structure

  static int set_cap_versioninfo returns 0 if ok, 1 if error

  dbi_conn conn connection to a database engine

  const char* drivername name of db driver

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int set_cap_versioninfo(dbi_conn conn, const char* drivername) {
  struct VERSIONINFO ver;

  /* try only once */
  ptr_dbcaps->has_versioninfo = 1;

  if (my_dbi_conn_get_versioninfo(conn, &ver)) {
    return 1;
  }

  if (!strcmp(drivername, "mysql")) {
    if (ver.major >= 4) {
      strcpy(ptr_dbcaps->sql_union, "t");
    }
  }
  /* else: nothing to do */

  return 0;
}
