/*++++++++++++++++++++
refdbda.c: refdb application server, administrative functions
markus@mhoenicka.de 6-5-00
$Id: refdbda.c,v 1.41.2.2 2004/10/12 00:18:46 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   +++++++++++++++++++++++++*/


#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include <time.h> /* for strftime */
#include <sys/types.h>
#include <sys/socket.h>
#include <dbi/dbi.h>

#include "backend.h"
#include "refdb.h"
#include "linklist.h"
#include "refdbd.h" /* depends on backend.h */
#include "tokenize.h"
#include "strfncs.h"
#include "connect.h"
#include "cgi.h"
#include "dbfncs.h"

/* these externs are all defined in refdbd.c */
extern struct BSTRING outomem;
extern struct BSTRING outomem_n;
extern struct BSTRING connerr;
extern struct BSTRING connerr_n;
extern struct BSTRING positive;
extern struct BSTRING negative;
extern struct BSTRING cgihead_html;
extern struct BSTRING cgihead_plain;
extern char server_ip[];
extern char log_file[];
extern char log_dest[];
extern char log_level[];
extern char port_address[];
extern char dbs_port_address[];
extern char refdb_timeout[];
extern char refdblib[];
extern char the_fifo[];
extern char pid_file[];
extern int n_log_level;
extern int n_log_dest;
extern int n_refdb_timeout;
extern int n_reopen_log;
extern int n_remote_admin;
extern int run;
extern int parent_pid;
extern int fd_fifo;

/* forward declarations of local functions */
static int is_reference_database(struct CLIENT_REQUEST* ptr_clrequest, dbi_conn conn, const char* dbname);
static int daemon_started_by_init(void);
static int daemon_started_by_inetd(void);
static long limit_open(void);

/* additional types */
#ifndef socklen_t
#  define socklen_t int
#endif


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  child_confserv(): configures the application server (child process)
                    this fn sends the changes to the parent process
                    via a FIFO/named pipe (via a temp file in CygWin).

  void child_confserv

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void child_confserv(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_result dbires1;
  char outbuffer[OUTBUF_LEN];
  char access_list[2][PRIV_LENGTH]; /* used for querying access rights */
  char* sql_command;
  char* new_sql_command;
  const char* item;
  int fd_fifo;
  size_t sql_command_len;
  int error;

  if (!n_remote_admin) {
    tiwrite(ptr_clrequest->fd, "remote administration disabled\n", TERM_NO);
    return;
  }

  sql_command_len = 1024;
  sql_command = (char*)malloc(sql_command_len);
  if (sql_command == NULL) {
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    return;
  }

  /* check our access rights */
  if (!is_refdb_admin(ptr_clrequest)) {
    free(sql_command); /* unauthorized attempt likely */
    return;
  }

  /* we have to treat ping and cleanvar separately as they should not
     create a temp file or use the FIFO */
  if (ptr_clrequest->inargc == *(ptr_clrequest->ptr_optind)+1 && strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "ping") == 0) {
    sprintf(outbuffer, "parent PID: %d  child PID: %d\n", parent_pid, getpid());
    tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);
    free(sql_command);
    return;
  }
  else if (ptr_clrequest->inargc == *(ptr_clrequest->ptr_optind)+1 && strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "cleanvar") == 0) {
    /* connect to the database */
    if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {

      /* get a list of the tables in the database */
      dbires = dbi_conn_get_table_list(conn, "refdbvar", "refdb%");
      if (dbires != NULL) {
	if (dbi_result_get_numrows(dbires)) { /* if there are temp tables */
	  strcpy(sql_command, "DROP TABLE ");

	  /* loop over all table names */
	  while (dbi_result_next_row(dbires)) {
	    item = dbi_result_get_string_idx(dbires, 1); /* 1-base index */
	    new_sql_command = mstrcat(sql_command, (char*)item, &sql_command_len, 0);
	    if (new_sql_command == NULL) {
	      dbi_result_free(dbires);
	      dbi_conn_close(conn);
	      iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
	      return;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	    new_sql_command = mstrcat(sql_command, ",", &sql_command_len, 0);
	    if (new_sql_command == NULL) {
	      dbi_result_free(dbires);
	      dbi_conn_close(conn);
	      iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
	      return;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	  } /* end while */

	  sql_command[strlen(sql_command)-1] = '\0'; /* remove trailing comma */

	  /* send query */
	  dbires1 = dbi_conn_query(conn, sql_command);
	  LOG_PRINT(LOG_DEBUG, sql_command);
	  if (!dbires1) {
	    tiwrite(ptr_clrequest->fd, "error while deleting temp tables\n", TERM_NO);
	    LOG_PRINT(LOG_WARNING, "error while deleting temp tables");
	  }
	  else {
	    tiwrite(ptr_clrequest->fd, "successfully deleted temp tables\n", TERM_NO);
	    LOG_PRINT(LOG_INFO, "temp tables deleted");
	    dbi_result_free(dbires1);
	  }
	}
	else { /* no temp tables in database */
	  tiwrite(ptr_clrequest->fd, "nothing to do\n", TERM_NO);
	}
	dbi_result_free(dbires);
      }
      dbi_conn_close(conn);
    }
    free(sql_command);
    return;
  }

#ifdef HAVE_MKFIFO
  if (fd_fifo != -1) {
    close(fd_fifo); /* this is still open for reading */
    if ((fd_fifo = open(the_fifo, O_WRONLY)) == -1) {
      LOG_PRINT(LOG_WARNING, "cannot open FIFO for writing");
      tiwrite(ptr_clrequest->fd, "confserv: remote configure not possible\n", TERM_NO);
      return;
    }
  }
  else {
    tiwrite(ptr_clrequest->fd, "confserv: remote configure not possible\n", TERM_NO);
    return;
  }
#else
  if (access(the_fifo, F_OK) != -1) { /* old file still exists */
    tiwrite(ptr_clrequest->fd, "confserv: appserver busy, try again later\n", TERM_NO);
    return;
  }
  else { /* create a temp file with exclusive rights */
    if ((fd_fifo = open(the_fifo, O_WRONLY|O_CREAT|O_EXCL, S_IRUSR|S_IWUSR)) == -1) {
      LOG_PRINT(LOG_WARNING, "cannot open FIFO for writing");
      tiwrite(ptr_clrequest->fd, "confserv: remote configure not possible\n", TERM_NO);
      return;
    }
  }
#endif /* HAVE_MKFIFO */
  
  /* the inargv array should have one or two entries: a command and
     an optional argument */
  if (ptr_clrequest->inargc == *(ptr_clrequest->ptr_optind)) { /* no argument given */
    tiwrite(ptr_clrequest->fd, "confserv: wrong or incomplete command\n", TERM_NO);
  }

  if (ptr_clrequest->inargc == *(ptr_clrequest->ptr_optind)+1) { /* one argument */
    if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "stop") == 0) {
      if (write(fd_fifo, "1stop", 6) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	tiwrite(ptr_clrequest->fd, "application server stop submitted\n", TERM_NO);
	LOG_PRINT(LOG_INFO, "application server stop submitted");
      }
    }
    else {
      tiwrite(ptr_clrequest->fd, "confserv: wrong or incomplete command\n", TERM_NO);
    }
  }
  else { /* more than one argument, use first two */
    if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "serverip") == 0) {
      strcpy(ptr_clrequest->server_ip, ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)+1]);
      sprintf(outbuffer, "set serverip to %s submitted\n", ptr_clrequest->server_ip);
      tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);

      sprintf(outbuffer, "2serverip %s", ptr_clrequest->server_ip);
      if (write(fd_fifo, outbuffer, strlen(outbuffer) + 1) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	LOG_PRINT(LOG_INFO, "set serverip submitted");
      }
    }
    else if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "timeout") == 0) {
      strcpy(refdb_timeout, ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)+1]);
      n_refdb_timeout = atoi(refdb_timeout);
      sprintf(outbuffer, "set timeout to %d seconds submitted\n", n_refdb_timeout);
      tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);

      sprintf(outbuffer, "2timeout %s", refdb_timeout);
      if (write(fd_fifo, outbuffer, strlen(outbuffer) + 1) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	LOG_PRINT(LOG_INFO, "set timeout submitted");
      }
    }
    else if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "logfile") == 0) {
      strcpy(log_file, ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)+1]);
      sprintf(outbuffer, "set logfile to %s submitted\n", log_file);
      tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);

      sprintf(outbuffer, "2logfile %s", log_file);
      if (write(fd_fifo, outbuffer, strlen(outbuffer) + 1) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	LOG_PRINT(LOG_INFO, "set logfile submitted");
      }
    }
    else if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "logdest") == 0) {
      strcpy(log_dest, ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)+1]);
      n_log_dest = atoi(log_dest);
      sprintf(outbuffer, "set logdest to %d\n", n_log_dest);
      tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);

      sprintf(outbuffer, "2logdest %s", log_dest);
      if (write(fd_fifo, outbuffer, strlen(outbuffer) + 1) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	LOG_PRINT(LOG_INFO, "set logdest submitted");
      }
    }
    else if (strcmp(ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)], "loglevel") == 0) {
      strcpy(log_level, ptr_clrequest->inargv[*(ptr_clrequest->ptr_optind)+1]);
      n_log_level = atoi(log_level);
      sprintf(outbuffer, "set loglevel to %d submitted\n", n_log_level);
      tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);

      sprintf(outbuffer, "2loglevel %s", log_level);
      if (write(fd_fifo, outbuffer, strlen(outbuffer) + 1) == -1) {
	tiwrite(ptr_clrequest->fd, "write error in FIFO\n", TERM_NO);
	LOG_PRINT(LOG_WARNING, "write error in FIFO");
      }
      else {
	LOG_PRINT(LOG_INFO, "set loglevel submitted");
      }
    }
    else {
      tiwrite(ptr_clrequest->fd, "confserv: wrong or incomplete command\n", TERM_NO);
    }
  }
  if (close(fd_fifo) != 0) {
    LOG_PRINT(LOG_WARNING, "write error in FIFO, detected when closing");
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  confserv(): configures the application server (parent process)

  void confserv

  char* command ptr to string with the command

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void confserv(char *command, char* server_ip) {
  char outbuffer[OUTBUF_LEN];

  /* the command buffer should have one or two entries: a command and
     an optional argument */
  if (*command == '1') { /* one argument */
    if (strncmp(command + 1, "stop", 4) == 0) {
      run = 0;
      LOG_PRINT(LOG_INFO, "application server stopped");
    }
  }
  else { /* more than one argument, use first two */
    if (strncmp(command + 1, "serverip", 8) == 0) {
      strcpy(server_ip, command + 10);
      sprintf(outbuffer, "set serverip to %s", server_ip);
      LOG_PRINT(LOG_INFO, outbuffer);
    }
    else if (strncmp(command + 1, "timeout", 7) == 0) {
      strcpy(refdb_timeout, command + 9);
      n_refdb_timeout = atoi(refdb_timeout);
      sprintf(outbuffer, "set timeout to %d seconds", n_refdb_timeout);
      LOG_PRINT(LOG_INFO, outbuffer);
    }
    else if (strncmp(command + 1, "logfile", 7) == 0) {
      strcpy(log_file, command + 9);
      sprintf(outbuffer, "set logfile to %s", log_file);
      LOG_PRINT(LOG_INFO, outbuffer);
      n_reopen_log = 1; /* signal the log fn to open new log file */
    }
    else if (strncmp(command + 1, "logdest", 7) == 0) {
      strcpy(log_dest, command + 9);
      n_log_dest = atoi(log_dest);
      sprintf(outbuffer, "set logdest to %d", n_log_dest);
      LOG_PRINT(LOG_INFO, outbuffer);
    }
    else if (strncmp(command + 1, "loglevel", 8) == 0) {
      strcpy(log_level, command + 10);
      n_log_level = atoi(log_level);
      sprintf(outbuffer, "set loglevel to %d", n_log_level);
      LOG_PRINT(LOG_INFO, outbuffer);
    }
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  connect_to_db(): establishes a connection with the SQL database server
  
  dbi_conn connect_to_db Returns connection ptr if successful,
                       NULL if failed      

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with connection info

  const char* special_db ptr to string containing the name of a special
              database to connect to if the latter is different from
	      the one in ptr_clrequest

  int nocheck if 1, don't verify the database is a RefDB database

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
dbi_conn connect_to_db(struct CLIENT_REQUEST* ptr_clrequest, const char* special_db, int nocheck) {
  dbi_conn conn;
  char *my_db;
  char main_db[] = MAIN_DB;

  LOG_PRINT(LOG_DEBUG, ptr_clrequest->server_ip);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->username);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->passwd);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->current_db);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->dbs_port_address);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->dbserver);
  LOG_PRINT(LOG_DEBUG, ptr_clrequest->db_path);

  conn = dbi_conn_new(ptr_clrequest->dbserver);

  if (!conn) {
    LOG_PRINT(LOG_WARNING, "creating database connection structure failed");
    return NULL;
  }

  my_db = (special_db && *special_db) ? (char*)special_db : ((*(ptr_clrequest->current_db)) ? ptr_clrequest->current_db : main_db);
  LOG_PRINT(LOG_DEBUG, my_db);
  /* common driver options */
  dbi_conn_set_option(conn, "username", ptr_clrequest->username);
  dbi_conn_set_option(conn, "password", (ptr_clrequest->passwd && *(ptr_clrequest->passwd)) ? ptr_clrequest->passwd : "");
  /* connect to refdb unless a specific reference database is requested
     special_db overrides ptr_clrequest->current_db */
  dbi_conn_set_option(conn, "dbname", my_db);
  dbi_conn_set_option_numeric(conn, "port", atoi(ptr_clrequest->dbs_port_address));

  /* driver-specific options */
  if (!strcmp(ptr_clrequest->dbserver, "mysql")) {
    dbi_conn_set_option_numeric(conn, "compression", 1);
    dbi_conn_set_option(conn, "host", ptr_clrequest->server_ip);
  }
  else if (!strcmp(ptr_clrequest->dbserver, "pgsql")) {
    /* don't set host if we want to connect locally */
    if (strcmp(ptr_clrequest->server_ip, "localhost")) {
      dbi_conn_set_option(conn, "host", ptr_clrequest->server_ip);
    }
  }
  else if (!strcmp(ptr_clrequest->dbserver, "sqlite")) {
    dbi_conn_set_option(conn, "sqlite_dbdir", ptr_clrequest->db_path);
  }

  if (dbi_conn_connect(conn) < 0) { /* -1 and -2 indicate errors */
    LOG_PRINT(LOG_WARNING, "failed to connect to database server using database:");
    LOG_PRINT(LOG_WARNING, my_db);
    dbi_conn_close(conn);
    return NULL;
  }

  if (!nocheck && strcmp(my_db, MAIN_DB) && !is_reference_database(NULL, conn, my_db)) {
    LOG_PRINT(LOG_WARNING, "failed to connect to database server using database:");
    LOG_PRINT(LOG_WARNING, my_db);
    dbi_conn_close(conn);
    return NULL;
  }

  LOG_PRINT(LOG_DEBUG, "connected to database server using database:");
  LOG_PRINT(LOG_DEBUG, my_db);
  return conn;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  viewstat(): executes client command viewstat

  void viewstat

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void viewstat(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn; /* libdbi connection structure */
  dbi_driver driver; /* libdbi driver structure */
  dbi_result dbires; /* libdbi result structure */
  char* outstring; /* ptr to a buffer for output */
  char* outbuffer; /* ptr to a buffer for output */
  const char* versioninfo = NULL;
  char sql_command[32];

  /* get a buffer to hold output data */
  if ((outbuffer = (char*)malloc((size_t)OUTBUF_LEN)) == NULL) {
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    return;
  }
    
  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    driver = dbi_conn_get_driver(conn);

    /* obtain database server version info */
    sprintf(sql_command, "SELECT VERSION()");
    dbires = dbi_conn_query(conn, sql_command);
    LOG_PRINT(LOG_DEBUG, sql_command);
    if (dbires) {
      if (dbi_result_next_row(dbires)) {
	versioninfo = dbi_result_get_string_idx(dbires, 1); /* 1-base index */
      }
    }

    sprintf(outbuffer, "You are served by: %s %s\nClient IP: %s\nConnected via %s driver (%s)\nto: %s\nserverip: %s\ntimeout: %s\ndbs_port: %s\nlogfile: %s\nlogdest: %s\nloglevel: %s\nremoteadmin: %s\npidfile: %s\n",
	    PACKAGE,
	    VERSION,
	    ptr_clrequest->client_ip,
	    dbi_driver_get_name(driver),
	    dbi_driver_get_version(driver),
	    (versioninfo) ? versioninfo : "unknown",
	    ptr_clrequest->server_ip,
	    refdb_timeout,
	    ptr_clrequest->dbs_port_address,
	    log_file,
	    log_dest,
	    log_level,
	    (n_remote_admin) ? "on" : "off",
	    pid_file);

    if (dbires) {
      dbi_result_free(dbires);
    }

    tiwrite(ptr_clrequest->fd, outbuffer, TERM_NO);
    dbi_conn_close(conn);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
  }
  free(outbuffer);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  listdb(): implements the client command listdb

  void listdb This function has no return value

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  int select if 1, answers selectdb command, if 0, answers listdb command

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void listdb(struct CLIENT_REQUEST* ptr_clrequest, int select) {
  char* html_head;
  char* html_foot;
  const char* dbname;
  dbi_conn conn;
  dbi_result dbires;

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    if (*(ptr_clrequest->argument)) {
      /* this will list the databases that match the provided pattern */
      dbires = dbi_conn_get_db_list(conn, ptr_clrequest->argument);
    }
    else {
      /* this will list all databases */
      dbires = dbi_conn_get_db_list(conn, NULL);
    }

    if (ptr_clrequest->n_cgi) {
      html_head = load_html("refdbq_head", refdblib);
      html_foot = load_html("refdbq_foot", refdblib);
      if (!html_head || !html_foot) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
	tiwrite(ptr_clrequest->fd, "could not load cgi templates\n", TERM_NO);
	free(html_head);
	free(html_foot);
	dbi_result_free(dbires);
	dbi_conn_close(conn);
	return;
      }
    }

    if (select) {
      /* selectdb */
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_html.text, cgihead_html.length);
	tiwrite(ptr_clrequest->fd, html_head, TERM_NO);
	tiwrite(ptr_clrequest->fd, "<thead><tr><th class=\"result\">Current Database</th></tr></thead><tbody><tr><td class=\"result\">", TERM_NO);
	free(html_head);
      }
      if (dbi_result_next_row(dbires)) {
	/* this is just to prevent problems if the SQL server
	   allows longer database names than we expect */
	dbname = dbi_result_get_string_idx(dbires, 1); /* 1-based index */
	if (dbname) {
	  if (is_reference_database(ptr_clrequest, NULL, dbname)) {
	    if (strlen(dbname) > DBNAME_LENGTH) {
	      iwrite(ptr_clrequest->fd, dbname, DBNAME_LENGTH);
	    }
	    else {
	      iwrite(ptr_clrequest->fd, dbname, strlen(dbname));
	    }
	  }
	  else {
	    tiwrite(ptr_clrequest->fd, "not a RefDB database\n", TERM_NO);
	  }
	}
	else {
	  tiwrite(ptr_clrequest->fd, "not a RefDB database\n", TERM_NO);
	}
      }
      else {
	tiwrite(ptr_clrequest->fd, "database does not exist\n", TERM_NO);
      }
      if (ptr_clrequest->n_cgi) {
	tiwrite(ptr_clrequest->fd, "</td></tr>", TERM_NO);
	tiwrite(ptr_clrequest->fd, html_foot, TERM_NO);
	free(html_foot);
      }
    }
    else {
      /* listdb */
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_html.text, cgihead_html.length);
	tiwrite(ptr_clrequest->fd, html_head, TERM_NO);
  	tiwrite(ptr_clrequest->fd, "<thead><tr><th class=\"result\">Available Databases</th></tr></thead><tbody><tr><td><input type=\"button\" value=\"Select database\" onclick=\"seldb()\"></td></tr>", TERM_NO);
	free(html_head);
      }
      while (dbi_result_next_row(dbires)) {
	dbname = dbi_result_get_string_idx(dbires, 1);
	if (dbname) {
	  if (is_reference_database(ptr_clrequest, NULL, dbname)) {
	    if (ptr_clrequest->n_cgi) {
	      tiwrite(ptr_clrequest->fd, "<tr><td class=\"result\"><input type=\"radio\" name=\"selectdb\" value=\"", TERM_NO);
	    }
	    tiwrite(ptr_clrequest->fd, dbname, TERM_NO);
	    if (ptr_clrequest->n_cgi) {
	      tiwrite(ptr_clrequest->fd, "\">", TERM_NO);
	      tiwrite(ptr_clrequest->fd, dbname, TERM_NO);
	      tiwrite(ptr_clrequest->fd, "</td></tr>", TERM_NO);
	    }
	    tiwrite(ptr_clrequest->fd, "\n", TERM_NO);
	  }
	}
      }
      if (ptr_clrequest->n_cgi) {
	tiwrite(ptr_clrequest->fd, html_foot, TERM_NO);
	free(html_foot);
      }
    }
    dbi_result_free(dbires);
    dbi_conn_close(conn);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr.text, connerr.length);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_reference_database(): checks whether a database is a RefDB
                           reference database

  int is_reference_database returns 1 if the database is a RefDB database,
                            0 if not

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info
                         set to NULL if using conn

  dbi_conn conn optional existing connection to database. Overrides
                         ptr_clrequest. Set to NULL if not used

  const char* dbname ptr to a string containing the database name

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int is_reference_database(struct CLIENT_REQUEST* ptr_clrequest, dbi_conn conn, const char* dbname) {
  dbi_conn conn_ref;
  dbi_result dbires;
  char sql_command[128];
  const char* app;
  const char* type;
  int retval = 0;
  int close_conn_ref = 0;

  if (conn) {
    conn_ref = conn; /* legal as conn and conn_ref are pointers */
  }
  else {
    if ((conn_ref = connect_to_db(ptr_clrequest, dbname, 1)) == NULL) {
      /* can't access */
      return 0;
    }
    close_conn_ref++;
  }

  sprintf(sql_command, "SELECT meta_app,meta_type from t_meta");

  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn_ref, sql_command);

  if (!dbires) {
    goto Finish;
  }

  if (!dbi_result_next_row(dbires)) {
    goto Finish;
  }

  app = dbi_result_get_string(dbires, "meta_app");
  type = dbi_result_get_string(dbires, "meta_type");

  if (!strcmp(app, "refdb")) {
    retval = 1;
  }

 Finish:
  if (close_conn_ref) {
    dbi_conn_close(conn_ref);
  }

  return retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  listuser(): implements the client command listuser

  void listuser

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void listuser(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  char sql_command[256];
  const char* username;
  char* myarg;
  int result;

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    driver = dbi_conn_get_driver(conn);

    if (ptr_clrequest->argument && *(ptr_clrequest->argument)) {
      myarg = strdup(ptr_clrequest->argument);
    }
    else {
      myarg = strdup(my_dbi_driver_get_cap(driver, "listall")); /* list all if no argument is specified */
    }
    if (!myarg) {
      dbi_conn_close(conn);
      LOG_PRINT(LOG_DEBUG, outomem.text);
      iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
      return;
    }

    /* escape any characters that the database server cannot digest */
    if (dbi_driver_quote_string(driver, &myarg) == -1) {
      dbi_conn_close(conn);
      free(myarg);
      LOG_PRINT(LOG_WARNING, outomem.text);
      iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
      return;
    }

    sprintf(sql_command, "SELECT DISTINCT user_name FROM t_user WHERE user_name %s %s", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    dbires = dbi_conn_query(conn, sql_command);
    LOG_PRINT(LOG_DEBUG, sql_command);
    if (dbires) {
      while (dbi_result_next_row(dbires)) {
	username = dbi_result_get_string(dbires, "user_name");
	if (username) {
	  tiwrite(ptr_clrequest->fd, username, TERM_NO);
	  tiwrite(ptr_clrequest->fd, "\n", TERM_NO);
	}
      }
      dbi_result_free(dbires);
    }
    dbi_conn_close(conn);
    free(myarg);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr.text, connerr.length);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  listword(): implements the client command listword

  void listword

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void listword(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  char sql_command[256];
  const char* wordname;
  char* myarg;
  int result;

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    driver = dbi_conn_get_driver(conn);

    if (ptr_clrequest->argument && *(ptr_clrequest->argument)) {
      myarg = strdup(ptr_clrequest->argument);
    }
    else {
      myarg = strdup(my_dbi_driver_get_cap(driver, "listall")); /* list all if no argument is specified */
    }
    if (!myarg) {
      dbi_conn_close(conn);
      LOG_PRINT(LOG_DEBUG, outomem.text);
      iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
      return;
    }

    /* escape any characters that the database server cannot digest */
    if (dbi_driver_quote_string(driver, &myarg) == -1) {
      dbi_conn_close(conn);
      free(myarg);
      LOG_PRINT(LOG_WARNING, outomem.text);
      iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
      return;
    }

    sprintf(sql_command, "SELECT DISTINCT name FROM t_journal_words WHERE name %s %s", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    dbires = dbi_conn_query(conn, sql_command);
    LOG_PRINT(LOG_DEBUG, sql_command);
    if (dbires) {
      while (dbi_result_next_row(dbires)) {
	wordname = dbi_result_get_string(dbires, "name");
	if (wordname) {
	  tiwrite(ptr_clrequest->fd, wordname, TERM_NO);
	  tiwrite(ptr_clrequest->fd, "\n", TERM_NO);
	}
      }
      dbi_result_free(dbires);
    }
    dbi_conn_close(conn);
    free(myarg);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr.text, connerr.length);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  createdb(): implements the client command createdb

  int createdb returns 0 if error, 1 if successful 

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  const char* dbname ptr to string with the database name

  const char* db_encoding name of the character encoding to use

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int createdb(struct CLIENT_REQUEST* ptr_clrequest, const char* dbname, const char* db_encoding) {
  dbi_conn conn;
  dbi_result dbires;
  int error;
  char date_buffer[24];
  char *sql_command;
  const char *drivername;
  time_t the_time;

  /* it is not necessary to quote database names */

  /* connect to the main database */
  if ((conn = connect_to_db(ptr_clrequest, MAIN_DB, 0)) == NULL) {
    iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
    return 0;
  }
  
  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  /* allocate a buffer to assemble the sql command (give some extra byte
     for the command and encoding strings */
  sql_command = (char*)malloc((size_t)(strlen(dbname) + 256));
  if (sql_command == NULL) { /* malloc failed */
    iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
    return 0;
  }

  /* create a new database. There is no such SQL command for sqlite.
   sqlite will create the database for us as soon as we try to access
   it */

  if (strcmp(drivername, "sqlite")) {
    /* assemble sql command */
    if (!strcmp(my_dbi_conn_get_cap(conn, "encoding"), "t")) {
      if (db_encoding && *db_encoding) {
	sprintf(sql_command, "CREATE DATABASE %s WITH ENCODING = '%s'", dbname, db_encoding);
      }
      else { /* use default encoding of db server */
	sprintf(sql_command, "CREATE DATABASE %s", dbname);
      }
    }
    else {
      /* mysql does not support database-based encodings (yet) */
      sprintf(sql_command, "CREATE DATABASE %s", dbname);
    }

    LOG_PRINT(LOG_DEBUG, sql_command);

    dbires = dbi_conn_query(conn, sql_command);

    if (dbires == NULL) {
      tiwrite(ptr_clrequest->fd, "createdb failed\n", TERM_NO);
      free(sql_command);
      dbi_conn_close(conn);
      return 0;
    }
    
    dbi_result_free(dbires);
  } /* end if not sqlite */

  /* now we have to create the tables */
  if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "t")) {
    /* we switch over to the new database */
    error = dbi_conn_select_db(conn, dbname);
    if (error) {
      tiwrite(ptr_clrequest->fd, "selectdb failed\n", TERM_NO);
      dbi_conn_close(conn);
      free(sql_command);
      return 0;
    }
    /* if (!strcmp(drivername, "mysql")) { */
    if (create_tables_mysql(conn, ptr_clrequest)) {
      /* function provides message to client */
      dbi_conn_close(conn);
      free(sql_command);
      LOG_PRINT(LOG_ERR, "create tables failed");
      return 0;
    }
  }
  else {
    /* we need a new connection to the fresh database */
    dbi_conn_close(conn);

    /* do not check the database as it is brand new */
    if ((conn = connect_to_db(ptr_clrequest, dbname, 1)) == NULL) {
      iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
      free(sql_command);
      return 0;
    }
  
    if (!strcmp(drivername, "pgsql")) {
      if (create_tables_pgsql(conn, ptr_clrequest)) {
	/* function provides message to client */
	dbi_conn_close(conn);
	LOG_PRINT(LOG_ERR, "create tables failed");
	free(sql_command);
	return 0;
      }
    }
    else if (!strcmp(drivername, "sqlite")) {
      if (create_tables_sqlite(conn, ptr_clrequest)) {
	/* function provides message to client */
	dbi_conn_close(conn);
	LOG_PRINT(LOG_ERR, "create tables failed");
	free(sql_command);
	return 0;
      }
    }
    /* else: should never happen */
  }

  /* insert the meta data */
  time(&the_time);
  strftime(date_buffer, 24, "%Y-%m-%d %H:%M:%S", gmtime(&the_time));
  sprintf(sql_command, "INSERT INTO t_meta (meta_app,meta_type,meta_version,meta_create_date,meta_modify_date) VALUES (\'%s\', \'risx\', \'%s\', \'%s\', \'%s\')", PACKAGE, VERSION, date_buffer, date_buffer);
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "insert into t_meta failed\n", TERM_NO);
    LOG_PRINT(LOG_ERR, "insert into t_meta failed");
    free(sql_command);
    return 1;
  }
   
  dbi_result_free(dbires);


  /* cleanup */
  dbi_conn_close(conn);
  free(sql_command);
      
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  deletedb(): implements the client command deletedb

  int deletedb returns 0 if failed, 1 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  const char* dbname ptr to string with the database name

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int deletedb(struct CLIENT_REQUEST* ptr_clrequest, const char* dbname) {
  dbi_conn conn;
  dbi_result dbires;
  int error;
  char *sql_command;
  const char *drivername;

  /* it is not necessary to quote the database name */

  /* allocate a buffer to assemble the sql command. We require
     _POSIX_PATH_MAX for the sqlite driver, but this will be more than
     enough for database names of other drivers */
  sql_command = (char*)malloc((size_t)_POSIX_PATH_MAX);
  if (sql_command == NULL) { /* malloc failed */
    LOG_PRINT(LOG_WARNING, "out of memory");
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    return 0;
  }

  /* connect to database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
    return 0;
  }

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  if (!strcmp(drivername, "sqlite")) {
    dbi_conn_close(conn);

    /* sqlite has no SQL way to drop databases. Try to remove the
       database file from the file system manually */
    strcpy(sql_command, ptr_clrequest->db_path);
    if (sql_command[strlen(sql_command)-1] != '/') {
      strcat(sql_command, "/");
    }
    strcat(sql_command, dbname);

    if (unlink(sql_command)) {
      tiwrite(ptr_clrequest->fd, "failed to delete database file\n", TERM_NO);
      free(sql_command);
      return 0;
    }
    else {
      free(sql_command);
      return 1;
    }
  }

  /* assemble sql command */
  strcpy(sql_command, "DROP DATABASE ");
  strcat(sql_command, dbname);

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "deletedb failed\n", TERM_NO);
    dbi_conn_close(conn);
    free(sql_command);
    return 0;
  }
  else {
    dbi_result_free(dbires);
  }

  if (!strcmp(drivername, "pgsql")) {
    /* With PostgreSQL we keep a group to manage access rights for each
       database. When dropping the database we drop the group too. This
       does not affect the existence of the users contained in this group */
    if (is_group_pgsql(conn, ptr_clrequest->argument) == 1) {
      sprintf(sql_command, "DROP GROUP %suser", ptr_clrequest->argument);
      
      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires = dbi_conn_query(conn, sql_command);
      if (error) {
	tiwrite(ptr_clrequest->fd, "DROP GROUP failed\n", TERM_NO);
	dbi_conn_close(conn);
	free(sql_command);
	return 0;
      }
      
      dbi_result_free(dbires);
    }
  }
  /* else: do nothing for other drivers */

  dbi_conn_close(conn);
  free(sql_command);
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  adduser(): implements the client command adduser

  int adduser returns 0 if failed, 1 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  char *user_host string containing the name of the host from which
                  username will access the database. This is actually
                  the host on which the refdbd runs that username
                  will use to access the database. if user_host is
                  NULL, it will be set to localhost

  char *newuser_passwd string with the password for new users

  int n_remove if 0, access will be granted. if 1, access will be revoked

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int adduser(struct CLIENT_REQUEST* ptr_clrequest, char *user_host, char *newuser_passwd, int n_remove) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_result dbires1;
  int num_users = 0;
  int failed = 0;
  int error = 0;
  int nread_done = 0;
  int inbuffer_len;
  int token_size;
  int numbyte;
  int nhave_refdb;
  int n_host_regex = 0;
  int n_user_regex = 0;
  char *sql_command;
  char *inbuffer;
  char *new_inbuffer;
  char *token;
  char *mytoken;
  char *userathost;
  const char *item;
  const char *drivername;
  char localhost[] = "localhost";
  char old_db[DBNAME_LENGTH+1];
  struct lilimem sentinel;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  sql_command = (char*)malloc(1024); 
  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    LOG_PRINT(LOG_WARNING, outomem.text);
    return 0;
  }

  inbuffer_len = (size_t)COMMAND_INBUF_LEN;
  inbuffer = (char*)malloc(inbuffer_len);
  if (inbuffer == NULL || insert_lilimem(&sentinel, (void**)&inbuffer, NULL)) {
    delete_all_lilimem(&sentinel);
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    LOG_PRINT(LOG_WARNING, outomem.text);
    return 0;
  }

  /* send back confirmation to the client */
  numbyte = tiwrite(ptr_clrequest->fd, positive.text, TERM_YES);
  if (numbyte == -1) {
    delete_all_lilimem(&sentinel);
    LOG_PRINT(LOG_WARNING, "could not write to client");
    return 0;
  }

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    /* read from client and increase buffer size until we have all arguments */
    while (!nread_done) {
      numbyte = tread(ptr_clrequest->fd, inbuffer, COMMAND_INBUF_LEN-1);
      if (get_trailz(inbuffer, numbyte) == TERM_LEN) { /* if transmission complete */
	nread_done = 1;
      }
      else {
	inbuffer_len += (size_t)COMMAND_INBUF_LEN;
	if ((new_inbuffer = (char*)realloc(inbuffer, inbuffer_len)) == NULL) {
	  delete_all_lilimem(&sentinel);
	  iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
	  dbi_conn_close(conn);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return 0;
	}
	else {
	  inbuffer = new_inbuffer;
	}
      }
    }

    /* see what driver we use */
    drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

    /* +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
    /* mysql */

    /* mysqld keeps authentication information in the mysql
       database. These data are used for all user databases.
       Therefore we need only one entry to allow access to
       the database server, and entries per database to
       allow access to the databases themselves */
    if (!strcmp(drivername, "mysql")) {
      int create_new = 0;
      char lockstring[13];
      struct VERSIONINFO ver;

      /* obtain database server version info */
      if (my_dbi_conn_get_versioninfo(conn, &ver)) {
	iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
	dbi_conn_close(conn);
	delete_all_lilimem(&sentinel);
	return 0;
      }

      /* MySQL 4.0.2 introduced a new privilege "LOCK TABLES". As older
	 versions will barf at it, add it only if we use a newer
	 version */
      if ((ver.major == 4 && ver.minor == 0 && ver.minuscule >=2)
	  || (ver.major == 4 && ver.minor > 0)
	  || ver.major > 4) {
	strcpy(lockstring, ",LOCK TABLES ");
      }
      else {
	strcpy(lockstring, " ");
      }
	
      /* use localhost if no host was specified */
      if (!*user_host) {
	tiwrite(ptr_clrequest->fd, "use localhost as host\n", TERM_NO);
	user_host = localhost;
      }

      /* select mysql database */
      if (dbi_conn_select_db(conn, "mysql") == -1) {
	iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
	dbi_conn_close(conn);
	delete_all_lilimem(&sentinel);
	return 0;
      }

      /* see whether the host info is a SQL regexp */
      if (strchr(user_host, (int)'%') || strchr(user_host, (int)'_')) {
	n_host_regex = 1;
      }
      else {
	n_host_regex = 0;
      }
      
      token = inbuffer;
      token_size = -1;

      userathost = (char*)malloc(USERNAME_LENGTH + HOSTNAME_LENGTH + 16); 
      if (userathost == NULL || insert_lilimem(&sentinel, (void**)&userathost, NULL)) {
	iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
	LOG_PRINT(LOG_WARNING, outomem.text);
	return 0;
      }

      do {
	token = nstrtok(token + token_size + 1, &token_size, " ");
	if (token != NULL) {
	  token[token_size] = '\0';
	  
	  /* see whether the user info is a SQL regexp */
	  if (strchr(token, (int)'%') || strchr(token, (int)'_')) {
	    n_user_regex = 1;
	  }
	  else {
	    n_user_regex = 0;
	  }

	  if (n_remove) {
	    if (n_user_regex || n_host_regex) {
	      sprintf(sql_command, "REVOKE ALL ON %s.* FROM \'%s\'@\'%s\'", ptr_clrequest->current_db, token, user_host);
	    }
	    else {
	      sprintf(sql_command, "REVOKE ALL ON %s.* FROM %s@%s", ptr_clrequest->current_db, token, user_host);
	    }
	  }
	  else {
	    /* check whether the given user/host combo exists in User table */
	    sprintf(sql_command, "SELECT User FROM user WHERE User LIKE \'%s\' AND Host LIKE \'%s\'", token, user_host);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    dbires = dbi_conn_query(conn, sql_command);
	    if (!dbires) {
	      failed++; /* ToDo: do something intelligent here */
	      LOG_PRINT(LOG_WARNING, "select from mysql.user failed");
	      continue; 
	    }

	    if (!dbi_result_get_numrows(dbires) && !n_user_regex) { /* user does not exist yet and does not contain regexp */
	      create_new++;
	    }

	    dbi_result_free(dbires);
	    
	    /* check whether this host/user combo has access rights to
	       refdb. If not, create as well */
	    /* reset variables */
	    nhave_refdb = 0;
	    sprintf(sql_command, "SELECT User,Db FROM db WHERE User=\'%s\' AND Host=\'%s\' AND Db=\'%s\'", token, user_host, MAIN_DB);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    dbires = dbi_conn_query(conn, sql_command);
	    if (!dbires) {
	      failed++; /* ToDo: do something intelligent here */
	      LOG_PRINT(LOG_WARNING, "select from mysql.db failed");
	      continue; 
	    }

	    while (dbi_result_next_row(dbires)) {
	      item = dbi_result_get_string_idx(dbires, 2);
	      if (!strcmp(item, MAIN_DB)) {
		nhave_refdb = 1;
	      }
	    }

	    dbi_result_free(dbires);
	    
	    if (n_user_regex || n_host_regex) {
	      sprintf(userathost, "\'%s\'@\'%s\'", token, user_host);
	    }
	    else {
	      sprintf(userathost, "%s@%s", token, user_host);
	    }

	    if (!nhave_refdb) { /* user does not have access rights to refdb */
	      if (newuser_passwd && *newuser_passwd) {
		/* if the user doesn't exist yet, this command will create
		   a new user with the given password. If the user already
		   exists, the password will be changed to the one given */
		sprintf(sql_command, "GRANT SELECT, INSERT, UPDATE, DELETE%s ON %s.* TO %s IDENTIFIED BY \'%s\'", lockstring, MAIN_DB, userathost, newuser_passwd);
	      }
	      else if (create_new) {
		/* avoid possible security risk: we provide a default
		   password to avoid creating users with no passwords */
		sprintf(sql_command, "GRANT SELECT, INSERT, UPDATE, DELETE%s ON %s.* TO %s IDENTIFIED BY \'refdb\'", lockstring, MAIN_DB, userathost);
	      }
	      else {
		/* the user already exists and we don't touch the password */
		sprintf(sql_command, "GRANT SELECT, INSERT, UPDATE, DELETE%s ON %s.* TO %s", lockstring, MAIN_DB, userathost);
	      }
	      LOG_PRINT(LOG_DEBUG, sql_command);
	      dbires = dbi_conn_query(conn, sql_command);
	      if (!dbires) {
		LOG_PRINT(LOG_WARNING, "insert into mysql.db failed");
		failed++;
		continue;
	      }
	      dbi_result_free(dbires);
	    }


	    /* now handle the requested database. Users need CREATE and 
	       DROP privileges for temp tables during bibliography creation*/
	    if (n_user_regex || n_host_regex) {
	      sprintf(sql_command, "GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP%s ON %s.* TO \'%s\'@\'%s\'", lockstring, ptr_clrequest->current_db, token, user_host);
	    }
	    else {
	      sprintf(sql_command, "GRANT SELECT, INSERT, UPDATE, DELETE, CREATE, DROP%s ON %s.* TO %s@%s", lockstring, ptr_clrequest->current_db, token, user_host);
	    }
	  }
	  LOG_PRINT(LOG_DEBUG, sql_command);
	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    failed++;
	  }
	  else {
	    num_users++;
	    dbi_result_free(dbires);
	  }
	}
      } while (token != NULL);

      /* reloading permission tables is not necessary when using 
	 the GRANT/REVOKE syntax */

    }

    /* +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
    /* pgsql */
    else if (!strcmp(drivername, "pgsql")) {
      /* PostgreSQL knows users and groups. Access is granted to tables
	 (and other stuff we don't use here). The system tables that
	 hold the user and group information are accessible through
	 connections to any existing database (at least for the
	 superuser, that is).
	 Each new refdb database creates a group and grants access for this
	 group. We simply add users to or remove them from this group.
	 Same for the group refdbuser which allows access to the common
	 database refdb
       */

      token = inbuffer;
      token_size = -1;

      /* loop over all usernames */
      /* ToDo: this loop is inefficient. Better construct a comma-separated
	 list of users and pass this as an argument to a single ALTER GROUP
	 command */
      do {
	token = nstrtok(token + token_size + 1, &token_size, " ");
	if (token != NULL) {
	  token[token_size] = '\0';
	  
	  if (n_remove) {
	    /* the current implementation removes only the group membership
	       for the given database. Users will still have access to the
	       refdb database. */
	    /* ToDo: maybe a client command switch can be used to force
	       removal from group refdbuser */
	    sprintf(sql_command, "ALTER GROUP %suser DROP USER %s", ptr_clrequest->current_db, token);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    dbires = dbi_conn_query(conn, sql_command);
	    if (!dbires) {
	      LOG_PRINT(LOG_WARNING, "ALTER GROUP failed");
	      failed++;
	    }
	    else {
	      num_users++;
	      dbi_result_free(dbires);
	    }
	  }
	  else {
	    if (is_user_pgsql(conn, token) != 1) {
	      sprintf(sql_command, "CREATE USER %s WITH PASSWORD \'%s\'", token, (*newuser_passwd) ? newuser_passwd : "refdb");
	      LOG_PRINT(LOG_DEBUG, sql_command);
	      dbires = dbi_conn_query(conn, sql_command);
	      if (!dbires) {
		LOG_PRINT(LOG_WARNING, "CREATE USER failed");
		failed++;
		continue;
	      }
	      dbi_result_free(dbires);
	    }

	    /* Add user to the group refdbuser */
	    /* Adding users that already are members will not cause
	       an error. Checking membership would be extremely tedious as
	       the user list is implemented as an array instead of
	       using a separate table */
	    sprintf(sql_command, "ALTER GROUP %suser ADD USER %s", /*MAIN_DB*/"refdb", token);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    dbires = dbi_conn_query(conn, sql_command);
	    if (!dbires) {
	      LOG_PRINT(LOG_WARNING, "ALTER GROUP failed");
	      failed++;
	      continue;
	    }
	    dbi_result_free(dbires);

	    /* now handle the requested database and add the user
	       to the group <dbname>user */
	    sprintf(sql_command, "ALTER GROUP %suser ADD USER %s", ptr_clrequest->current_db, token);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    dbires = dbi_conn_query(conn, sql_command);
	    if (!dbires) {
	      LOG_PRINT(LOG_WARNING, "ALTER GROUP failed");
	      failed++;
	    }
	    else {
	      num_users++;
	      dbi_result_free(dbires);
	    }
	  }
	}
      } while (token != NULL); /* end loop over all usernames */
    }
    else { /* driver does not support SQL-based access control */
      tiwrite(ptr_clrequest->fd, "access control not supported", TERM_NO);
      LOG_PRINT(LOG_WARNING, "access control not supported");
    }
    dbi_conn_close(conn);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr.text, connerr.length);
  }
  
  /* reuse sql_command for a return message */
  if (n_remove) {
    sprintf(sql_command, "%d user entries removed, %d failed\n", num_users, failed);
  }
  else {
    sprintf(sql_command, "%d user entries added, %d failed\n", num_users, failed);
  }
  tiwrite(ptr_clrequest->fd, sql_command, TERM_NO);
  delete_all_lilimem(&sentinel);
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  addword(): implements the client command addword

  int addword returns 0 if failed, 1 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  int n_remove if 0, access wil be granted. if 1, access will be revoked

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int addword(struct CLIENT_REQUEST* ptr_clrequest, int n_remove) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  int num_words = 0;
  int failed = 0;
  int error = 0;
  int nread_done = 0;
  int inbuffer_len;
  int token_size;
  int numbyte;
  int n_word_regex = 0;
  char *sql_command;
  char *inbuffer;
  char *new_inbuffer;
  char *token;
  char* mytoken;
  struct lilimem sentinel;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';


  sql_command = (char*)malloc(1024); 
  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    LOG_PRINT(LOG_WARNING, outomem.text);
    return 0;
  }

  inbuffer_len = (size_t)COMMAND_INBUF_LEN;
  inbuffer = (char*)malloc(inbuffer_len);
  if (inbuffer == NULL || insert_lilimem(&sentinel, (void**)&inbuffer, NULL)) {
    delete_all_lilimem(&sentinel);
    iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
    LOG_PRINT(LOG_WARNING, outomem.text);
    return 0;
  }

  /* send back confirmation to the client */
  numbyte = tiwrite(ptr_clrequest->fd, positive.text, TERM_YES);
  if (numbyte == -1) {
    delete_all_lilimem(&sentinel);
    LOG_PRINT(LOG_WARNING, "could not write to client");
    return 0;
  }

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {

    /* read from client and increase buffer size until we have all arguments */
    while (!nread_done) {
      numbyte = tread(ptr_clrequest->fd, inbuffer, COMMAND_INBUF_LEN-1);
      if (get_trailz(inbuffer, numbyte) >= TERM_LEN) { /* if transmission complete */
	nread_done = 1;
      }
      else {
	inbuffer_len += (size_t)COMMAND_INBUF_LEN;
	if ((new_inbuffer = (char*)realloc(inbuffer, inbuffer_len)) == NULL) {
	  delete_all_lilimem(&sentinel);
	  iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
	  dbi_conn_close(conn);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return 0;
	}
	else {
	  inbuffer = new_inbuffer;
	}
      }
    } /* end while(!nread_done) */

/*      printf("%s\n", inbuffer); */

    token = inbuffer;
    token_size = -1;

    do {
      token = nstrtok(token + token_size + 1, &token_size, " ");
      if (token != NULL) {
	token[token_size] = '\0';

	/* make sure we have uppercase words only */
	mytoken = strdup(token);
	if (!mytoken) {
	  failed++;
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  continue;
	}
	strup(mytoken);

	driver = dbi_conn_get_driver(conn);

	/* escape any characters that the database server cannot digest */
	if (dbi_driver_quote_string(driver, &mytoken) == -1) {
	  free(mytoken);
	  failed++;
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  continue;
	}

	if (n_remove) {
  	  sprintf(sql_command, "DELETE FROM t_journal_words WHERE name %s %s", my_dbi_conn_get_cap(conn, "rlike"), mytoken);
	}
	else {
	  sprintf(sql_command, "INSERT INTO t_journal_words (name) VALUES (%s)", mytoken);
	}
	LOG_PRINT(LOG_DEBUG, sql_command);
	free(mytoken);
	dbires = dbi_conn_query(conn, sql_command);
	if (!dbires) {
	  failed++;
	}
	else {
	  num_words++;
	  dbi_result_free(dbires);
	}
      }
    } while (token != NULL);
    dbi_conn_close(conn);
  }
  else {
    iwrite(ptr_clrequest->fd, connerr.text, connerr.length);
  }
  
  /* reuse sql_command for a return message */
  if (n_remove) {
    sprintf(sql_command, "%d word entries removed, %d failed\n", num_words, failed);
  }
  else {
    sprintf(sql_command, "%d word entries added, %d failed\n", num_words, failed);
  }
  tiwrite(ptr_clrequest->fd, sql_command, TERM_NO);
  delete_all_lilimem(&sentinel);
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  getfoo(): implements the client commands getau, getkw, and getjo

  void getfoo

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  int type the type of user command 0 = getau, 1 = getkw, 2 = getjo,
                                    3 = getjf, 4 = getj1, 5 = getj2

  int all if 1, return all synonyms of a journal name query

  char* db_encoding output encoding

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void getfoo(struct CLIENT_REQUEST* ptr_clrequest, int type, int all, char* db_encoding) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  int n_sql_command_len;
  int error;
  char *sql_command;
  char *new_sql_command;
  char *html_head;
  char *html_foot;
  char *myarg;
  const char *foostring;

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) != NULL) {
    driver = dbi_conn_get_driver(conn);

    if (ptr_clrequest->argument && *(ptr_clrequest->argument)) {
      myarg = strdup(ptr_clrequest->argument);
    }
    else {
      myarg = strdup(my_dbi_driver_get_cap(driver, "listall")); /* list all if no argument is specified */
    }
    if (!myarg) {
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
      }
      iwrite(ptr_clrequest->fd, outomem_n.text, outomem_n.length);
      dbi_conn_close(conn);
      return;
    }

    /* escape any characters that the database server cannot digest */
    if (dbi_driver_quote_string(driver, &myarg) == -1) {
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
      }
      iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
      dbi_conn_close(conn);
      free(myarg);
      return;
    }

    n_sql_command_len = 256+strlen(myarg);
    sql_command = (char*)malloc(n_sql_command_len); 
    if (sql_command == NULL) {
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
      }
      iwrite(ptr_clrequest->fd, outomem.text, outomem.length);
      dbi_conn_close(conn);
      free(myarg);
      return;
    }

    /* assemble query */
    if (type == 0) {
      sprintf(sql_command, "SELECT DISTINCT t_author.author_name FROM t_author, t_xauthor WHERE t_author.author_id=t_xauthor.author_id AND t_xauthor.xauthor_type=%s AND t_author.author_name %s %s ORDER BY t_author.author_name", get_author_type_string(driver, 1), my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 1) {
      sprintf(sql_command, "SELECT DISTINCT keyword_name FROM t_keyword WHERE keyword_name %s %s ORDER BY keyword_name", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 2) {
      sprintf(sql_command, "SELECT DISTINCT periodical_abbrev, periodical_name, periodical_custabbrev1, periodical_custabbrev2 FROM t_periodical WHERE periodical_abbrev %s %s ORDER BY periodical_abbrev,periodical_name,periodical_custabbrev1,periodical_custabbrev2", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 3) {
      sprintf(sql_command, "SELECT DISTINCT periodical_name, periodical_abbrev, periodical_custabbrev1, periodical_custabbrev2 FROM t_periodical WHERE periodical_name %s %s ORDER BY periodical_abbrev,periodical_name,periodical_custabbrev1,periodical_custabbrev2", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 4) {
      sprintf(sql_command, "SELECT DISTINCT periodical_custabbrev1, periodical_name, periodical_abbrev, periodical_custabbrev2 FROM t_periodical WHERE periodical_custabbrev1 %s %s ORDER BY periodical_abbrev,periodical_name,periodical_custabbrev1,periodical_custabbrev2", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 5) {
      sprintf(sql_command, "SELECT DISTINCT periodical_custabbrev2, periodical_name, periodical_abbrev, periodical_custabbrev1 FROM t_periodical WHERE periodical_custabbrev2 %s %s ORDER BY periodical_abbrev,periodical_name,periodical_custabbrev1,periodical_custabbrev2", my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 6) {
      sprintf(sql_command, "SELECT DISTINCT t_author.author_name FROM t_author, t_xauthor WHERE t_author.author_id=t_xauthor.author_id AND t_xauthor.xauthor_type=%s AND t_author.author_name %s %s ORDER BY t_author.author_name", get_author_type_string(driver, 2), my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }
    else if (type == 7) {
      sprintf(sql_command, "SELECT DISTINCT t_author.author_name FROM t_author, t_xauthor WHERE t_author.author_id=t_xauthor.author_id AND t_xauthor.xauthor_type=%s AND t_author.author_name %s %s ORDER BY t_author.author_name", get_author_type_string(driver, 3), my_dbi_conn_get_cap(conn, "rlike"), myarg);
    }

    free(myarg);

    /* send query */
    LOG_PRINT(LOG_DEBUG, sql_command);
    dbires = dbi_conn_query(conn, sql_command);
    if (!dbires) {
      LOG_PRINT(LOG_WARNING, "getfoo failed");
      if (ptr_clrequest->n_cgi) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
      }
      tiwrite(ptr_clrequest->fd, "getfoo failed\n", TERM_NO);
      dbi_conn_close(conn);
      free(sql_command);
      return;
    }

    free(sql_command);

    if (ptr_clrequest->n_cgi) {
      /* load cgi templates */
      html_head = load_html("refdbqp_head", refdblib);
      html_foot = load_html("refdbqp_foot", refdblib);
      if (!html_head || !html_foot) {
	iwrite(ptr_clrequest->fd, cgihead_plain.text, cgihead_plain.length);
	tiwrite(ptr_clrequest->fd, "could not load cgi templates\n", TERM_NO);
	free(html_head);
	free(html_foot);
	dbi_result_free(dbires);
	dbi_conn_close(conn);
	return;
      }
      iwrite(ptr_clrequest->fd, cgihead_html.text, cgihead_html.length);
      tiwrite(ptr_clrequest->fd, html_head, TERM_NO);

      switch (type) {
      case 1:
	tiwrite(ptr_clrequest->fd, "<thead><tr><th class=\"result\">Matching keywords</th></tr></thead><tbody>", TERM_NO);
	break;
      case 2: /* fall through for all journal types */
      case 3:
      case 4:
      case 5:
	tiwrite(ptr_clrequest->fd, "<thead><tr><th class=\"result\">Matching journals</th></tr></thead><tbody>", TERM_NO);
	break;
      case 6: /* fall through for all author/editor types */
      case 7:
      case 0: 
      default: /* author is default */
	tiwrite(ptr_clrequest->fd, "<thead><tr><th class=\"result\">Matching authors</th></tr></thead><tbody>", TERM_NO);
	break;
      }

      free(html_head);
    }

    /* send results to client */
    while (dbi_result_next_row(dbires)) {
      if (!all) {
	foostring = dbi_result_get_string_idx(dbires, 1); /* 1-base index */
	if (foostring) {
	  if (ptr_clrequest->n_cgi) {
	    tiwrite(ptr_clrequest->fd, "<tr><td class=\"result\">", TERM_NO);
	  }
	  iwrite(ptr_clrequest->fd, foostring, strlen(foostring));
	  if (ptr_clrequest->n_cgi) {
	    tiwrite(ptr_clrequest->fd, "</td></tr>", TERM_NO);
	  }
	}
      }
      else {
	/* display all synonyms. If one is missing, we still print
	   the colon to indicate which is missing */
	foostring = dbi_result_get_string_idx(dbires, 1); /* 1-base index */
	if (foostring) {
	  if (ptr_clrequest->n_cgi) {
	    tiwrite(ptr_clrequest->fd, "<tr><td class=\"result\">", TERM_NO);
	  }
	  tiwrite(ptr_clrequest->fd, foostring, TERM_NO);
	}

	tiwrite(ptr_clrequest->fd, ":", TERM_NO);
	foostring = dbi_result_get_string_idx(dbires, 2); /* 1-base index */
	if (foostring) {
	  tiwrite(ptr_clrequest->fd, foostring, TERM_NO);
	}

	tiwrite(ptr_clrequest->fd, ":", TERM_NO);
	foostring = dbi_result_get_string_idx(dbires, 3); /* 1-base index */
	if (foostring) {
	  tiwrite(ptr_clrequest->fd, foostring, TERM_NO);
	}

	tiwrite(ptr_clrequest->fd, ":", TERM_NO);
	foostring = dbi_result_get_string_idx(dbires, 4); /* 1-base index */
	if (foostring) {
	  tiwrite(ptr_clrequest->fd, foostring, TERM_NO);
	}
	tiwrite(ptr_clrequest->fd, "|", TERM_NO);
	if (ptr_clrequest->n_cgi) {
	  tiwrite(ptr_clrequest->fd, "</td></tr>", TERM_NO);
	}
      }
      tiwrite(ptr_clrequest->fd, "\n", TERM_NO);
    }

    if (ptr_clrequest->n_cgi) {
      tiwrite(ptr_clrequest->fd, html_foot, TERM_NO);
      free(html_foot);
    }

    dbi_result_free(dbires);
    dbi_conn_close(conn);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_refdb_admin(): tests whether a user is likely to be a refdbd
                    administrator. We test whether the user has read
		    access to a system table readable only for
		    database administrators
		     
  int is_refdb_admin returns 0 if access is denied. returns 1 if
               access is allowed. If a database error occurs other
	       than denial of access, it will also return 0.

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int is_refdb_admin(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn;
  dbi_result dbires;
  int error;
  char outbuffer[256];
  const char *drivername;

  /* connect to database server */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    iwrite(ptr_clrequest->fd, connerr_n.text, connerr_n.length);
    LOG_PRINT(LOG_ERR, connerr.text);
    return 0;
  }

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

  /* try to access a system table that only db administrators
     may access */
  if (!strcmp(drivername, "mysql")) {
    dbires = dbi_conn_query(conn, "SELECT User FROM mysql.user");
  }
  else if (!strcmp(drivername, "pgsql")) {
    dbires = dbi_conn_query(conn, "SELECT * FROM pg_shadow");
  }
  else {
    /* no access control */
    dbi_conn_close(conn);
    return 1;
  }

  if (!dbires) {
    tiwrite(ptr_clrequest->fd, "access denied\n", TERM_NO);
    sprintf(outbuffer, "unauthorized administration attempt by %s detected", ptr_clrequest->username);
    LOG_PRINT(LOG_WARNING, outbuffer);
    dbi_conn_close(conn);
    return 0;
  }
  dbi_result_free(dbires);
  dbi_conn_close(conn);
  return 1;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  daemonize(): puts the process properly into the background 

  int daemonize returns 1 in case of an error, 0 if all went fine

  char* progname optional string containing the name of the program

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int daemonize(char* progname) {
  /* this whole fn is more or less stolen from the libslack package */
  /* libslack - http://libslack.org/ */
  /* Copyright (C) 1999-2001 raf <raf@raf.org> (GPL) */


  pid_t pid;
  long nopen;
  int fd;

  /*
  ** Don't setup a daemon-friendly process context
  ** if started by init(8) or inetd(8).
  */

  if (!(daemon_started_by_init() || daemon_started_by_inetd())) {
      /*
      ** Background the process.
      ** Lose process group leadership.
      */

    if ((pid = fork()) == -1) {
      return 1;
    }

    if (pid) {
      exit(0);
    }

    /* Become a process session leader. */

    setsid();

#ifndef NO_EXTRA_SVR4_FORK
#ifdef SVR4
    /*
    ** Lose process session leadership
    ** to prevent gaining a controlling
    ** terminal in SVR4.
    */

    if ((pid = fork()) == -1) {
      return 1;
    }

    if (pid) {
      exit(0);
    }
#endif
#endif
  }

  /* Enter the root directory to prevent hampering umounts. */

  if (chdir(ROOT_DIR) == -1) {
    return 1;
  }

  /* Clear umask to enable explicit file modes. */

  umask(0);

  /*
  ** We need to close all open file descriptors. Check how
  ** many file descriptors we have (If indefinite, a usable
  ** number (1024) will be returned).
  */

  if ((nopen = limit_open()) == -1) {
    return 1;
  }

  /*
  ** Close all open file descriptors. If started by inetd,
  ** we don't close stdin, stdout and stderr.
  ** Don't forget to open any future tty devices with O_NOCTTY
  ** so as to prevent gaining a controlling terminal
  ** (not necessary with SVR4).
  */

  if (daemon_started_by_inetd()) {
    for (fd = 0; fd < nopen; ++fd) {
      switch (fd) {
      case STDIN_FILENO:
      case STDOUT_FILENO:
      case STDERR_FILENO:
	break;
      default:
	close(fd);
      }
    }
  }
  else {
    for (fd = 0; fd < nopen; ++fd) {
      close(fd);
    }

    /*
    ** Open stdin, stdout and stderr to /dev/null just in case some
    ** code buried in a library somewhere expects them to be open.
    */

    if ((fd = open("/dev/null", O_RDWR)) == -1) {
      return 1;
    }

    /*
    ** This is only needed for very strange (hypothetical)
    ** POSIX implementations where STDIN_FILENO != 0 or
    ** STDOUT_FILE != 1 or STERR_FILENO != 2 (yeah, right).
    */

    if (fd != STDIN_FILENO) {
      if (dup2(fd, STDIN_FILENO) == -1) {
	return 1;
      }

      close(fd);
    }
    
    if (dup2(STDIN_FILENO, STDOUT_FILENO) == -1) {
      return 1;
    }

    if (dup2(STDIN_FILENO, STDERR_FILENO) == -1) {
      return 1;
    }
  }

  /* ToDo: deal with this later */
  /* Place our process id in the file system and lock it. */

/*    if (name) */
/*      { */
/*        int rc; */

/*        ptry(pthread_mutex_lock(&g.lock)) */
/*  	rc = daemon_pidfile(name); */
/*        ptry(pthread_mutex_unlock(&g.lock)) */

/*  	return rc; */
/*      } */

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  daemon_started_by_init(): returns true if the daemon was started
                            by init(8)

  static int daemon_started_by_init returns 1 if the daemon was started by init
                             returns 0 if the daemon was started by
                             any other means

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int daemon_started_by_init(void) {
  /* this code also taken from libslack */
/*    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER; */
  static int rc = -1;

  if (rc == -1) {
/*        ptry(pthread_mutex_lock(&lock)) */

    if (rc == -1) {
      rc = (getppid() == 1);
    }

/*        ptry(pthread_mutex_unlock(&lock)) */
  }

  return rc;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  daemon_started_by_inetd(): returns true if the daemon was started
                             by inetd(8)

  static int daemon_started_by_inetd returns 1 if the daemon was started by
                             intetd. returns 0 otherwise.

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int daemon_started_by_inetd(void) {
/* this code also taken from libslack */
/*    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER; */
  static int rc = -1;

  if (rc == -1) {
    /* socklen_t is not available on all platforms (OSX). We define it
     as 'int' at the top of this file if it is missing */
    socklen_t optlen = (socklen_t)sizeof(int);
    int optval;

/*      ptry(pthread_mutex_lock(&lock)) */

    if (rc == -1) {
      rc = (getsockopt(STDIN_FILENO, SOL_SOCKET, SO_TYPE, &optval, &optlen) == 0);
    }

/*      ptry(pthread_mutex_unlock(&lock)) */
  }

  return rc;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  limit_open(): Returns the maximum number of files that a process can
                have open at any time. If indeterminate, a usable
                guess (1024) is returned.

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static long limit_open(void) {
  long value;

  if ((value = sysconf(_SC_OPEN_MAX)) == -1) {
    return 1024L; /* educated guess */
  }

  return value;
}


