/*+++++++++++++++++
  refdbib - the refdb bibliography client console application
  markus@mhoenicka.de 9-25-00
  $Id: refdbib.c,v 1.41.2.2 2004/11/27 23:15:40 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   +++++++++++++++++*/



/* general includes */
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <netdb.h>

#include <string.h>

#include <unistd.h> 
#include <limits.h>
#include <syslog.h>

#include <signal.h>

#include "getopt.h"

/* our own stuff */
#include "refdb.h"  /* common stuff for all refdb applications*/
#include "linklist.h" /* for linked lists */
#include "pref.h" /* for init file, depends on linklist.h */
#include "strfncs.h" /* for is functions, LOG_PRINT definition */
#include "readln.h"  /* readline-related  stuff, need only for command dummy */
#include "page.h" /* pager functions */
#include "refdb-client.h" /* stuff common to all clients */
#include "refdbib.h" /* stuff specific to this client */
#include "readris.h" /* fncs to access RIS files */
#include "connect.h" /* modified read/write for sockets */
#include "tokenize.h" /* cuts command line into tokens */

/* Globals */

char current_db[PREFS_BUF_LEN] = "";

/*+ this array will hold the user preferences +*/
Prefs prefs[19] = {
  {"serverip", ""},
  {"port", ""},
  {"verbose", ""},
  {"pager", ""},
  {"username", ""},
  {"passwd", ""},
  {"defaultdb", ""},
  {"timeout", ""},
  {"outtype", ""},
  {"outformat", ""},
  {"logfile", ""},
  {"logdest", ""},
  {"loglevel", ""},
  {"stylespecdir", ""},
  {"refdblib", ""},
  {"startnumber", ""},
  {"toencoding", ""},
  {"ignore_missing", ""},
  {"", ""}
};

/* these are the configurable variables with the compile-time defaults */
char server_ip[PREFS_BUF_LEN] = "127.0.0.1"; /*+ default IP address of refdbd +*/
char port_address[PREFS_BUF_LEN] = "9734"; /*+ default port address of refdbd +*/
char the_pager[PREFS_BUF_LEN] = "stdout"; /*+ default pager +*/
char username[PREFS_BUF_LEN] = ""; /*+ default username (emtpy) +*/
char passwd[PREFS_BUF_LEN] = "*"; /*+ default password (ask user) +*/
char refdb_timeout[PREFS_BUF_LEN] = "180"; /* 180 seconds default timeout */
char output_type[PREFS_BUF_LEN] = "db31"; /* docbook is default output fmt */
char output_format[PREFS_BUF_LEN] = ""; /* format specification */
char log_file[PREFS_BUF_LEN] = "/var/log/refdbib.log"; /*+ default log file +*/
char log_dest[PREFS_BUF_LEN] = "1"; /*+ default log destination (0 = stderr, 1 = syslog, 2 = log_file +*/
char log_level[PREFS_BUF_LEN] = "6"; /*+ default level up to which messages are logged (0 through 7). -1 means no logging +*/
char stylespec_dir[PREFS_BUF_LEN] = "."; /* use PWD for stylespec files */
char refdblib[PREFS_BUF_LEN] = ""; /* path to shareable files */
char verbose[PREFS_BUF_LEN] = ""; /* verbose output? */
char startnumber[PREFS_BUF_LEN] = ""; /* first number of reference */
char encoding[PREFS_BUF_LEN] = ""; /* output encoding */
char ignore_missing[PREFS_BUF_LEN] = "f"; /* ignore missing refs */
char confdir[_POSIX_PATH_MAX+1] = ""; /* path to the config files */

int n_refdb_timeout;
int n_broken_pipe; /*+ 1 indicates that we attempted to write to a broken pipe +*/
int n_abort_connect; /*+ 1 indicates that we want to abort a connection +*/
int n_log_dest = 1; /* destination of log output */
int n_log_level = 0; /* level of log information that will be printed */
int n_read_stdin = 0; /* if 1, data try to squeeze in at stdin */
int n_exit_code = 0; /* any other value than 0 indicates an error */
int n_send_stylespec = 1; /* set to 0 if we don't need the style spec */
int n_cgi = 0; /* if 1, we run as a cgi app */
int n_verbose = 0; /* verbose output if 1 */
FILE* fp_log_file = NULL; /* ptr to logfile structure */

struct BSTRING readerr = {"could not read from refdbd", 26};
struct BSTRING readerr_n = {"could not read from refdbd\n", 27};
struct BSTRING writerr = {"could not write to refdbd", 25};
struct BSTRING writerr_n = {"could not write to refdbd\n", 26};

extern const char cs_term[];

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  The one and only main function
  
  int main returns 0 if successful, 1 if error
  
  int argc number of arguments

  char** argv ptr to array of strings with the command line arguments
  
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int main (int argc, char** argv)
{
  char *line, *s;
  char *processed_infile;
  char *my_refdblib;
  char msgbuf[512]; /* buffer to assemble error messages */
  char infile[_POSIX_PATH_MAX];
  int n_opt, i, j;
  int n_readinit = 1; /* if 1, read config file. If 0, skip config file */
  int n_inbytes = 0; /* number of bytes waiting on stdin */
  int n_type = 0; /* output type, numerical version
		     0 = db31
		     1 = db31x
		     2 = teix
		     3 = bibtex */

  FILE* fp_infile = NULL;
  struct stat statbuf;
  struct sigaction act, oldact, intact, oldintact;

  /* initialize signal handler */
  n_broken_pipe = 0;
  n_abort_connect = 0;

  act.sa_handler = pipehandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  intact.sa_handler = inthandler;
  sigemptyset(&intact.sa_mask);
  intact.sa_flags = 0;

  if (sigaction(SIGPIPE, &act, &oldact) != 0 ||
      sigaction(SIGINT, &intact, &oldintact) != 0) {
    fprintf(stderr, "initializing signal handlers failed\n");
    exit(1);
  }

  /* switch on reading from stdin if we receive data on stdin via a
     redirection or a pipe. */
/*   if (!isatty(fileno(stdin))) {  */
/*     n_read_stdin = 1; */
/*   } */
#ifdef FIONREAD
  if (ioctl(fileno(stdin), FIONREAD, &n_inbytes) != -1) {
    if (n_inbytes > 0) {
      n_read_stdin = 1;
    }
  }
  /* else: there was an error, and we can't safely assume that we can
     read from stdin */
#endif
  /* else: if FIONREAD is not supported, commands that accept input
     from stdin must be called with the '-f stdin' switch */
    
  /* initialize the array of preference values */
  prefs[0].varvalue = server_ip;
  prefs[1].varvalue = port_address;
  prefs[2].varvalue = verbose;
  prefs[3].varvalue = the_pager;
  prefs[4].varvalue = username;
  prefs[5].varvalue = passwd;
  prefs[6].varvalue = current_db;
  prefs[7].varvalue = refdb_timeout;
  prefs[8].varvalue = output_type;
  prefs[9].varvalue = output_format;
  prefs[10].varvalue = log_file;
  prefs[11].varvalue = log_dest;
  prefs[12].varvalue = log_level;
  prefs[13].varvalue = stylespec_dir;
  prefs[14].varvalue = refdblib;
  prefs[15].varvalue = startnumber;
  prefs[16].varvalue = encoding;
  prefs[17].varvalue = ignore_missing;

  /* a slimy hack to detect options before getopt runs */
  for (i = 0; i < argc; i++) {
    if (argv[i][0] == '-' && argv[i][1] == 'q') {
      n_readinit = 0;
      j--;
      if (!j) {
	break;
      }
    }
    if (argv[i][0] == '-' && argv[i][1] == 'y') {
      strncpy(confdir, argv[i+1], _POSIX_PATH_MAX);
      confdir[_POSIX_PATH_MAX] = '\0';
      j--;
      if (!j) {
	break;
      }
    }
  }

  if (n_readinit) {
    /* read config file settings */
    read_prefs(prefs, "refdbibrc", 0);
  }

  /* read command line settings. These may override the config file settings */
  while ((n_opt = getopt(argc, argv, "c:d:D:e:E:f:hi:l:L:mnN:p:qS:t:T:u:vVw:y:")) != -1) {
    switch (n_opt) {
    case 'c':
      strncpy(the_pager, optarg, PREFS_BUF_LEN);
      the_pager[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'd':
      strncpy(current_db, optarg, PREFS_BUF_LEN);
      current_db[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'D':
      strncpy(stylespec_dir, optarg, PREFS_BUF_LEN);
      stylespec_dir[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'e':
      strncpy(log_dest, optarg, PREFS_BUF_LEN);
      log_dest[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'E':
      strncpy(encoding, optarg, PREFS_BUF_LEN);
      encoding[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'f':
      if (!strcmp(optarg, "stdin")) {
	n_read_stdin = 1;
      }
      break;
    case 'h':
      fprintf(stderr, "Create a bibliography file based on an ID or citation key list\nThe list must be provided as an XML document using the citationlistx DTD either on stdin or by passing the file\'s path as an argument\nUsage: refdbib [-c pager ] [-d db] [-D dir] [-e logdest] [-E encoding] [-f stdin] [-h] [-i address] [-l log-level] [-L logfile] [-m] [-n] [-N number] [-p port] [-q] [-S style] [-t type] [-T time] [-u name] [-v] [-V] [-w password] [-y confdir] file\nOptions: -c pager command\n         -d use database db\n         -D save driver files in dir\n         -e log destination (0=stderr, 1=syslog, 2=custom file)\n         -E encoding set the output character encoding\n         -f stdin read data from stdin (kludge only required for Cygwin)\n         -h prints this help\n         -i set server IP address\n         -l set the log level\n         -L set the log file\n         -m ignore missing references\n         -n send no driver file\n         -N start numbering bibliography entries with number\n         -p set server port\n         -q ignore init-file\n         -S use bibliography style\n         -t output type (db31|db31x|teix|bibtex)\n         -T set timeout in seconds\n         -u set username\n         -v show version information\n         -V switch to verbose mode\n         -w set password\n         -y look for configuration files in confdir\n");
      exit (0);
      break;
    case 'i':
      strncpy(server_ip, optarg, PREFS_BUF_LEN);
      server_ip[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'l':
      strncpy(log_level, optarg, PREFS_BUF_LEN);
      log_level[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'L':
      strncpy(log_file, optarg, PREFS_BUF_LEN);
      log_file[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'm':
      *ignore_missing = 't';
      break;
    case 'n':
      n_send_stylespec = 0;
      break;
    case 'N':
      strncpy(startnumber, optarg, PREFS_BUF_LEN);
      break;
    case 'p':
      strncpy(port_address, optarg, PREFS_BUF_LEN);
      port_address[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'q':
      n_readinit = 0;
      break;
    case 'S':
      strncpy(output_format, optarg, PREFS_BUF_LEN);
      output_format[PREFS_BUF_LEN-1] = '\0';
      break;
    case 't':
      strncpy(output_type, optarg, PREFS_BUF_LEN);
      output_type[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'T':
      strncpy(refdb_timeout, optarg, PREFS_BUF_LEN);
      refdb_timeout[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'u':
      strncpy(username, optarg, PREFS_BUF_LEN);
      username[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'v':
      printf("refdbib %s markus@mhoenicka.de\nYou may redistribute and modify this software under the terms of the GNU General Public License.\n", VERSION);
      exit (0);
      break;
    case 'V':
      strcpy(verbose, "t");
      break;
    case 'w':
      strncpy(passwd, optarg, PREFS_BUF_LEN);
      passwd[PREFS_BUF_LEN-1] = '\0';
      break;
    case 'y':
      /* do nothing, this option is used before getopt runs */
      break;
    case ':':
      fprintf(stderr, "Usage: refdbib [-c pager ] [-d db] [-D dir] [-e logdest] [-E encoding] [-f stdin] [-h] [-i address] [-l log-level] [-L logfile] [-m] [-n] [-N number] [-p port] [-q] [-S style] [-t type] [-T time] [-u name] [-v] [-V] [-w password] [-y confdir] file\nOptions: -c pager command\n         -d use database db\n         -D save driver files in dir\n         -e log destination (0=stderr, 1=syslog, 2=custom file)\n         -E encoding set the output character encoding\n         -f stdin read data from stdin (kludge only required for Cygwin)\n         -h prints this help\n         -i set server IP address\n         -l set the log level\n         -L set the log file\n         -m ignore missing references\n         -n send no driver file\n         -N start numbering bibliography entries with number\n         -p set server port\n         -q ignore init-file\n         -S use bibliography style\n         -t output type (db31|db31x|teix|bibtex)\n         -T set timeout in seconds\n         -u set username\n         -v show version information\n         -V switch to verbose mode\n         -w set password\n         -y look for configuration files in confdir\n");
      exit (1);
      break;
    case '?':
      fprintf(stderr, "unknown option %c: use refdbib -h to display usage\n", optopt);
      break;
    }
  }

  /* a smart but simple hack to hide the password in the ps ax output */
  for (i = 0; i < argc; i++) {
    if (argv[i][0] == '-' && argv[i][1] == 'w') {
      j = 0;
      while (argv[i+1][j]) {
	argv[i+1][j] = 'x';
	j++;
      }
      break;
    }
  }

  /* post-process configuration variables */

  if (*verbose == 't') {
    n_verbose = 1;
  }

  i = check_ip(server_ip); /* reuse i */
  if (i>0) {
    if (i==1) {
      fprintf(stderr, "\'%s\' cannot be resolved as a hostname\n", server_ip);
    }
    else if (i==2) {
      fprintf(stderr, "\'%s\' does not appear to be an IP host\n", server_ip);
    }
    else if (i==3) {
      fprintf(stderr, "\'%s\' does not appear to have a valid IP address\n", server_ip);
    }
    else if (i==4) {
      fprintf(stderr, "\'%s\' has more than one network interface. Please specify one IP address\n", server_ip);
    }
    exit (1);
  }
  else if (i==-1) {
    /* silently change 'localhost' to '127.0.0.1' */
    strcpy(server_ip, "127.0.0.1");
  }

  if (!is_port(port_address)) {
    fprintf(stderr, "option -p needs a value: %s is no valid port\n", optarg);
    if (*verbose == 't') {
      fprintf(stderr, "Port addresses below 1024 are reserved for system use. refdb should use a port higher than 1024. The server and all clients must use the same port.\n");
    }
    exit (1);
  }

  /* see whether we have a username */
  if (!username[0] && getlogin()) {
    strcpy(username, getlogin()); /* although not recommended, the login name
				     is a good guess */
  }

  /* see whether we need to ask for a password */
  if (strcmp(passwd, "*") == 0) {
    ask_for_passwd(passwd);
  }

  /* make some conversions */
  n_log_level = num_loglevel(log_level);
  n_log_dest = num_logdest(log_dest);
  n_refdb_timeout = atoi(refdb_timeout);

  /* fill in stylespec_dir if not explicitly set */
  if (!*stylespec_dir) {
    /* use REFDBLIB if possible */
    if (!*refdblib) {
      my_refdblib = getenv("REFDBLIB");
    }
    else {
      my_refdblib = refdblib;
    }
    if (my_refdblib && *my_refdblib) {
      sprintf(stylespec_dir, "%s/stylespec/", refdblib);
    }
    else {
      stylespec_dir[0] = '\0'; /* use PWD if nothing else helps */
    }
  }
  else if (!strcmp(stylespec_dir, ".")) {
    stylespec_dir[0] = '\0'; /* use PWD if the path is a single dot */
  }

  if (!strcmp(output_type, "db31")) {
    n_type = 0;
  }
  else if (strcmp(output_type, "db31x") == 0) {
    n_type = 1;
  }
  else if (strcmp(output_type, "teix") == 0) {
    n_type = 2;
  }
  else if (strcmp(output_type, "bibtex") == 0) {
    n_type = 3;
  }
  /* maybe there'll be RTF output someday */
/*    else if (strcmp(output_type, "rtf") == 0) { */
/*      n_type = 4; */
/*    } */
  else {
    n_type = 0; /* temporary default, may change or be removed */
  }

/*   fprintf(stderr, "username: %s; password: %s\n", username, passwd); */
  
  /* set up logging */
  if (n_log_dest == 2) { /* use custom log file */
    if ((fp_log_file = fopen(log_file, "ab")) == NULL) {
      n_log_dest = 1; /* fall back to syslog */
      openlog("refdbib", LOG_PID|LOG_ODELAY, LOG_USER);
      LOG_PRINT(LOG_WARNING, "could not open custom log file");
    }
  }
  else if (n_log_dest == 1) { /* use syslog */
    openlog("refdbib", LOG_PID|LOG_ODELAY, LOG_USER);
  }

  LOG_PRINT(LOG_INFO, "refdbib started");

  /* some checks */
  if (*current_db == '\0') {
    LOG_PRINT(LOG_CRIT, "no database specified");
    n_exit_code = 127;
  }
  else {
    if (n_read_stdin) {
      if (n_type == 3) { /* bibtex */
	n_exit_code = make_texbib(stdin);
      }
      else { /* 0 is default */
	n_exit_code = make_dbib(stdin, n_type);
      }
    }
    else if (argv[optind] && *(argv[optind])) { /* read from file */
      strncpy(infile, argv[optind], (size_t)_POSIX_PATH_MAX-1);
      infile[_POSIX_PATH_MAX-1] = '\0'; /* terminate just in case */

      processed_infile = canonicalize_path(infile);
      if (processed_infile == NULL) {
	if (n_type == 3) { /* bibtex */
	  strcat(infile, ".aux");
	  processed_infile = canonicalize_path(infile);
	  if (processed_infile == NULL) {
	    sprintf(msgbuf, "could not access input file %s", argv[optind]);
	    LOG_PRINT(LOG_WARNING, msgbuf);
	    n_exit_code = 1;
	    goto the_end;
	  }
	}
	else {
	  sprintf(msgbuf, "could not access input file %s", argv[optind]);
	  LOG_PRINT(LOG_WARNING, msgbuf);
	  n_exit_code = 1;
	  goto the_end;
	}
      }
      strcpy(infile, processed_infile);
      free(processed_infile);

      if ((fp_infile = fopen(infile, "rb")) == NULL) {
	sprintf(msgbuf, "could not open input file %s", argv[optind]);
	LOG_PRINT(LOG_WARNING, msgbuf);
	n_exit_code = 1;
      }
      else {
	if (n_type == 3) {
	  n_exit_code = make_texbib(fp_infile);
	}
	else { /* 0 is default */
	  n_exit_code = make_dbib(fp_infile, n_type);
	}
	fclose(fp_infile);
      }
    }
    /* else: no file specified, do nothing */
  }
    
 the_end:
  if (n_exit_code) {
    LOG_PRINT(LOG_INFO, "refdbib finished with errors");
  }
  else {
    LOG_PRINT(LOG_INFO, "refdbib finished");
  }
  
  if (fp_log_file) {
    fclose(fp_log_file); /* not strictly necessary, but more polite */
  }
  exit (n_exit_code);
}


/* **************************************************************** */
/*                                                                  */
/*                   refdbib functions                              */
/*                                                                  */
/* **************************************************************** */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  make_dbib() requests a docbook bibliography

  int make_dbib returns 0 if ok, 1 if error

  FILE *fp_infile pointer to a stream to read the id data from

  int n_type type of requested output: 0 = db31, 1 = db31x,
		     2 = teix, (3 = bibtex)
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int make_dbib(FILE* fp_infile, int n_type) {
  int n_sockfd; /* file descriptor of the socket */
  char *outbuffer;
  char *new_outbuffer;
  char inbuffer[COMMAND_INBUF_LEN];
  char srv_inbuffer[COMMAND_INBUF_LEN] = "";
  char specfile[_POSIX_PATH_MAX];
  char msg_buffer[OUTBUF_LEN];
  char read_buffer[256];
  char *read_result;
  char *stylesenv;
  char thebytes[11];
  char scrambled_passwd[PASSWD_LENGTH*3+1];
  size_t outbuf_len = COMMAND_INBUF_LEN;
  int numbyte;
  int n_read_done = 0;
  int n_result;
  int i;
  int n_xmlfile_done = 0;
  int n_chunk_count = 0;
  int n_retval = 0;
  int n_curr_trailing_z = 0;
  int n_last_trailing_z = 0;
  size_t byte_written = 0;
  FILE *pagerfp;
  FILE *specfilefp;
  struct linger mylinger;

  outbuffer = (char*)malloc(outbuf_len);
  if (outbuffer == NULL) {
    return 1;
  }

  strcpy(outbuffer, "getbib ");

  if (connect_to_server(&n_sockfd, server_ip, port_address) != 0) {
    LOG_PRINT(LOG_ERR, "could not connect to server");
    free(outbuffer);
    return 1;
  }

  strcpy(scrambled_passwd, passwd);

  if (init_dialog(n_sockfd, scrambled_passwd, srv_inbuffer)) {
    LOG_PRINT(LOG_ERR, "server authentication failed");
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }

  /* assemble command */
  switch (n_type) {
  case 2:
    strcat(outbuffer, " -t teix -u ");
    break;
  case 1:
    strcat(outbuffer, " -t db31x -u ");
    break;
  case 0:
    /* fall through */
  default:
    strcat(outbuffer, " -t db31 -u ");
    break;
  }

  strcat(outbuffer, username);
  if (!n_send_stylespec) {
    strcat(outbuffer, " -n");
  }
  if (*passwd) {
    strcat(outbuffer, " -w ");
    strcat(outbuffer, scrambled_passwd);
  }
  strcat(outbuffer, " -d ");
  strcat(outbuffer, current_db);
  if (*output_format) {
    strcat(outbuffer, " -s ");
    strcat(outbuffer, output_format);
  }

  if (*startnumber) {
    strcat(outbuffer, " -o ");
    strcat(outbuffer, startnumber);
  }

  if (*encoding) {
    strcat(outbuffer, " -E ");
    strcat(outbuffer, encoding);
  }

  /* now send the command to the application server */
/*      printf("%s\n", outbuffer); */
/*    free(outbuffer); */
/*    return 0; */

  LOG_PRINT(LOG_DEBUG, outbuffer);
  numbyte = tiwrite(n_sockfd, outbuffer, TERM_YES);
  if (numbyte == -1) {
    free(outbuffer);
    fprintf(stderr, writerr_n.text);
    LOG_PRINT(LOG_WARNING, writerr.text);
    close(n_sockfd);
    return 1;
  }

  numbyte = tread(n_sockfd, inbuffer, COMMAND_INBUF_LEN);
/*     fprintf(stderr, "first server reply:%s<<\n", inbuffer);    */
  if (numbyte == -1) {
    fprintf(stderr, readerr_n.text);
    LOG_PRINT(LOG_WARNING, readerr.text);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }
  
  if (n_send_stylespec) {
    if (strncmp(inbuffer, "SPEC", 4) != 0) {



      /* find define for maximum server error length. sprintf error message to log */


      numbyte = tiwrite(n_sockfd, "NEG", TERM_YES);
      sprintf(msg_buffer, "server error: %s", inbuffer);
      LOG_PRINT(LOG_WARNING, msg_buffer);
      fprintf(stderr, "server error: %s.\n", inbuffer);
      numbyte = tiwrite(n_sockfd, "NEG", TERM_YES);
      free(outbuffer);
      close(n_sockfd);
      return 1;
    }
    
    numbyte = tiwrite(n_sockfd, "POS", TERM_YES);
    if (numbyte == -1) {
      free(outbuffer);
      fprintf(stderr, writerr_n.text);
      LOG_PRINT(LOG_WARNING, writerr.text);
      close(n_sockfd);
      return 1;
    }

    if (stylespec_dir && *stylespec_dir) {
      strcpy(specfile, stylespec_dir);
      if (specfile[0] && specfile[strlen(specfile)-1] != '/') {
	strcat(specfile, "/");
      }
    }
    else {
      specfile[0] = '\0';
    }
    strcat(specfile, output_format);
    if (specfile[strlen(specfile)-1] != '.') {
      strcat(specfile, ".");
    }
    if (!n_type) {
      strcat(specfile, "dsl"); 
    }
    else {
      strcat(specfile, "xsl"); 
    }

    /* this is guaranteed to return a file ptr. This may be stdout
       if nothing else helps */
    specfilefp = open_outfile(specfile, 0);
/*      printf("specfile: %s<<\n", specfile); */

    n_read_done = 0;

    do {
      numbyte = tread(n_sockfd, inbuffer, OUTBUF_LEN);
/*        printf("in read specfile loop:%s<<\n", inbuffer); */
      if (numbyte == -1) {
	fprintf(stderr, readerr_n.text);
	LOG_PRINT(LOG_WARNING, readerr.text);
	close(n_sockfd);
	free(outbuffer);
	close_outfile(specfilefp);
	n_broken_pipe = 0;
	return 1;
      }

      n_curr_trailing_z = get_trailz(inbuffer, numbyte);

      if (numbyte >= TERM_LEN) {
	if (n_curr_trailing_z >= TERM_LEN) {
	  n_read_done++;
	}
      }
      else if (n_curr_trailing_z == numbyte
	     && n_curr_trailing_z + n_last_trailing_z >= TERM_LEN) {
	n_read_done++;
      }
      else if (n_curr_trailing_z == numbyte) {
	/* terminator is still incomplete */
	n_last_trailing_z += n_curr_trailing_z;
	continue;
      }

      if (!n_broken_pipe) {
	/* write numbyte chars to output, unless this is the last chunk:
	   we do not want to write the terminating \0 */
	if (n_last_trailing_z) {
	  byte_written += fwrite(cs_term, sizeof(char), n_last_trailing_z, pagerfp);
	}
	byte_written += fwrite(inbuffer, sizeof(char), numbyte-n_curr_trailing_z, specfilefp);
      }

      if (!n_read_done) {
	n_last_trailing_z = n_curr_trailing_z;
      }
/*        printf("%s<< byte_written:%d\n", inbuffer, byte_written); */
    } while (!n_read_done);

    close_outfile(specfilefp);

    numbyte = tiwrite(n_sockfd, "POS", TERM_YES);
    if (numbyte == -1) {
      free(outbuffer);
      fprintf(stderr, writerr_n.text);
      LOG_PRINT(LOG_WARNING, writerr.text);
      close(n_sockfd);
      return 1;
    }

    numbyte = tread(n_sockfd, inbuffer, COMMAND_INBUF_LEN);
    
    if (numbyte == -1) {
      fprintf(stderr, readerr_n.text);
      LOG_PRINT(LOG_WARNING, readerr.text);
      free(outbuffer);
      close(n_sockfd);
      return 1;
    }
  } /* end if (n_send_stylespec) */


  if (strncmp(inbuffer, "POS", 3) != 0) {
    sprintf(msg_buffer, "server error: %s", inbuffer);
    LOG_PRINT(LOG_WARNING, msg_buffer);
    fprintf(stderr, "server error: %s\n", inbuffer);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }

  /* run phases 1 through 4 of client/server protocol */
  n_result = send_xml_data(fp_infile, pagerfp, stderr, n_sockfd, &byte_written);

  if (n_result) {
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }

  /* ------------------------------------------------------------ */
  /* PHASE 5 */
  /*  signal server that we're done */
  sprintf(thebytes, "%d", 0); 
  numbyte = tiwrite(n_sockfd, thebytes, TERM_YES);
  /*        printf("%s\n", thebytes); */
  if (numbyte == -1) {
    fprintf(stderr, writerr_n.text);
    LOG_PRINT(LOG_WARNING, writerr.text);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }

  numbyte = tread(n_sockfd, inbuffer, COMMAND_INBUF_LEN);
    
  if (numbyte == -1) {
    fprintf(stderr, readerr_n.text);
    LOG_PRINT(LOG_WARNING, readerr.text);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }
  
  if (strncmp(inbuffer, "POS", 3) != 0) {
    fprintf(stderr, "server error: %s\n", inbuffer);
    sprintf(msg_buffer, "server error: %s", inbuffer);
    LOG_PRINT(LOG_WARNING, msg_buffer);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }

  numbyte = tiwrite(n_sockfd, "POS", TERM_YES);
  if (numbyte == -1) {
    fprintf(stderr, writerr_n.text);
    LOG_PRINT(LOG_WARNING, writerr.text);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }
  
  /* openpager and open_outfile are guaranteed to return a valid file/pipe - 
     and be it stdout */
  pagerfp = openpager(the_pager);

  n_read_done = 0;
  n_last_trailing_z = 0;
  n_curr_trailing_z = 0;

  do {
    numbyte = tread(n_sockfd, inbuffer, COMMAND_INBUF_LEN);
/*      printf("numbyte:%d\n", numbyte); */
    if (numbyte == -1) {
      fprintf(stderr, readerr_n.text);
      LOG_PRINT(LOG_WARNING, readerr.text);
      free(outbuffer);
      close(n_sockfd);
      closepager(pagerfp);
      n_broken_pipe = 0;
      return 1;
    }

    n_curr_trailing_z = get_trailz(inbuffer, numbyte);

    if (numbyte >= TERM_LEN) {
      if (n_curr_trailing_z >= TERM_LEN) {
	n_read_done++;
      }
    }
    else if (n_curr_trailing_z == numbyte
	     && n_curr_trailing_z + n_last_trailing_z >= TERM_LEN) {
      n_read_done++;
    }
    else if (n_curr_trailing_z == numbyte) {
      /* terminator is still incomplete */
      n_last_trailing_z += n_curr_trailing_z;
      continue;
    }

    /* write numbyte chars to output, unless this is the last chunk: we do not
       want to write the terminating \0 */
    if (!n_broken_pipe) {
      if (n_last_trailing_z) {
	byte_written += fwrite(cs_term, sizeof(char), n_last_trailing_z, pagerfp);
      }
      byte_written += fwrite(inbuffer, sizeof(char), numbyte-n_curr_trailing_z, pagerfp);
    }

    if (!n_read_done) {
      n_last_trailing_z = n_curr_trailing_z;
    }

  } while (!n_read_done);

  /* send back confirmation to the server */
  numbyte = tiwrite(n_sockfd, "POS", TERM_YES);
  if (numbyte == -1) {
    fprintf(stderr, writerr_n.text);
    LOG_PRINT(LOG_WARNING, writerr.text);
    free(outbuffer);
    close(n_sockfd);
    return 1;
  }
  
/*    printf("byte written:%d\n", byte_written); */
/*    fflush(stdout); */


  numbyte = tread(n_sockfd, inbuffer, COMMAND_INBUF_LEN);
  if (numbyte == -1) {
    fprintf(stderr, readerr_n.text);
    LOG_PRINT(LOG_WARNING, readerr.text);
  }
  else {
    fprintf(stderr, "%s\n", inbuffer);
    if (*ignore_missing == 'f' && !strncmp(inbuffer, "getbib error", 12)) {
      n_retval = 1;
    }
  }
  
  /* send back confirmation to the server */
  tiwrite(n_sockfd, "POS", TERM_YES);

  closepager(pagerfp);

  /* closing the socket in the background apparently does not work if the whole app exits. So we force the close() call to wait until the data are read by the server */
  mylinger.l_onoff = 1; /* let close wait */
  mylinger.l_linger = 5; /* 5 seconds timeout */
  setsockopt(n_sockfd, SOL_SOCKET, SO_LINGER, (const void*)&mylinger, sizeof(mylinger));
  n_broken_pipe = 0;
  free(outbuffer);
  close(n_sockfd);
  return n_retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  make_texbib() requests a LaTeX/BibTeX bibliography

  int make_texbib returns 0 if ok, 1 if error

  FILE *fp_infile pointer to a stream to read the id data from

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int make_texbib(FILE* fp_infile) {
  char *new_outbuffer = NULL;
  char cmd_buffer[OUTBUF_LEN] = "";
  char inbuffer[COMMAND_INBUF_LEN] = "";
  char *newarg = NULL;
  char* infile = NULL;
  char scrambled_passwd[PASSWD_LENGTH*3+1] = "";
  char msg_buffer[OUTBUF_LEN];
  int numbyte = 0;
  int n_read_done = 0;
  int result;
  int n_opt;
  int i, j;
  size_t outbuf_len;
  size_t byte_written = 0;
  FILE *pagerfp;
  FILE* errstream;
  struct linger mylinger;
  struct simplelistvals slvals;
  struct liliform* ptr_current;
  struct lilimem sentinel;

  errstream = stderr;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  outbuf_len = 128; /* something to start with */
  slvals.outbuffer = (char*)malloc(outbuf_len); 
  if (slvals.outbuffer == NULL) {
    return 1;
  }
  slvals.outbuffer[0] = '\0'; /* terminate string */

  slvals.n_file_open = 0;
  slvals.n_file_append = 0;
  slvals.n_pipe = 0;
  slvals.outfile = NULL;
  slvals.outpipe = NULL;

  if (insert_lilimem(&sentinel, (void**)&(slvals.outbuffer), NULL)) {
    return 1;
  }

  strcpy(cmd_buffer, "gettexbib ");

  if (connect_to_server(&(slvals.n_sockfd), server_ip, port_address) != 0) {
    LOG_PRINT(LOG_ERR, "could not connect to server");
    delete_all_lilimem(&sentinel);
    return 1;
  }

  strcpy(scrambled_passwd, passwd);

  if (init_dialog(slvals.n_sockfd, scrambled_passwd, inbuffer)) {
    LOG_PRINT(LOG_ERR, "server authentication failed");
    close(slvals.n_sockfd);
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* prepare data */
  result = add_id_from_aux(fp_infile, &(slvals.outbuffer), &outbuf_len);

  if (result) {
    LOG_PRINT(LOG_WARNING, "could not read ID data from .aux file");
    delete_all_lilimem(&sentinel);
    close(slvals.n_sockfd);
    return 1;
  }

  /* assemble command */
  strcat(cmd_buffer, " -u ");
  strcat(cmd_buffer, username);
  if (strlen(passwd) > 0) {
    strcat(cmd_buffer, " -w ");
    strcat(cmd_buffer, scrambled_passwd);
  }
  strcat(cmd_buffer, " -d ");
  strcat(cmd_buffer, current_db);

  if (*output_format) {
    strcat(cmd_buffer, " -s ");
    strcat(cmd_buffer, output_format);
  }

  sprintf(cmd_buffer+strlen(cmd_buffer), " %d", strlen(slvals.outbuffer)+1); 

/*    printf("outbuffer:%s<<\ncmd_buffer:%s<<\n", slvals.outbuffer, cmd_buffer); */

/*    delete_all_lilimem(&sentinel); */
/*    return 0; */

  /* send command to application server */
  numbyte = tiwrite(slvals.n_sockfd, cmd_buffer, TERM_YES);
  LOG_PRINT(LOG_DEBUG, cmd_buffer);
  if (numbyte == -1) {
    fprintf(stderr, writerr_n.text);
    LOG_PRINT(LOG_WARNING, writerr.text);
    close(slvals.n_sockfd);
    delete_all_lilimem(&sentinel);
    return 1;
  }

  numbyte = 0;

  /* read acknowledgement from application server */
  numbyte = tread(slvals.n_sockfd, slvals.inbuffer, OUTBUF_LEN);
  if (numbyte == -1) {
    fprintf(stderr, readerr_n.text);
    LOG_PRINT(LOG_WARNING, readerr.text);
    close(slvals.n_sockfd);
    delete_all_lilimem(&sentinel);
    return 1;
  }

  if (strncmp(slvals.inbuffer, "POS", 3) != 0) {
    sprintf(msg_buffer, "server error: %s", slvals.inbuffer);
    LOG_PRINT(LOG_WARNING, msg_buffer);
    fprintf(stderr, "server error: %s\n", slvals.inbuffer);
    delete_all_lilimem(&sentinel);
    close(slvals.n_sockfd);
    return 1;
  }

  getsimplelist(&slvals, 1);

  /* getsimplelist() retrieves the data proper. Now we read the summary
     which should not appear in the bib file but on stderr */
/*   numbyte = tread(slvals.n_sockfd, slvals.inbuffer, COMMAND_INBUF_LEN); */
    
/*   if (numbyte == -1) { */
/*     fprintf(stderr, readerr_n.text); */
/*     LOG_PRINT(LOG_WARNING, readerr.text); */
/*     delete_all_lilimem(&sentinel); */
/*     close(slvals.n_sockfd); */
/*     return 1; */
/*   } */
  
/*   fprintf(stderr, slvals.inbuffer); */

/*   iwrite(slvals.n_sockfd, "POS", 4); */

  /* closing the socket in the background apparently does not work if the whole app exits. So we force the close() call to wait until the data are read by the server */
  mylinger.l_onoff = 1; /* let close wait */
  mylinger.l_linger = 5; /* 5 seconds timeout */
  setsockopt(slvals.n_sockfd, SOL_SOCKET, SO_LINGER, (const void*)&mylinger, sizeof(mylinger));

  close(slvals.n_sockfd);
  delete_all_lilimem(&sentinel);
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  log_print(): writes a log message

  void log_print

  int priority the priority level of the log message as in syslog.h

  char* string a string containing the message to be logged

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void log_print(int priority, char* string) {
  /* we must have this fn in the file with main() because FILE* 
     cannot be declared extern (??) */
  time_t the_time;
  char timestring[256];
  FILE* new_fp_log_file;

  if (n_log_dest == 0) { /* output on stderr */
    fprintf(stderr, "%s\n", string);
  }
  else if (n_log_dest == 1) { /* output via syslog */
    syslog(priority, string);
  }
  else { /* output in user-defined logfile */
    time(&the_time);
    strftime(timestring, 256, "%a %b %d %H:%M:%S %Y", gmtime(&the_time));
    fprintf(fp_log_file, "%d:pid=%d:%s:%s\n", priority, getpid(), timestring, string);
  }
}













